/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: AVEncoder.cpp,v 1.3 2004/03/27 13:47:16 gadibergman Exp $
 */

/************************************************************
				AVEncoder.cpp
				=============

  An object responsible of integrating:
  - audio input
  - video input
  - an audio encoder
  - a video encoder
  - an AV-Format writer

  This object uses
  - AV Format writer -- FFMPEG implementation of AV Format writer
  - audio encoder    -- whatever implements CAudioEncoder
  - video encoder    -- FFMPEG implementation of MPEG-4 Video
                        Used internally by AV Format writer

************************************************************/

// PCH includes
#include "StdAfx.h"

// Usual includes
#include "MPEGRecorder.h"
#include "AVEncoder.h"

// MPEG Includes
#include "AVFormatWriter.h"
#include "VideoCodecTrackWriter.h"

// DirectX based receivers
#include "DXSoundReceiver.h"
#include "DXVideoReceiver.h"

// Project includes
#include "TimeBasedPipe.h"
#include "FileNameGenerator.h"
#include "MediaDivision.h"

// Encoders includes
#include "AudioEncoder.h"
#include "AudioEncoderGenerator.h"
#include "rgb2yuv/RGB2YUVConvertor.h"
#include "RGBBlockCopier.h"
#include "XVidEncoder.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAVEncoder::~CAVEncoder()
{
	m_pAudioProvider->Stop();
	m_pVideoProvider->Stop();

	delete m_pAudioEncoder;
	delete m_pRGBBlockCopier;

	delete m_pPCMAudioPipe;
	delete m_pRGBVideoPipe;

	delete m_pAudioTrackWriter;
	delete m_pVideoTrackWriter;
	delete m_pAVFormatWriter;
}

void CAVEncoder::Init()
{
	CString sDirectory, sFileName;
	m_FileNameGenerator.GenerateFileName(sDirectory, sFileName);

	CString sFilePath = sDirectory;
	sFilePath += sFileName;

	m_pAVFormatWriter = new CAVFormatWriter(
		m_pSoundReceiver->GetSampleRate(),
		m_pSoundReceiver->GetChannels(),
		m_AudioEncoderGenerator.GetBitRate(),
		m_pVideoReceiver->GetWidth(),
		m_pVideoReceiver->GetHeight(),
		m_pVideoReceiver->GetFrameRate(),
		m_uDesiredFrameRate,
		sFilePath);

	m_pVideoTrackWriter = 
		m_pAVFormatWriter->CreateVideoCodecTrackWriter(CODEC_ID_MPEG4);

	m_pAudioTrackWriter = 
		m_pAVFormatWriter->CreateTrackWriter(CODEC_TYPE_AUDIO, m_AudioEncoderGenerator.GetCodecID());

	m_pPCMAudioPipe = new CTimeBasedPipe();
	m_pRGBVideoPipe = new CTimeBasedPipe();

	enum {
		RGB_ALLOCATION_SIZE_LIMIT = 10000000
	};

	m_pRGBVideoPipe->SetAllocationSizeLimit( RGB_ALLOCATION_SIZE_LIMIT );

	m_pAudioEncoder = 
		m_AudioEncoderGenerator.CreateAudioEncoder(		
			m_pPCMAudioPipe,
			m_pAudioTrackWriter,
			m_pSoundReceiver->GetChannels(),
			m_pSoundReceiver->GetBytesPerSample(),
			m_pSoundReceiver->GetSampleRate(),
			CAudioEncoderGenerator::AVI_FILE_CONTAINER
			);

	m_pRGBBlockCopier = new CRGBBlockCopier(
		m_pRGBVideoPipe,
		m_pVideoTrackWriter,
		m_pVideoReceiver->GetWidth(),
		m_pVideoReceiver->GetHeight()
		);

	m_pAVFormatWriter->Start();

	m_pAudioProvider = m_pSoundReceiver->CreateMediaProvider(m_pPCMAudioPipe);
	m_pVideoProvider = m_pVideoReceiver->CreateMediaProvider(m_pRGBVideoPipe);

}

void CAVEncoder::GetProgress(
	double&	time,
	LONGLONG &pcmBytes, 
	LONGLONG &aacBytes, 
	LONGLONG &rgbBytes, 
	LONGLONG &yuvBytes, 
	LONGLONG &xvidBytes
	) const
{
	time = m_pAVFormatWriter->GetTime();
	pcmBytes = m_pPCMAudioPipe->GetNumBytesWritten();
	aacBytes = m_pAudioTrackWriter->GetNumBytesWritten();
	rgbBytes = m_pRGBVideoPipe->GetNumBytesWritten();
	yuvBytes = m_pVideoTrackWriter->GetNumYUVBytesWritten();
	xvidBytes = m_pVideoTrackWriter->GetNumXVidBytesWritten();
}

void CAVEncoder::Stop()
{
	m_pAudioProvider->Stop();
	m_pVideoProvider->Stop();	
}

void CAVEncoder::PlayOrPause()
{
	m_pAudioProvider->PlayOrPause();
	m_pVideoProvider->PlayOrPause();
}
