/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: AVFormatWriter.cpp,v 1.3 2004/02/20 08:50:34 gadibergman Exp $
 */

/************************************************************
				AVFormatWriter.cpp
				==================

  The combination of two channels - one video and one audio
  into one file, such as Microsoft's AVI.

  Uses FFMPEG library (libmedia.dll).
************************************************************/

#include "stdafx.h"
#include "AVFormatWriter.h"
#include <assert.h>
#include "ffmpeg/avformat.h"
#include "TimeKeepingReaderWriter.h"
#include "TimeBasedProgressSelection.h"
#include "mpeg4/combined/media_time.h"
#include "ProgramLog.h"
#include "VideoCodecTrackWriter.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

class FFMPEGAccess {
public:
	static FFMPEGAccess& Get() {
		static FFMPEGAccess obj;
		return obj;
	}

	FFMPEGAccess() {
		av_register_all();
	}

	enum {
		VIDEO_STREAM_INDEX = 0,
		AUDIO_STREAM_INDEX = 1
	};

    AVOutputFormat* GetFormatByFileName(const char * filename) {
		AVOutputFormat* pFormat = guess_format(NULL, filename, NULL);
		assert(pFormat);
		return pFormat;
	}

	AVFormatContext* AllocateFormatContext(AVOutputFormat* pFormat) {
		AVFormatContext* pContext = (AVFormatContext*) av_mallocz(sizeof(AVFormatContext));
		assert(pContext);
		pContext->oformat = pFormat;
		return pContext;
	}

	void SetOutputFileName(AVFormatContext* pContext, const char * filename) {
		snprintf(pContext->filename, 
			sizeof(pContext->filename), "%s", filename);
	}

	void OpenOutputFile(AVOutputFormat* pFormat, 
		AVFormatContext* pContext, const char * filename) {
		if (!(pFormat->flags & AVFMT_NOFILE)) {
			int iRes = url_fopen(&pContext->pb, filename, URL_WRONLY);
			assert(iRes >= 0);
		}
    }

	void CloseOutputFile(AVOutputFormat* pFormat, 
		AVFormatContext* pContext) {
		if (!(pFormat->flags & AVFMT_NOFILE)) {
			url_fclose(&pContext->pb);
		}
	}

	void WriteContextHeader(AVFormatContext* pContext) {
		av_write_header(pContext);
	}

	void WriteContextTrailer(AVFormatContext* pContext) {
		av_write_trailer(pContext);
	}

	void SetContextParameters(AVFormatContext* pContext) {
		int iRes = av_set_parameters(pContext, NULL);
		assert(iRes >= 0);
	}

	void FreeAllocatedMemory(void* p) {
		av_free(p);
	}

	bool FormatSupportsVideo(AVOutputFormat* pFormat) {
		return pFormat->video_codec != CODEC_ID_NONE;
	}

	bool FormatSupportsAudio(AVOutputFormat* pFormat) {
		return pFormat->audio_codec != CODEC_ID_NONE;
	}

	AVStream* AddVideoStream(AVFormatContext* pContext,
		CodecID codecID, int iFrameWidth, int iFrameHeight,
		double dFrameRate) {

		AVCodecContext *pCodec;
		AVStream *pStream;

		pStream = av_new_stream(pContext, VIDEO_STREAM_INDEX);

		assert(pStream);

		pCodec = &pStream->codec;
		pCodec->codec_id = codecID;
		pCodec->codec_type = CODEC_TYPE_VIDEO;

		/* put sample parameters */
		//pCodec->bit_rate = 400000;
		pCodec->bit_rate = iFrameWidth * iFrameHeight * 5;

		/* resolution must be a multiple of two */
		pCodec->width = iFrameWidth;  
		pCodec->height = iFrameHeight;
		/* frames per second */
		pCodec->frame_rate = 1000 * dFrameRate;  
		pCodec->frame_rate_base = 1000;
		pCodec->gop_size = 12; /* emit one intra frame every twelve frames */

		return pStream;
	}

	AVStream* AddAudioStream(AVFormatContext* pContext, 
		CodecID codecID, int iChannels, int iSampleRate,
		int iBitRate) {

		AVCodecContext *pCodec;
		AVStream *pStream;

		pStream = av_new_stream(pContext, AUDIO_STREAM_INDEX);

		assert(pStream);

		pCodec = &pStream->codec;
		pCodec->codec_id = codecID;
		pCodec->codec_type = CODEC_TYPE_AUDIO;

		/* put sample parameters */
		pCodec->bit_rate = iBitRate;
		pCodec->sample_rate = iSampleRate;
		pCodec->channels = iChannels;

		return pStream;
	}

	void FreeStreams(AVFormatContext * pContext)
	{
		for(int i = 0; i < pContext->nb_streams; i++) {
			av_freep(&pContext->streams[i]);
		}
	}

	void WriteFrame(AVFormatContext *pContext, 
		int iStreamIndex, const void * buf, int iSize)
	{
		int writeResult = av_write_frame(pContext, 
			iStreamIndex, (const uint8_t*) buf, 
			iSize);

		assert(writeResult >= 0);
	}

	void WriteAudio(AVFormatContext *pContext,
		const void * buf, int iSize)
	{
		int writeResult = av_write_frame(pContext, 
			AUDIO_STREAM_INDEX, (const uint8_t*) buf, 
			iSize);

		assert(writeResult >= 0);
	}

	unsigned long GetNumberOfExpectedFrames(
		AVStream* pStream, Timestamp timestamp, Duration& dExpectedDuration)
	{
		double dFrameRate = (double) pStream->codec.frame_rate / 
			pStream->codec.frame_rate_base;

		double dSeconds = (double) (int64_t) timestamp / 
			TimestampTicks;

		dExpectedDuration = (double) (int64_t) TimestampTicks / dFrameRate;

		return dFrameRate * dSeconds;
	}

	void WriteVideoFrame(AVFormatContext *pContext,
		const void * buf, int iSize, 
		Timestamp timestamp, unsigned long uVideoFramesNumber)
	{
		AVStream* pStream = pContext->streams[VIDEO_STREAM_INDEX];
		int64_t pts_mask = ((int64_t) 1 << pContext->pts_wrap_bits) - 1;

		double dFrameRate = (double) pStream->codec.frame_rate / 
			pStream->codec.frame_rate_base;

		double dSeconds = (double) (int64_t) timestamp / 
			TimestampTicks;

		LOG_STREAM("Time = " << dSeconds << " seconds; "
			"Frame rate: " << (double) uVideoFramesNumber / dSeconds << 
			" frames/sec");

		pStream->pts.val = (int) ( dFrameRate * dSeconds );

		int iWriteRes = pContext->oformat->write_packet(
			pContext, VIDEO_STREAM_INDEX, (unsigned char *)buf, iSize, 
			pStream->pts.val & pts_mask);
		
		assert(iWriteRes >= 0);	
	}

	int EncodeVideo(AVCodecContext *pCodecContext,
		uint8_t *pOutputFrame, int iOutputFrameSize,
		const AVFrame *pFrame)
	{
		return avcodec_encode_video(pCodecContext, 
			pOutputFrame, iOutputFrameSize, pFrame);
	}

	AVFrame *AllocatePicture(int pix_fmt, 
			int width, int height, int& size)
	{
		AVFrame *pFrame = avcodec_alloc_frame();

		if (!pFrame)
			return NULL;

		size = avpicture_get_size(pix_fmt, width, height);

		return pFrame;
	}

	AVCodec* FindEncoderByCodec(CodecID codecID)
	{
		return avcodec_find_encoder(codecID);
	}

	virtual ~FFMPEGAccess() {

	}

};

CAVFormatWriter::CAVFormatWriter(
		unsigned int uAudioSampleRate,
		unsigned int uAudioChannels,
		unsigned long uAudioBitRate,
		unsigned short uFrameWidth,
		unsigned short uFrameHeight,
		double dFrameRate,
		unsigned int uDesiredFrameRate,
		LPCTSTR sFileName):
m_uAudioSampleRate(uAudioSampleRate),
m_uAudioChannels(uAudioChannels),
m_uAudioBitRate(uAudioBitRate),
m_uFrameWidth(uFrameWidth),
m_uFrameHeight(uFrameHeight),
m_dFrameRate(dFrameRate),
m_uDesiredFrameRate(uDesiredFrameRate),
m_ffmpeg(FFMPEGAccess::Get()),
m_sFileName(sFileName),
m_pFormat(m_ffmpeg.GetFormatByFileName(sFileName)),
m_pContext(m_ffmpeg.AllocateFormatContext(m_pFormat)),
m_pVideoStream(NULL),
m_pAudioStream(NULL),
m_bVideoDone(true),
m_bAudioDone(true),
m_pProgressSelection(new CTimeBasedProgressSelection(2)),
m_uVideoFramesNumber(0),
m_tsRecent(0)
{
	m_dFrameRate = uDesiredFrameRate;

	m_ffmpeg.SetOutputFileName(m_pContext, sFileName);	

}

CAVFormatWriter::~CAVFormatWriter()
{
	m_ffmpeg.FreeStreams(m_pContext);
	m_ffmpeg.FreeAllocatedMemory(m_pContext);
	delete m_pProgressSelection;
}

void CAVFormatWriter::Start()
{
	m_bAudioDone = false; m_bVideoDone = false;

	m_ffmpeg.SetContextParameters(m_pContext);
	m_ffmpeg.OpenOutputFile(m_pFormat, m_pContext, (LPCTSTR) m_sFileName);
	m_ffmpeg.WriteContextHeader(m_pContext);
}

void FlipPicture(	const void *pSource,
					void *pTarget,
					int iWidth,
					int iHeight,
					int iBytesPerPixel
				)
{
	const unsigned char * pSourcePtr = (const unsigned char*) pSource;
	unsigned char * pTargetPtr = (unsigned char*) pTarget;
	int iPixelsInRow = iWidth * iBytesPerPixel;

	for (int c = 0; c < iWidth; ++c) {
		for (int r = 0; r < iHeight; ++r) {
			int iPos1 = r * iWidth + c;
			int iPos2 = (iHeight - r - 1) * iWidth + c;
			for (int i = 0; i < iBytesPerPixel; ++i) {
				pTargetPtr[iPos2 * iBytesPerPixel + i] = 
					pSourcePtr[(iPos1 + 1) * iBytesPerPixel - i - 1];
			}
		}
	}
}

unsigned int CVideoCodecTrackWriter::Write(
	const void * buf, 
	unsigned int numBytes,
	Timestamp timestamp,
	Duration duration
	) 
{
	assert(numBytes == m_iOriginalSize);

	unsigned long uNumExpectedFrames = 
		m_ffmpeg.GetNumberOfExpectedFrames(m_pStream,
			timestamp + duration, duration);

	LOG_STREAM("Frame time = " << 
		(double) (__int64) timestamp / TimestampTicks );

	LOG_STREAM("Num frames = " << m_uNumFrames << 
		"; Expected frames = " << uNumExpectedFrames)

	if (m_uNumFrames >= uNumExpectedFrames)
		return numBytes;

	FlipPicture	(	buf, 
					m_pInitialTransformBuffer,
					m_pCodecContext->width, 
					m_pCodecContext->height,
					INITIAL_BYTES_PER_PIXEL
				);

	int iYUVSize(0);

	if (m_pCodecContext->pix_fmt == m_pfIn) {
		// place the buf into the frame
		avpicture_fill((AVPicture *)m_pFrame, 
			(uint8_t*) m_pInitialTransformBuffer, 
			m_pCodecContext->pix_fmt, 
			m_pCodecContext->width, 
			m_pCodecContext->height);
	}
	else {
		iYUVSize = avpicture_fill((AVPicture *) m_pOriginalFrame, 
			(uint8_t*) m_pInitialTransformBuffer, 
			m_pfIn, 
			m_pCodecContext->width, 
			m_pCodecContext->height);

		img_convert((AVPicture *) m_pFrame, 
			m_pCodecContext->pix_fmt, 
            (AVPicture *) m_pOriginalFrame, 
			m_pfIn,
            m_pCodecContext->width, 
			m_pCodecContext->height);
	}

	m_uYUVBytesWritten += iYUVSize;

	for (; m_uNumFrames < uNumExpectedFrames; ++m_uNumFrames)
	{
		/* encode the image */
		int iOutputSize = m_ffmpeg.EncodeVideo(
			m_pCodecContext, m_pOutBuffer, m_iOutBufferSize, m_pFrame);
    
		/* write the compressed frame in the media file */
		m_AVFormatWriter.WriteFrame(CODEC_TYPE_VIDEO, 
			m_pOutBuffer, iOutputSize, m_timestamp, duration);

		m_uXVidBytesWritten += iOutputSize;
		m_timestamp += duration;
	}

	return numBytes;
}

CVideoCodecTrackWriter::CVideoCodecTrackWriter(
	CAVFormatWriter& avFormatWriter,
	AVFormatContext *pFormatContext,
	AVStream* pAVStream,
	PixelFormat pfIn):
m_ffmpeg(FFMPEGAccess::Get()),
m_AVFormatWriter(avFormatWriter),
m_pStream(pAVStream),
m_pfIn(pfIn),
m_pInitialTransformBuffer(NULL),
m_uNumFrames(0),
m_timestamp(0),
m_uYUVBytesWritten(0),
m_uXVidBytesWritten(0)
{
	m_pCodecContext = &pAVStream->codec;
	m_pAVCodec = m_ffmpeg.FindEncoderByCodec(
					m_pCodecContext->codec_id);
	assert(m_pAVCodec);
	int iOpenRes = avcodec_open(m_pCodecContext, m_pAVCodec);
	assert(iOpenRes >= 0);

	m_pOutBuffer = NULL;
	if (!(pFormatContext->oformat->flags & AVFMT_RAWPICTURE)) {
		/* allocate output buffer */
		/* XXX: API change will be done */
		m_iOutBufferSize = 200000;
		m_pOutBuffer = (uint8_t*) malloc(m_iOutBufferSize);
	}

	m_pFrame = m_ffmpeg.AllocatePicture(
		m_pCodecContext->pix_fmt, 
		m_pCodecContext->width, 
		m_pCodecContext->height,
		m_iPictureSize);

	assert(m_pFrame);

	m_pInitialTransformBuffer = 
		malloc(	m_pCodecContext->width *
				m_pCodecContext->height *
				INITIAL_BYTES_PER_PIXEL	);

	if (m_pCodecContext->pix_fmt == m_pfIn) {
		m_pOriginalFrame = NULL;
		m_pConversionBuffer = NULL;
		m_iOriginalSize = m_iPictureSize;
	}
	else {
		m_pOriginalFrame = m_ffmpeg.AllocatePicture(
		m_pfIn, 
		m_pCodecContext->width, 
		m_pCodecContext->height,
		m_iOriginalSize);

		assert(m_pOriginalFrame);

		m_pConversionBuffer = (uint8_t*) malloc(m_iPictureSize);

		avpicture_fill((AVPicture *)m_pFrame, 
			(uint8_t*) m_pConversionBuffer, 
			m_pCodecContext->pix_fmt, 
			m_pCodecContext->width, 
			m_pCodecContext->height);
	}
}

CVideoCodecTrackWriter::~CVideoCodecTrackWriter()
{
	av_free(m_pOriginalFrame);
    av_free(m_pFrame);
	free(m_pOutBuffer);
	free(m_pConversionBuffer);
	free(m_pInitialTransformBuffer);
}

void CVideoCodecTrackWriter::DeclareWritingEnded() {
	m_AVFormatWriter.DeclareWritingEnded(CODEC_TYPE_VIDEO);
}

class CAVTrackWriter : public TimeKeepingWriter {
private:
	unsigned int Write(
		const void * buf, 
		unsigned int numBytes,
		Timestamp timestamp,
		Duration duration
		) 
	{
		m_AVFormatWriter.WriteFrame(m_CodecType, buf, numBytes,
			timestamp, duration);

		return numBytes;
	}
	
	CAVFormatWriter& m_AVFormatWriter;
	CodecType m_CodecType;

public:
	CAVTrackWriter(CAVFormatWriter& avFormatWriter,
		CodecType codecType):
	m_AVFormatWriter(avFormatWriter),
	m_CodecType(codecType)
	{}

	void DeclareWritingEnded() {
		m_AVFormatWriter.DeclareWritingEnded(m_CodecType);
	}

	virtual ~CAVTrackWriter() {}
};

CVideoCodecTrackWriter* CAVFormatWriter::CreateVideoCodecTrackWriter(CodecID codecID)
{
	assert(m_ffmpeg.FormatSupportsVideo(m_pFormat));
	m_pVideoStream = m_ffmpeg.AddVideoStream(m_pContext,
		codecID, m_uFrameWidth, m_uFrameHeight, m_dFrameRate);

	return new CVideoCodecTrackWriter(*this, m_pContext, m_pVideoStream, PIX_FMT_RGB24);
}

TimeKeepingWriter* CAVFormatWriter::CreateTrackWriter(CodecType codecType, CodecID codecID)
{
	switch (codecType) {
    case CODEC_TYPE_VIDEO:
		{
			assert(m_ffmpeg.FormatSupportsVideo(m_pFormat));
			m_pVideoStream = m_ffmpeg.AddVideoStream(m_pContext,
				codecID, m_uFrameWidth, m_uFrameHeight, m_dFrameRate);

			return new CVideoCodecTrackWriter(*this, m_pContext, m_pVideoStream, PIX_FMT_RGB24);

		} break;
    case CODEC_TYPE_AUDIO:
		{
			assert(m_ffmpeg.FormatSupportsAudio(m_pFormat));
			m_pAudioStream = m_ffmpeg.AddAudioStream(m_pContext,
				codecID, m_uAudioChannels, m_uAudioSampleRate, m_uAudioBitRate);

		} break;

	default:
		assert(false);
	}

	return new CAVTrackWriter(*this, codecType);
}

void CAVFormatWriter::WriteFrame
(CodecType codecType, const void *buf, unsigned int numBytes,
 Timestamp timestamp, Duration duration)
{
	switch (codecType) {
	case CODEC_TYPE_VIDEO:
		m_pProgressSelection->Wait(FFMPEGAccess::VIDEO_STREAM_INDEX, timestamp + duration);
		++m_uVideoFramesNumber;
		m_ffmpeg.WriteFrame(m_pContext, 
			FFMPEGAccess::VIDEO_STREAM_INDEX, buf, numBytes);
		//m_ffmpeg.WriteVideoFrame(m_pContext,
		//	buf, numBytes, timestamp, m_uVideoFramesNumber);
		break;
	case CODEC_TYPE_AUDIO:
		m_pProgressSelection->Wait(FFMPEGAccess::AUDIO_STREAM_INDEX, timestamp + duration);
		m_ffmpeg.WriteFrame(m_pContext, 
			FFMPEGAccess::AUDIO_STREAM_INDEX, buf, numBytes);
		//m_ffmpeg.WriteAudio(m_pContext, buf, numBytes);
		break;
	default:
		assert(false);
	};

	m_tsRecent = timestamp + duration;
	m_pProgressSelection->Release();
}


void CAVFormatWriter::DeclareWritingEnded(CodecType codecType)
{
	switch (codecType) {
	case CODEC_TYPE_VIDEO:
		m_pProgressSelection->DeclareInactive(FFMPEGAccess::VIDEO_STREAM_INDEX);
		m_bVideoDone = true;
		break;
	case CODEC_TYPE_AUDIO:
		m_pProgressSelection->DeclareInactive(FFMPEGAccess::AUDIO_STREAM_INDEX);
		m_bAudioDone = true;
		break;
	default:
		assert(false);
	};

	if (m_bAudioDone && m_bVideoDone)
		Stop();

	m_pProgressSelection->Release();
}

void CAVFormatWriter::Stop()
{
	m_ffmpeg.WriteContextTrailer(m_pContext);
	m_ffmpeg.CloseOutputFile(m_pFormat, m_pContext);
}
