/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: ActiveApplicationObjects.cpp,v 1.3 2004/04/12 15:36:49 gadibergman Exp $
 */

/************************************************************
				ActiveApplicationObjects.cpp
				============================

  A functional layer that stores pointers to active UI 
  objects - the video viewer and the audio recorder dialog.

************************************************************/

// PCH includes
#include "stdafx.h"

// STL includes
#include <string>

// ID includes
#include "MPEGRecorder_i.h"

// Project includes
#include "ActiveApplicationObjects.h"
#include "ApplicationPropertiesSheet.h"
#include "MPEG4AVRecorder.h"
#include "RecorderDlg.h"
#include "RecordingEventsUI.h"
#include "SoundCaptureDevices.h"
#include "VideoCaptureDevices.h"
#include "VideoViewer.h"
#include "VolumeMeterDialog.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CActiveApplicationObjects::CActiveApplicationObjects():
	m_SpecificApplicationData(),
	m_AudioFileNameGenerator(m_SpecificApplicationData.GetRecordingDirectory(),
	_T(".mp3")),
	m_MPEGFileNameGenerator(m_SpecificApplicationData.GetRecordingDirectory(),
	_T(".mp4")),
	m_AudioEncoderGenerator(m_SpecificApplicationData.GetAudioKBitRate()),
	m_pCOMAVRecorder(),
	m_bInTransition(false),
	m_pVideoViewer(NULL),
	m_pAudioRecorderDlg(NULL),
	m_pVolumeMeterDlg(NULL),
	m_bHasNextEvent(false)
{

}

CActiveApplicationObjects::~CActiveApplicationObjects()
{
}

bool CActiveApplicationObjects::EditProperties()
{
	try {
		CApplicationPropertiesSheet props(_T("Settings"));
		props.SetApplicationData(&m_SpecificApplicationData);
		return (props.DoModal() == IDOK);
	}
	catch (std::string * errMess) {
		AfxMessageBox(errMess->c_str(), MB_OK, 0);
		return false;
	}
}

bool CActiveApplicationObjects::UseProperties()
{
	m_AudioFileNameGenerator.SetRootDirectory(
		m_SpecificApplicationData.GetRecordingDirectory());
	m_MPEGFileNameGenerator.SetRootDirectory(
		m_SpecificApplicationData.GetRecordingDirectory());

	CString sMPEGExtension(_T("."));
	sMPEGExtension += m_SpecificApplicationData.GetAVFormatName();

	m_MPEGFileNameGenerator.SetExtension(sMPEGExtension);

	m_AudioEncoderGenerator.SetEncoderType(
		(CAudioEncoderGenerator::AudioEncoderType)
		m_SpecificApplicationData.GetAudioEncoderNumber());

	m_AudioEncoderGenerator.SetKBitRate(
		m_SpecificApplicationData.GetAudioKBitRate());

	m_AudioFileNameGenerator.SetExtension(
		m_AudioEncoderGenerator.GetDefaultEncodedFileExtension());

	const CString & devName = 
		m_SpecificApplicationData.GetRecordingDeviceName();

	bool success;

	CSoundCaptureDevices recordingDevices;
	success = recordingDevices.GetDeviceIDByName(
		(LPCTSTR) devName, m_gRecordingDeviceID);

	if (!success)
		return false;

	const CString & vidDevName = 
		m_SpecificApplicationData.GetVideoCaptureDevName();

	CVideoCaptureDevices videoDevices;
	m_pVideoFilter = videoDevices.GetVideoFilterByName(
		(LPCTSTR) vidDevName);

	if (!m_pVideoFilter)
		return false;

	return true;
}

bool CActiveApplicationObjects::LoadApplicationData()
{
	return m_SpecificApplicationData.Load();
}

#define MULTIMEDIA_CHECK  			\
	if (m_bInTransition ||			\
		m_pAudioRecorderDlg ||		\
		m_pVideoViewer ||			\
		m_pVolumeMeterDlg)			\
		return;

void CActiveApplicationObjects::RecordAudio()
{
	MULTIMEDIA_CHECK

	try {
		new CRecorderDlg(&m_gRecordingDeviceID, 
		m_AudioFileNameGenerator, 
		m_AudioEncoderGenerator,
		m_bInTransition,
		&m_pAudioRecorderDlg,
		NULL, AfxGetApp( )->m_pMainWnd);
	}
	catch (std::string* pstrError)
	{
		AfxMessageBox(pstrError->c_str());		
	}
}

void CActiveApplicationObjects::WatchVideo(bool bIsManual)
{
	MULTIMEDIA_CHECK

	CVideoViewer::RecorderControlMode controlMode = 
		(bIsManual ? CVideoViewer::MANUAL_CONTROL : CVideoViewer::EXTERNAL_CONTROL);

	new CVideoViewer(
		m_pVideoFilter, 
		m_gRecordingDeviceID, 
		m_AudioEncoderGenerator,
		m_MPEGFileNameGenerator,
		controlMode,
		m_SpecificApplicationData.GetTVVideoFormatNumber(),
		m_SpecificApplicationData.GetTVChannelRef(),
		m_SpecificApplicationData.GetVideoFrameRate(),
		std::string((LPCSTR)m_SpecificApplicationData.GetPreferredImageSize()),
		m_bInTransition,
		&m_pVideoViewer,
		AfxGetApp( )->m_pMainWnd);
}

void CActiveApplicationObjects::DisplayVolumeUnitMeter()
{
	MULTIMEDIA_CHECK

	new CVolumeMeterDialog( 
		&m_gRecordingDeviceID,
		m_bInTransition,
		&m_pVolumeMeterDlg,
		AfxGetApp( )->m_pMainWnd
		);

}

void CActiveApplicationObjects::ShowMyFiles()
{
	CString sDirectory, sFileName;
	m_MPEGFileNameGenerator.GenerateFileName(sDirectory, sFileName);
	ShellExecute(AfxGetApp( )->m_pMainWnd->m_hWnd, _T("open"), 
		sDirectory, NULL, NULL, SW_SHOW);
}

void CActiveApplicationObjects::RecordVideo(bool bIsAutomatic)
{
	WatchVideo(!bIsAutomatic);

	if (m_pVideoViewer)
	{
		if (bIsAutomatic)
			m_pVideoViewer->ExternalStartRecording();
		else
			m_pVideoViewer->ManualStartRecording();
	}
}

void CActiveApplicationObjects::ManageRecordingEvents()
{
	CRecordingEventsUI recordingEventsDlg;
	recordingEventsDlg.EditTimeEventsVector(m_vTimeEvents);

	UpdateNextRecordingEvent();
}


void CActiveApplicationObjects::OnMinutePassed()
{
	if (m_pVideoViewer)
		m_pVideoViewer->OnMinutePassed();

	if (m_bHasNextEvent)
	{
		CSimpleTimePoint tpCurrent;
		tpCurrent.SetLocalTime();
		if (tpCurrent == m_tpNextEvent)
		{
			if (m_bNextEventIsEnding)
			{
				if (m_pVideoViewer)
					m_pVideoViewer->ExternalStopRecording();
			}
			else
			{
				RecordVideo(true);
			}

			UpdateNextRecordingEvent();
		}
	}	
}

void CActiveApplicationObjects::OnTimeChanged()
{
	UpdateNextRecordingEvent();
}

void CActiveApplicationObjects::UpdateNextRecordingEvent()
{
	CSimpleTimePoint tpCurrent;
	tpCurrent.SetLocalTime();
	CRepeatingTimeEvent tEvent;

	m_bHasNextEvent = 
		m_vTimeEvents.GetEarliestStateChange(
		tpCurrent, m_tpNextEvent, m_bNextEventIsEnding, tEvent);
}

