/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: BlocksPipe.cpp,v 1.2 2004/01/25 17:45:33 gadibergman Exp $
 */

/************************************************************
				BlocksPipe.cpp
				==============

  A standard queue of data blocks that implements standard
  access methods - reading a block and writing a block.

  The reading is performed from the head of the list
  and the writing is performed to the tail of the list.

  This pipe is thread-safe, with a built-in mutual exclusion
  mechanism.

************************************************************/

// PCH includes
#include "StdAfx.h"

#include <winbase.h>

// Project includes
#include "BlocksPipe.h"

CBlocksPipe::CBlocksPipe():
m_bWritingEnded(false),
m_hDataEvent(CreateEvent(NULL, TRUE, FALSE, NULL)),
m_uTotalRead(0),
m_uTotalWritten(0),
m_uBytesAllocated(0),
m_uAllocationSizeLimit(0)
{}

CBlocksPipe::~CBlocksPipe() {
	CloseHandle(m_hDataEvent);
}

CBasicPipeBlock::CBasicPipeBlock(const void * pBuffer, unsigned long uSize):
m_pBuffer(new char[uSize]),
m_pPosition(m_pBuffer),
m_uSize(uSize)
{
	ASSERT(m_pBuffer);

	memcpy(m_pBuffer, pBuffer, uSize);	  
}

CBasicPipeBlock::~CBasicPipeBlock() {
  delete [] m_pBuffer;
}

unsigned int CBasicPipeBlock::Read(void * buf, unsigned int numBytes) {
	unsigned int uAvailable = 
		(m_uSize - (m_pPosition - m_pBuffer));
	unsigned int uTempSize = (uAvailable > numBytes) ? 
			numBytes : uAvailable;
	memcpy(buf, m_pPosition, uTempSize);
	m_pPosition += uTempSize;
	return uTempSize;
}

bool CBasicPipeBlock::IsDone() const {
	return ((m_pBuffer + m_uSize) == m_pPosition);
}

unsigned int CBasicPipeBlock::GetBlockSize() const {
	return m_uSize;
}

CBasicPipeBlock * CBlocksPipe::ReadBlockAdvance() {

	BOOL bResult;
	DWORD dwResult = WaitForSingleObject(m_Mutex._mutexHandle, INFINITE);
	ASSERT(dwResult == WAIT_OBJECT_0);
	CBasicPipeBlock * pBlock = NULL;

	do {

		if (m_blocks.size()) {
			BlocksList::iterator iter = m_blocks.begin();
			pBlock = *iter;
			if (pBlock->IsDone()) {
				m_uBytesAllocated -= pBlock->GetBlockSize();
				delete pBlock;
				m_blocks.erase(iter); // Delete the first block
			}
			else
				break; // We have a block - the first one
		}
		else {
			
			if (m_bWritingEnded) {
				pBlock = NULL; // No more blocks
				break;
			}
			
			bResult = ResetEvent(m_hDataEvent);
			ASSERT(bResult);

			bResult = ReleaseMutex(m_Mutex._mutexHandle);
			ASSERT(bResult);

			// Wait for data to arrive from the writer
			dwResult = WaitForSingleObject(
				m_hDataEvent, INFINITE);

			ASSERT(dwResult == WAIT_OBJECT_0);

			// Wait again for the buffer mutex
			dwResult = WaitForSingleObject(m_Mutex._mutexHandle, 
				INFINITE);
			ASSERT(dwResult == WAIT_OBJECT_0);
		}

	} while (1);

	bResult = ReleaseMutex(m_Mutex._mutexHandle);
	ASSERT(bResult);
	return pBlock;
}

unsigned int CBlocksPipe::ReadBlock(CBasicPipeBlock * pBlock, void * buf, unsigned int numBytes) {
	return pBlock->Read(buf, numBytes);
}

unsigned int CBlocksPipe::Read(void * buf, unsigned int numBytes) {
	
	unsigned int numRead = 0;
	char * pBuffer = (char *) buf;

	while (numBytes > 0) {
		CBasicPipeBlock * pBlock = ReadBlockAdvance();

		if (pBlock == NULL)
			break;

		unsigned int uTempSize = ReadBlock(pBlock, pBuffer, numBytes);
		
		pBuffer += uTempSize;
		numBytes -= uTempSize;
		numRead += uTempSize;
	}

	return numRead;
}

CBasicPipeBlock * CBlocksPipe::CreateBasicBlock(const void * pBuffer, unsigned long uSize) {
	return new CBasicPipeBlock(pBuffer, uSize);
}

unsigned int CBlocksPipe::Write(const void * buf, unsigned int numBytes) {
	ASSERT(!m_bWritingEnded);
	
	MutexProtection prot(m_Mutex);

	// Check allocation size limit
	if ( m_uAllocationSizeLimit &&
		 ( m_uBytesAllocated + numBytes > m_uAllocationSizeLimit ) )
		return 0;

	m_blocks.push_back(CreateBasicBlock(buf, numBytes));
	m_uBytesAllocated += numBytes;
	SetEvent(m_hDataEvent);

	return numBytes;
}

void CBlocksPipe::DeclareWritingEnded()
{
	MutexProtection prot(m_Mutex);
	m_bWritingEnded = true;
	SetEvent(m_hDataEvent);
}

unsigned long CBlocksPipe::HowMuchWritten() const {
	MutexProtection prot(m_Mutex);
	return m_uTotalWritten;
}

unsigned long CBlocksPipe::HowMuchRead() const {
	MutexProtection prot(m_Mutex);
	return m_uTotalRead;
}

void CBlocksPipe::SetAllocationSizeLimit(unsigned long uAllocationSizeLimit)
{
	MutexProtection prot(m_Mutex);
	m_uAllocationSizeLimit = uAllocationSizeLimit;
}

