/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: BlocksPipe.h,v 1.2 2004/01/25 17:45:33 gadibergman Exp $
 */

/************************************************************
				BlocksPipe.h
				============

  A standard queue of data blocks that implements standard
  access methods - reading a block and writing a block.

  The reading is performed from the head of the list
  and the writing is performed to the tail of the list.

  This pipe is thread-safe, with a built-in mutual exclusion
  mechanism.

************************************************************/

#ifndef __BLOCKS_PIPE_H
#define __BLOCKS_PIPE_H

#include "Common/ReaderWriter.h"
#include "Common/MutexProtection.h"

#include <list>

class CBasicPipeBlock {
private:
	unsigned int m_uSize;
	char * m_pBuffer, * m_pPosition;

public:
	CBasicPipeBlock(const void * pBuffer, unsigned long uSize);
	virtual ~CBasicPipeBlock();
	unsigned int Read(void * buf, unsigned int numBytes);
	bool IsDone() const;
	unsigned int GetBlockSize() const;

};

class CBlocksPipe : public GeneralReader, public GeneralWriter {

private:

	bool m_bWritingEnded;
	
	Mutex m_Mutex;
	HANDLE m_hDataEvent;
	unsigned long m_uTotalRead;
	unsigned long m_uTotalWritten;
	unsigned long m_uBytesAllocated;
	unsigned long m_uAllocationSizeLimit;

	typedef std::list<CBasicPipeBlock*> BlocksList;	
	BlocksList m_blocks;

	// returns the next block.
	// waits for next block if there is something to wait for.
	// returns NULL if there are no more blocks coming.
	CBasicPipeBlock * ReadBlockAdvance();

protected:

	unsigned int Read(void * buf, unsigned int numBytes);
	unsigned int Write(const void * buf, unsigned int numBytes);
	void DeclareWritingEnded();

	virtual CBasicPipeBlock * CreateBasicBlock(const void * pBuffer, unsigned long uSize);
	virtual unsigned int ReadBlock(CBasicPipeBlock * pBlock, void * buf, unsigned int numBytes);


public:

	CBlocksPipe();

	virtual ~CBlocksPipe();

	unsigned long HowMuchWritten() const;
	unsigned long HowMuchRead() const;

	void SetAllocationSizeLimit(unsigned long uAllocationSizeLimit);

};


#endif