/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: DXSoundReceiver.cpp,v 1.2 2004/04/12 15:36:49 gadibergman Exp $
 */

/************************************************************
			DXSoundReceiver.cpp
			===================

  An object that attaches to a DirectX sound device,
  receives buffers of raw samples, and redirects them
  to a standard buffer writing object.

************************************************************/

// PCH includes
#include "StdAfx.h"

// WIN32 includes
#include <process.h>

// STL includes
#include <string>

// IO includes
#include <iostream.h>

// Common includes
#include "Common/ReaderWriter.h"
#include "Common/MutexProtection.h"

// Project includes
#include "DXSoundReceiver.h"


using namespace std;

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------
#define SAFE_DELETE(p)  { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_RELEASE(p) { if(p) { (p)->Release(); (p)=NULL; } }

#ifndef MAX
#define MAX(a,b)        ( (a) > (b) ? (a) : (b) )
#endif

CDXSoundReceiver::CDXSoundReceiver( 
	const GUID * pDeviceID,
	unsigned short usChannels, 
	unsigned short usBytesPerSample,
	unsigned short usSampleRate):
// Base class
CSoundReceiverBase(),

// Local variables
m_usChannels(usChannels),
m_usBytesPerSample(usBytesPerSample),
m_usSampleRate(usSampleRate),
m_bTerminationRequested(false),
m_hTermination(CreateEvent(NULL, TRUE, TRUE, NULL)),
m_hNotificationEvent(CreateEvent( NULL, FALSE, FALSE, NULL )),

// DirectX specific
m_gDeviceID(*pDeviceID),
m_pDSCapture(NULL),
m_pDSBCapture(NULL),
m_pDSNotify(NULL)
{
	Init();	
}

// Destructor
CDXSoundReceiver::~CDXSoundReceiver() {

	// Release the thread from waiting to a buffer
	BOOL bResult = SetEvent(m_hNotificationEvent);
	ASSERT(bResult);

	// Wait for the thread to complete
	DWORD dwResult = WaitForSingleObject(m_hTermination, INFINITE);
	ASSERT(dwResult == WAIT_OBJECT_0);

	// Release DirectSound interfaces
	SAFE_RELEASE( m_pDSNotify );
	SAFE_RELEASE( m_pDSBCapture );
	SAFE_RELEASE( m_pDSCapture ); 

	CloseHandle(m_hTermination);
}

static void CDXSoundReceiver_RecordingThreadFunc(LPVOID pParam) {
	
	CDXSoundReceiver * pSoundReceiver = (CDXSoundReceiver*) pParam;
	ThreadCompletionAnnouncement announce(pSoundReceiver->m_hTermination);

	try {
		pSoundReceiver->Record();
	}
	catch (string * errMess) {
		cerr << errMess << endl;
		AfxMessageBox(_T(errMess->c_str()));
		delete errMess;
		return /*-1*/;
	}
	catch (...) {
		AfxMessageBox(_T("Unhandled exception in Sound Receiver Thread"));
	}

	/*return 0;*/
}


// Initializations
void CDXSoundReceiver::Init() {

	InitDirectSound();

	CreateCaptureBuffer();

}

void CDXSoundReceiver::Start()
{
	m_tsLast = GetTimestamp();

	unsigned long ulSampleSize = m_usChannels * m_usBytesPerSample;

	CreateMediaDivision(ulSampleSize, m_usSampleRate, false);

	_beginthread(CDXSoundReceiver_RecordingThreadFunc, 0, this);

	HRESULT hr;
	if( FAILED( hr = m_pDSBCapture->Start( DSCBSTART_LOOPING ) ) )
		throw new string("Failed to start recording");

}

void CDXSoundReceiver::InitDirectSound()
{
    HRESULT hr;

    ZeroMemory( &m_aPosNotify, sizeof(DSBPOSITIONNOTIFY) * 
                               (NUM_REC_NOTIFICATIONS + 1) );
    m_dwCaptureBufferSize = 0;
    m_dwNotifySize        = 0;

    // Create IDirectSoundCapture using the preferred capture device
    if( FAILED( hr = DirectSoundCaptureCreate( &m_gDeviceID, &m_pDSCapture, NULL ) ) )
		throw new string("Failed to create direct sound capture");
}

void CDXSoundReceiver::CreateCaptureBuffer()
{
    HRESULT hr;
    DSCBUFFERDESC dscbd;

    SAFE_RELEASE( m_pDSNotify );
    SAFE_RELEASE( m_pDSBCapture );

	// Prepare wave format for sound card tuning
	WAVEFORMATEX wf = {
		WAVE_FORMAT_PCM,
		m_usChannels,
		m_usSampleRate,
		m_usChannels * m_usBytesPerSample * m_usSampleRate,
		m_usChannels * m_usBytesPerSample,
		m_usBytesPerSample << 3,
		0
	};

    // Set the notification size
    m_dwNotifySize = MAX( 1024, wf.nAvgBytesPerSec / 8 );
    m_dwNotifySize -= m_dwNotifySize % wf.nBlockAlign;   

    // Set the buffer sizes 
    m_dwCaptureBufferSize = m_dwNotifySize * NUM_REC_NOTIFICATIONS;

    SAFE_RELEASE( m_pDSNotify );
    SAFE_RELEASE( m_pDSBCapture );

    // Create the capture buffer
    ZeroMemory( &dscbd, sizeof(dscbd) );
    dscbd.dwSize        = sizeof(dscbd);
    dscbd.dwBufferBytes = m_dwCaptureBufferSize;
    dscbd.lpwfxFormat   = &wf; // Set the format during creatation

    if( FAILED( hr = m_pDSCapture->CreateCaptureBuffer( &dscbd, 
                                                        &m_pDSBCapture, 
                                                        NULL ) ) )
		throw new string("Failed to create capture buffer");

    m_dwNextCaptureOffset = 0;

	InitNotifications();
}

void CDXSoundReceiver::InitNotifications()
{
    HRESULT hr; 

    if( !m_pDSBCapture )
		throw new string("No capture buffer");

    // Create a notification event, for when the sound stops playing
    if( FAILED( hr = m_pDSBCapture->QueryInterface( IID_IDirectSoundNotify, 
                                                    (VOID**)&m_pDSNotify ) ) )
		throw new string("The capture buffer does not implement IDirectSoundNotify interface");

    // Setup the notification positions
    for( int i = 0; i < NUM_REC_NOTIFICATIONS; i++ )
    {
        m_aPosNotify[i].dwOffset = (m_dwNotifySize * i) + m_dwNotifySize - 1;
        m_aPosNotify[i].hEventNotify = m_hNotificationEvent;             
    }
    
    // Tell DirectSound when to notify us. the notification will come in the from 
    // of signaled events that are handled in WinMain()
    if( FAILED( hr = m_pDSNotify->SetNotificationPositions( NUM_REC_NOTIFICATIONS, 
                                                            m_aPosNotify ) ) )
		throw new string("Failed to set notification positions");
}

void CDXSoundReceiver::Stop()
{
	HRESULT hr;
    // Stop the capture and read any data that 
    // was not caught by a notification
    if( NULL == m_pDSBCapture )
		throw new string("No capture buffer found for termination");

    // Stop the buffer, and read any data that was not 
    // caught by a notification
    if( FAILED( hr = m_pDSBCapture->Stop() ) )
		throw new string("Failed to stop capture buffer");

	RecordCapturedData();

	DeclareWritingEnded();

	m_bTerminationRequested = true;
}

void CDXSoundReceiver::RecordCapturedData()
{
    HRESULT hr;
    VOID*   pbCaptureData    = NULL;
    DWORD   dwCaptureLength;
    VOID*   pbCaptureData2   = NULL;
    DWORD   dwCaptureLength2;
    VOID*   pbPlayData       = NULL;
    UINT    dwDataWrote;
    DWORD   dwReadPos;
    DWORD   dwCapturePos;
    LONG lLockSize;


	Timestamp tsCurrent = m_tsLast;

	m_tsLast = GetTimestamp();

    ASSERT(m_pDSBCapture);
	
    if( FAILED( hr = m_pDSBCapture->GetCurrentPosition( &dwCapturePos, &dwReadPos ) ) )
		throw new string("Failed to get current position");

    lLockSize = dwReadPos - m_dwNextCaptureOffset;
    if( lLockSize < 0 )
        lLockSize += m_dwCaptureBufferSize;

    // Block align lock size so that we are always write on a boundary
    lLockSize -= (lLockSize % m_dwNotifySize);

    if( lLockSize == 0 )
        return /*S_FALSE*/;

    // Lock the capture buffer down
	hr = m_pDSBCapture->Lock( m_dwNextCaptureOffset, lLockSize, 
                              &pbCaptureData, &dwCaptureLength, 
                              &pbCaptureData2, &dwCaptureLength2, 
							  0L );

    if( FAILED( hr ) )
		return;
		//throw new string("Failed to lock capture buffer");

    // Write the data into a writer
	dwDataWrote = WriteBuffer(pbCaptureData, dwCaptureLength, tsCurrent);

    // Move the capture offset along
    m_dwNextCaptureOffset += dwCaptureLength; 
    m_dwNextCaptureOffset %= m_dwCaptureBufferSize; // Circular buffer

    if( pbCaptureData2 != NULL )
    {
		unsigned long dwBytesPerSec = 
			m_usSampleRate * m_usChannels * m_usBytesPerSample;
		tsCurrent += TimestampTicks * dwCaptureLength2 / dwBytesPerSec;

        // Write the data into a writer
		WriteBuffer(pbCaptureData2, dwCaptureLength2, tsCurrent);

        // Move the capture offset along
        m_dwNextCaptureOffset += dwCaptureLength2; 
        m_dwNextCaptureOffset %= m_dwCaptureBufferSize; // Circular buffer
    }

    // Unlock the capture buffer
    m_pDSBCapture->Unlock( pbCaptureData,  dwCaptureLength, 
                           pbCaptureData2, dwCaptureLength2 );

}

void CDXSoundReceiver::Record()
{
	DWORD dwResult;

	do {

		dwResult = WaitForSingleObject(m_hNotificationEvent, INFINITE);
		ASSERT(dwResult == WAIT_OBJECT_0);

		if (m_bTerminationRequested)
			break;
	
		RecordCapturedData();


	} while(1);

}

