/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: DXVideoReceiver.cpp,v 1.4 2004/04/12 15:36:49 gadibergman Exp $
 */

/************************************************************
			DXVideoReceiver.cpp
			===================

  An object that attaches to a DirectX video capture device,
  receives frames, and redirects them to a standard buffer 
  writing object.

************************************************************/

// PCH includes
#include "stdafx.h"

// Project includes
#include "DXVideoReceiver.h"
#include "MPEGRecorder.h"
#include "ProgramLog.h"
#include "SpecificApplicationData.h"

// Personal Direct Show includes
#include <mtype.h>
#include "Common/DirectShow/dshowutil.h"
#include <amvideo.h>
#include <control.h>
#include <Dshow.h>

// STL includes
#include <string>

using namespace std;


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

class CSampleGrabberCB : public ISampleGrabberCB 
{
public:
	virtual  ~CSampleGrabberCB();
    long m_lWidth;
    long m_lHeight;

	void SetVideoReceiver(CDXVideoReceiver * pReceiver);

    CSampleGrabberCB();

	// fake out any COM ref counting
	//
	STDMETHODIMP_(ULONG) AddRef() { return 2; }
	STDMETHODIMP_(ULONG) Release() { return 1; }
	STDMETHODIMP QueryInterface(REFIID riid, void ** ppv);
	STDMETHODIMP SampleCB( double SampleTime, IMediaSample * pSample );
	STDMETHODIMP BufferCB( double dblSampleTime, BYTE * pBuffer, long lBufferSize );

private:
	CDXVideoReceiver * m_pReceiver;
};

// fake out any COM QI'ing
//
STDMETHODIMP CSampleGrabberCB::QueryInterface(REFIID riid, void ** ppv)
{
    if( riid == IID_ISampleGrabberCB || riid == IID_IUnknown ) 
    {
        *ppv = (void *) static_cast<ISampleGrabberCB*> ( this );
        return NOERROR;
    }    
    return E_NOINTERFACE;
}

// we don't implement this interface for this example
//
STDMETHODIMP CSampleGrabberCB::SampleCB( double SampleTime, IMediaSample * pSample )
{
    return 0;
}

STDMETHODIMP CSampleGrabberCB::BufferCB( double dblSampleTime, BYTE * pBuffer, long lBufferSize )
{
	m_pReceiver->WriteBuffer(pBuffer, lBufferSize, dblSampleTime * TimestampTicks);
	return 0;
}

CSampleGrabberCB::CSampleGrabberCB():
m_pReceiver(NULL)
{
}

CSampleGrabberCB::~CSampleGrabberCB()
{
}

void CSampleGrabberCB::SetVideoReceiver(CDXVideoReceiver * pReceiver) {
	m_pReceiver = pReceiver;
}

class CActiveCaptureObjects {
public:
	CActiveCaptureObjects(CSampleGrabberCB * pGrabber,
		IBaseFilter * pVideoIn, HWND hPrevWindow,
		const std::string & strPrefImageSize);

	virtual ~CActiveCaptureObjects();

    // the capture still graph
    CComPtr< IGraphBuilder > m_pGraph;

	// the graph builder
	CComPtr< ICaptureGraphBuilder2 > m_pBuilder;

    // the sample grabber for grabbing stills
    CComPtr< ISampleGrabber > m_pGrabber;

	// Callback samples coming in
	CSampleGrabberCB * m_pGrabberCB;

	void Start();
	void Stop();
private:
	long m_lWidth, m_lHeight;
	double m_dFramesPerSecond;

public:
	long GetWidth() const { return m_lWidth; }
	long GetHeight() const { return m_lHeight; }
	double GetFrameRate() const { return m_dFramesPerSecond; }
	void SetPreviewWindow(HWND hPrevWindow);
	void GetGraphBuilderRef(IGraphBuilder** pGraph, ICaptureGraphBuilder2** pBuilder);
};

void CActiveCaptureObjects::GetGraphBuilderRef(IGraphBuilder** pGraph, ICaptureGraphBuilder2** pBuilder)
{
	m_pGraph.CopyTo(pGraph);
	ATLASSERT(pGraph);
	m_pBuilder.CopyTo(pBuilder);
	ATLASSERT(pBuilder);
}

#define TCE(f,errmess) \
{\
	HRESULT hr = (f);\
	if (FAILED(hr))\
		throw new string(errmess);\
}

void EnumPinMediaTypes(IPin* pPin)
{
	IEnumMediaTypes *pEnum;
	pPin->EnumMediaTypes(&pEnum);
	AM_MEDIA_TYPE *pMediaType;

	while(pEnum->Next(1, &pMediaType, 0) == S_OK)
	{
		VIDEOINFOHEADER * vih = (VIDEOINFOHEADER*) pMediaType->pbFormat;
		long
			lWidth  = vih->bmiHeader.biWidth,
			lHeight = vih->bmiHeader.biHeight;

		FreeMediaType( *pMediaType );
	}

	pEnum->Release();
}

void ApplyPreferredImageSize(IPin* pPin, LPCTSTR ptszPrefImageSize)
{
	if ( 0 == _tcsicmp( ptszPrefImageSize, DEFAULT_PREF_IMAGE_SIZE ) )
		return; // Default means do nothing here

	CComQIPtr< IAMStreamConfig > pStreamConfig( pPin );
	
	if ( !pStreamConfig )
		return;

	int iNumCaps(0), iCapsSize(0);
	pStreamConfig->GetNumberOfCapabilities( &iNumCaps, &iCapsSize );

	AM_MEDIA_TYPE *pmtBest = NULL;
	DWORD dwBestImageSize(0);

	for (int iIndex = 0; iIndex < iNumCaps; ++iIndex)
	{
		VIDEO_STREAM_CONFIG_CAPS vcVideoCaps;
		AM_MEDIA_TYPE *pmt;
		pStreamConfig->GetStreamCaps( iIndex, &pmt, (BYTE*) &vcVideoCaps );

		VIDEOINFOHEADER * pVIH = (VIDEOINFOHEADER*) pmt->pbFormat;
		BITMAPINFOHEADER * pBIH = &pVIH->bmiHeader;

		DWORD
			dwImageSize = pBIH->biWidth * pBIH->biHeight;

		if ( pBIH->biBitCount == 24 && pBIH->biCompression == 0 &&
			( !pmtBest || dwBestImageSize < dwImageSize ) )
		{
			if ( pmtBest )
				DeleteMediaType(pmtBest);

			pmtBest = pmt;
			dwBestImageSize = dwImageSize;
		}
		else
		{
			DeleteMediaType(pmt);
		}
	}

	if ( pmtBest )
	{
		if ( 0 == _tcsicmp( ptszPrefImageSize, BEST_PREF_IMAGE_SIZE ) )
		{
			// Do nothing here
		}
		else
		{
			int iWidth, iHeight;
			int iRes = _stscanf( ptszPrefImageSize, "%dx%d", &iWidth, &iHeight);

			if ( iRes == 2 )
			{
				VIDEOINFOHEADER * pVIH = (VIDEOINFOHEADER*) pmtBest->pbFormat;
				BITMAPINFOHEADER * pBIH = &pVIH->bmiHeader;

				pBIH->biBitCount /= pmtBest->lSampleSize;
				pVIH->dwBitRate /= pmtBest->lSampleSize;

				pBIH->biWidth = iWidth;
				pBIH->biHeight = iHeight;
				pmtBest->lSampleSize = pBIH->biWidth * pBIH->biHeight;

				pVIH->dwBitRate *= pmtBest->lSampleSize;
				pBIH->biBitCount *= pmtBest->lSampleSize;
			}
		}
		
		HRESULT hRes = pStreamConfig->SetFormat( pmtBest );

		DeleteMediaType(pmtBest);
	}
}

CActiveCaptureObjects::
CActiveCaptureObjects(CSampleGrabberCB * pGrabber, 
					  IBaseFilter * pVideoIn, HWND hPrevWindow,
					  const std::string & strPrefImageSize):
m_pGrabberCB(pGrabber)
{
    // create a filter graph
    TCE(m_pGraph.CoCreateInstance( CLSID_FilterGraph ),
		"Failed to create filter graph");

	// create graph builder
	TCE(m_pBuilder.CoCreateInstance( CLSID_CaptureGraphBuilder2 ),
		"Failed to create capture graph builder");

	// Set the builder's graph to be our filter graph
	m_pBuilder->SetFiltergraph( m_pGraph );

    // get capture device filter
    CComPtr< IBaseFilter > pCap = pVideoIn;
    if( !pCap )
		throw new string("No video capture device was detected on your system");

    // add the capture filter to the graph
    TCE(m_pGraph->AddFilter( pCap, L"Cap" ),
		"Failed to put capture device in graph");

    // create a sample grabber
    TCE(m_pGrabber.CoCreateInstance( CLSID_SampleGrabber ),
		"Failed to create SampleGrabber (is qedit.dll registered?)");

    CComQIPtr< IBaseFilter, &IID_IBaseFilter > pGrabBase( m_pGrabber );

    // force it to connect to video, 24 bit
    CMediaType VideoType;
    VideoType.SetType( &MEDIATYPE_Video );
    VideoType.SetSubtype( &MEDIASUBTYPE_RGB24 );

    TCE(m_pGrabber->SetMediaType( &VideoType ),
		"Failed to set media type");

    // add the grabber to the graph
    TCE(m_pGraph->AddFilter( pGrabBase, L"Grabber" ),
		"Failed to put sample grabber in graph");

    // find the two pins and connect them
    IPin * pCapOut = GetOutPin( pCap, 0 );
    IPin * pGrabIn = GetInPin( pGrabBase, 0 );

	ApplyPreferredImageSize( pCapOut, strPrefImageSize.c_str() );
	
    TCE(m_pGraph->Connect( pCapOut, pGrabIn ),
		"Failed to connect capture pin #0 to grabber.\r\n"
        "Is the capture device being used by another application?");

    // render the sample grabber output pin, so we get a preview window
    IPin * pGrabOut = GetOutPin( pGrabBase, 0 );
    TCE(m_pGraph->Render( pGrabOut ),
		"Failed to render sample grabber output pin");

    // ask for the connection media type so we know how big
    // it is, so we can write out bitmaps
    //
    AM_MEDIA_TYPE mt;
    m_pGrabber->GetConnectedMediaType( &mt );
    VIDEOINFOHEADER * pVih = (VIDEOINFOHEADER*) mt.pbFormat;
    m_lWidth  = pVih->bmiHeader.biWidth;
    m_lHeight = pVih->bmiHeader.biHeight;

	m_dFramesPerSecond = 1.0e7 / (pVih->AvgTimePerFrame); 
    FreeMediaType( mt );

    // do buffer the samples as they pass through
    m_pGrabber->SetBufferSamples( FALSE/*TRUE*/ );

    // only grab one at a time, stop stream after
    // grabbing one sample
    m_pGrabber->SetOneShot( FALSE );

    // set the callback, so we can grab the one sample
    m_pGrabber->SetCallback( m_pGrabberCB, 1 );

	SetPreviewWindow(hPrevWindow);
}

void CActiveCaptureObjects::SetPreviewWindow(HWND hPrevWindow)
{
    // find the video window and stuff it in our window
    CComQIPtr< IVideoWindow, &IID_IVideoWindow > pWindow = m_pGraph;
    if( !pWindow )
		throw new string("Failed to get video window interface");

    // set up the preview window to be in our dialog
    // instead of floating popup
    RECT rc;
    ::GetWindowRect( hPrevWindow, &rc );
    pWindow->put_Owner( (long) hPrevWindow );
    pWindow->put_Left( 0 );
    pWindow->put_Top( 0 );
    pWindow->put_Width( rc.right - rc.left );
    pWindow->put_Height( rc.bottom - rc.top );
    pWindow->put_Visible( OATRUE );
    pWindow->put_WindowStyle( WS_CHILD | WS_CLIPSIBLINGS );
}

void CActiveCaptureObjects::Start() {
    // run the graph
    CComQIPtr< IMediaControl, &IID_IMediaControl > pControl = m_pGraph;
    TCE(pControl->Run( ), "Failed to run the filter graph");
}

void CActiveCaptureObjects::Stop() {
    // have to wait for the graphs to stop first
    CComQIPtr< IMediaControl, &IID_IMediaControl > pControl = m_pGraph;
    if( pControl ) 
        pControl->Stop( );
}

CActiveCaptureObjects::~CActiveCaptureObjects() {
	// Clear graphs
    if( m_pGraph )
    {
        // make the window go away before we release graph
        // or we'll leak memory/resources
        // 
        CComQIPtr< IVideoWindow, &IID_IVideoWindow > pWindow = m_pGraph;
        if( pWindow )
        {
            pWindow->put_Visible( OAFALSE );
            pWindow->put_Owner( NULL );
        }

        m_pGrabber.Release();
		m_pBuilder.Release();
        m_pGraph.Release();
    }	
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CDXVideoReceiver::
CDXVideoReceiver(IBaseFilter * pVideoIn,
				 HWND hPrevWindow,
				 const std::string & strPrefImageSize):
CVideoReceiverBase(),
m_pSampleGrabberCB(NULL),
m_pCaptureObjects(NULL)
{
	try
	{
		m_pSampleGrabberCB = new CSampleGrabberCB();
		m_pSampleGrabberCB->SetVideoReceiver(this);
		m_pCaptureObjects = new CActiveCaptureObjects(m_pSampleGrabberCB,
			pVideoIn, hPrevWindow, strPrefImageSize);
	}
	catch (std::string* pstrError)
	{
		delete m_pCaptureObjects;
		delete m_pSampleGrabberCB;
		throw pstrError;
	}
}

CDXVideoReceiver::~CDXVideoReceiver()
{
	delete m_pCaptureObjects;
	delete m_pSampleGrabberCB;
}

void CDXVideoReceiver::GetGraphBuilderRef(IGraphBuilder** pGraph, ICaptureGraphBuilder2** pBuilder)
{
	m_pCaptureObjects->GetGraphBuilderRef(pGraph, pBuilder);
}

void CDXVideoReceiver::Start() {
	unsigned long ulAtomSize = GetWidth() * GetHeight() * 3;
	CreateMediaDivision(ulAtomSize, GetFrameRate(), true);
	m_pCaptureObjects->Start();
}

void CDXVideoReceiver::Stop() {	
	m_pCaptureObjects->Stop();
	DeclareWritingEnded();
}

long CDXVideoReceiver::GetWidth() const {
	return m_pCaptureObjects->GetWidth();
}

long CDXVideoReceiver::GetHeight() const {
	return m_pCaptureObjects->GetHeight();
}

double CDXVideoReceiver::GetFrameRate() const
{
	return m_pCaptureObjects->GetFrameRate();
}

void CDXVideoReceiver::SetPreviewWindow(HWND hPrevWindow)
{
	m_pCaptureObjects->SetPreviewWindow(hPrevWindow);
}
