/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: FAACEncoder.cpp,v 1.2 2004/03/27 13:47:16 gadibergman Exp $
 */

/************************************************************
			FAACEncoder.cpp
			===============

  An audio encoder that receives standard raw input buffers
  and writes standard encoded output buffers.
  Uses code from the Freeware Advanced Audio Coder.

************************************************************/

// PCH includes
#include "stdafx.h"

// Project include
#include "MPEGRecorder.h"
#include "FAACEncoder.h"
#include "TimeKeepingReaderWriter.h"
#include "Common/ObjPtr.h"
#include "Common/MutexProtection.h"
#include "ProgramLog.h"

#define HAVE_INT32_T

#include "libfaac/frame.h"
#include "libfaac/util.h"
#include "mpeg4/mp4av/mp4av_aac.h"
#include "mpeg4/mp4v2/mp4.h"

#include <process.h>
#include <string>

using namespace std;

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CFAACEncoder::
CFAACEncoder(	TimeKeepingReader * pReader, 
				TimeKeepingWriter * pWriter,
				unsigned long uSampleRate, 
				unsigned int uChannels,
				unsigned short usKBitRate,
				bool bIsMP4
			):
m_pReader(pReader),
m_pWriter(pWriter),
m_uSampleRate(uSampleRate),
m_uChannels(uChannels),
m_usKBitRate(usKBitRate),
m_bIsMP4(bIsMP4),
m_hThreadCompletion(CreateEvent(NULL, FALSE, TRUE, NULL))
{
	Init();
}

CFAACEncoder::~CFAACEncoder()
{
	DWORD dwResult = WaitForSingleObject(m_hThreadCompletion, INFINITE);
	ASSERT(dwResult == WAIT_OBJECT_0);

	CloseHandle(m_hThreadCompletion);
}

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

void CFAACEncoder::Encode()
{
	unsigned long totalDesiredBitRate = 1000 * m_usKBitRate;

	unsigned long samplesInput, maxBytesOutput;

	faacEncHandle hEncoder;

	/* Construction. Set main parameters */
	hEncoder = faacEncOpen(m_uSampleRate, m_uChannels, &samplesInput, &maxBytesOutput);

	/* Set other parameters */
	faacEncConfiguration* myFormat = faacEncGetCurrentConfiguration(hEncoder);
    myFormat->aacObjectType = LOW;
    myFormat->mpegVersion = MPEG4;
    myFormat->useTns = 1;
    myFormat->allowMidside = 1;
    myFormat->bandWidth = 16000;
    myFormat->quantqual = 100;
    myFormat->outputFormat = m_bIsMP4 ? 0 : 1;
	myFormat->bitRate = totalDesiredBitRate / m_uChannels;
    myFormat->inputFormat = FAAC_INPUT_FLOAT;
	int iRes = faacEncSetConfiguration(hEncoder, myFormat);
    ATLASSERT(iRes);

	ObjPtr<float> fpcmRaw( samplesInput );
	ObjPtr<short> spcmRaw( samplesInput );
	ObjPtr<unsigned char> bitbuf( maxBytesOutput );

		/* encoding loop */
		while(true)
		{
			Timestamp timestamp;
			Duration duration;

			int bytesInput = m_pReader->ReadBytes(
				spcmRaw, 
				sizeof(short)*samplesInput,
				timestamp,
				duration
				);

			int iSamplesRead = bytesInput/sizeof(short);

			LOG_STREAM( "Samples read = " << iSamplesRead );

			// Convert the integer samples to floating points
			for (int iSampleIndex = 0; iSampleIndex < samplesInput; ++iSampleIndex)
			{
				fpcmRaw[iSampleIndex] = spcmRaw[iSampleIndex];				
			}

			/* call the actual encoding routine */
			int bytesWritten = faacEncEncode(hEncoder,
				(int32_t*) (float*) fpcmRaw,
				iSamplesRead,
				bitbuf,
				maxBytesOutput);

			LOG_STREAM( "Bytes written = " << bytesWritten );

			if (bytesWritten < 0)
			{
				faacEncClose(hEncoder);
				throw new string("faacEncEncode() failed");
			}

			/* write bitstream to aac file */
			if ( bytesWritten )
			{
				m_pWriter->WriteBytes(bitbuf, bytesWritten,
					timestamp, duration);
			}
			else
			{
				if (!bytesInput)
					break;
			}
		}

	faacEncClose(hEncoder);

	m_pWriter->DeclareWritingEnded();
}

static void CFAACEncoder_EncodingThread(LPVOID pParam) {
	CFAACEncoder * pEncoder = (CFAACEncoder *) pParam;
	ThreadCompletionAnnouncement announce(pEncoder->m_hThreadCompletion);

	try {
		pEncoder->Encode();
	}
	catch (string * errMess) {
		AfxMessageBox(errMess->c_str());
		delete errMess;
	}
}

void CFAACEncoder::Init()
{
	_beginthread(CFAACEncoder_EncodingThread, 0, this);
}

MediaType CFAACEncoder::get_audio_mp4_fileinfo (
				  bool *mpeg4,
				  bool *isma_compliant,
				  uint8_t *audioProfile,
				  uint8_t **audioConfig,
				  uint32_t *audioConfigLen,
				  uint8_t *mp4AudioType)
{
  *mpeg4 = true;
  *isma_compliant = true;
  *audioProfile = 0x0f;
  if (mp4AudioType) *mp4AudioType = MP4_MPEG4_AUDIO_TYPE;

  MP4AV_AacGetConfiguration(audioConfig,
			    audioConfigLen,
			    MP4AV_AAC_LC_PROFILE,
			    m_uSampleRate,
				m_uChannels);
  return AACAUDIOFRAME;
}
