/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: LameEncoder.cpp,v 1.1.1.1 2003/12/24 19:41:39 gadibergman Exp $
 */

/************************************************************
			LameEncoder.cpp
			===============

  An MP3 audio encoder that receives standard raw input buffers
  and writes standard encoded output buffers.
  Uses code from the Lame project.

************************************************************/

// PCH include
#include "StdAfx.h"

// WIN32 includes
#include <process.h>

// STL includes
#include <string>

// C++ includes
#include <iostream.h>

// Project includes
#include "LameEncoder.h"
#include "Common/MutexProtection.h"
#include "TimeKeepingReaderWriter.h"

// MPEG4 Includes
#include "mpeg4/mp4v2/mp4.h"

using namespace std;

CLameEncoder::CLameEncoder(
						   TimeKeepingReader * pReader, 
						   TimeKeepingWriter * pWriter,
						   unsigned short usChannels, 
							unsigned short usBytesPerSample,
							unsigned short usSampleRate,
							unsigned short usKBitRate
						):
m_pReader(pReader),
m_pWriter(pWriter),
m_pGlobalFlags(NULL),
m_usChannels(usChannels),
m_usBytesPerSample(usBytesPerSample),
m_usSampleRate(usSampleRate),
m_usKBitRate(usKBitRate),
m_uSamplesRead(0),
m_hTermination(CreateEvent(NULL, TRUE, TRUE, NULL))
{
	Init();
}

CLameEncoder::~CLameEncoder() 
{
	// Wait for the thread to exit
	DWORD dwResult = WaitForSingleObject(m_hTermination, INFINITE);
	ASSERT(dwResult == WAIT_OBJECT_0);

	CloseHandle(m_hTermination);

	lame_close(m_pGlobalFlags);
}

static void CLameEncoder_EncodingThreadFunc(LPVOID pParam) {
	
	CLameEncoder * pLameEncoder = (CLameEncoder*) pParam;
	ThreadCompletionAnnouncement announcement(pLameEncoder->m_hTermination);

	try {
		pLameEncoder->Encode();
	}
	catch (string * errMess) {
		cerr << errMess << endl;
		AfxMessageBox(_T(errMess->c_str()));
		delete errMess;
		return;
	}
	catch (...) {
		AfxMessageBox(_T("Unhandled exception in Encoding thread"));
		return;
	}
}

void CLameEncoder::Init()
{
	m_pGlobalFlags = lame_init();

	// Set in-parameters
	lame_set_in_samplerate(m_pGlobalFlags, 	m_usSampleRate);
	lame_set_num_channels(m_pGlobalFlags, m_usChannels);
	lame_set_brate(m_pGlobalFlags, m_usKBitRate);

    int i = lame_init_params(m_pGlobalFlags);
    if (i < 0)
		throw new string("fatal error during initialization");
	
	_beginthread(CLameEncoder_EncodingThreadFunc, 0, this);
}

void CLameEncoder::Encode()
{
    unsigned char mp3buffer[LAME_MAXMP3BUFFER];
    int     Buffer[2][1152];
    int     iread, imp3, frames;

     /* encode until we hit eof */
    do {
        /* read in 'iread' samples */
        iread = GetAudio(Buffer);
        frames = lame_get_frameNum(m_pGlobalFlags);

        /* encode */
        imp3 = lame_encode_buffer_int(m_pGlobalFlags, Buffer[0], Buffer[1], iread,
                                      mp3buffer, sizeof(mp3buffer));

        /* was our output buffer big enough? */
        if (imp3 < 0) {
            if (imp3 == -1)
                throw new string("mp3 buffer is not big enough");
            else {
				string * errMess = new string("mp3 internal error:  error code=");
				(*errMess) += imp3;
				throw errMess;
			}
        }
		
		if (m_pWriter->WriteBytes(mp3buffer, imp3, m_timestamp, m_duration) != imp3)
			throw new string("Failed to write MP3 output");

    } while (iread);

    imp3 = lame_encode_flush(m_pGlobalFlags, mp3buffer, sizeof(mp3buffer)); /* may return one more mp3 frame */

    if (imp3 < 0) {
        if (imp3 == -1)
            throw new string("mp3 buffer is not big enough");
        else {
			string * errMess = new string("mp3 internal error:  error code=");
			(*errMess) += imp3;
			throw errMess;
		}
    }

	m_pWriter->WriteBytes(mp3buffer, imp3, m_timestamp, m_duration);

	m_pWriter->DeclareWritingEnded();
}

#define         Min(A, B)       ((A) < (B) ? (A) : (B))
#define         Max(A, B)       ((A) > (B) ? (A) : (B))
#define			MAX_U_32_NUM            0xFFFFFFFF

int CLameEncoder::GetAudio(int buffer[][1152])
{
    int     num_channels = lame_get_num_channels( m_pGlobalFlags );
    int     samples_read;
    int     framesize;
    int     samples_to_read;
    unsigned int tmp_num_samples;
    int     i;
    int     *p;

    /* 
     * NOTE: LAME can now handle arbritray size input data packets,
     * so there is no reason to read the input data in chuncks of
     * size "framesize".  EXCEPT:  the LAME graphical frame analyzer 
     * will get out of sync if we read more than framesize worth of data.
     */

    samples_to_read = framesize = lame_get_framesize(m_pGlobalFlags);
    ASSERT(framesize <= 1152);

    /* get num_samples */
    tmp_num_samples = lame_get_num_samples( m_pGlobalFlags );

    int     insamp[2 * 1152];

    samples_read =
		ReadPCMSamples(insamp, num_channels * framesize, num_channels * samples_to_read);

	p = insamp + samples_read;
        samples_read /= num_channels;
	if( num_channels == 2 ) {
		for( i = samples_read; --i >= 0; ) {
			buffer[1][i] = *--p;
 			buffer[0][i] = *--p;
		}
	} else if( num_channels == 1 ) {
		memset( buffer[1], 0, samples_read * sizeof(int) );
		for( i = samples_read; --i >= 0; ) {
 			buffer[0][i] = *--p;
		}
	} else
		throw new string("Unsupported number of channels");

    /* if num_samples = MAX_U_32_NUM, then it is considered infinitely long.
       Don't count the samples */
    if ( tmp_num_samples != MAX_U_32_NUM )
        m_uSamplesRead += samples_read;

    return samples_read;

}

int CLameEncoder::ReadPCMSamples(int sample_buffer[], int frame_size, int samples_to_read)
{
    int     samples_read;
    int     iswav = true;
	int		swapbytes = false;
    int     hi_lo_order;	/* byte order of input stream */

	if ((m_usBytesPerSample >= 2) && (m_usBytesPerSample <= 4)) {
				/* assume only recognized wav files are */
				/*  in little endian byte order */
		hi_lo_order = (!iswav == !swapbytes);
        samples_read = GetUnpackedSamples(samples_to_read, m_usBytesPerSample, 
                                           hi_lo_order,sample_buffer);
       
    } else if( 1 == m_usBytesPerSample ) {
		samples_read = GetUnpackedSamples( samples_to_read, 1, 0,
							sample_buffer);
    } else {
        throw new string("Only 8, 16, 24 and 32 bit input files supported");
    }

    return samples_read;
}

int CLameEncoder::GetUnpackedSamples(int samples_to_read, int bytes_per_sample, int swap_order, int *sample_buffer)
{
    int samples_read;
    int i;
    int *op;			/* output pointer */
    unsigned char *ip = (unsigned char *) sample_buffer; /* input pointer */
    const int b = sizeof(int) * 8;

#define GA_URS_IFLOOP( ga_urs_bps ) \
    if( bytes_per_sample == ga_urs_bps ) \
	for( i = samples_read * bytes_per_sample; (i -= bytes_per_sample) >=0;)
    

    samples_read = 	m_pReader->ReadBytes(sample_buffer, 
		bytes_per_sample * samples_to_read, 
		m_timestamp, m_duration) / bytes_per_sample;

    op = sample_buffer + samples_read;

    GA_URS_IFLOOP( 1 )
	*--op = (ip[i] ^ 0x80)<<(b-8) | 0x7f<<(b-16);/* convert from unsigned*/
    if( swap_order == 0 ) {
	GA_URS_IFLOOP( 2 )
	    *--op = ip[i]<<(b-16) | ip[i+1]<<(b-8); 
	GA_URS_IFLOOP( 3 )
	    *--op = ip[i]<<(b-24) | ip[i+1]<<(b-16) | ip[i+2]<<(b-8);
	GA_URS_IFLOOP( 4 )
	    *--op = ip[i]<<(b-32) | ip[i+1]<<(b-24) | ip[i+2]<<(b-16) | ip[i+3] << (b-8);
    } else {
	GA_URS_IFLOOP( 2 )
	    *--op = ip[i]<<(b-8) | ip[i+1]<<(b-16); 
	GA_URS_IFLOOP( 3 )
	    *--op = ip[i]<<(b-8) | ip[i+1]<<(b-16) | ip[i+2]<<(b-24);
	GA_URS_IFLOOP( 4 )
	    *--op = ip[i]<<(b-8) | ip[i+1]<<(b-16) | ip[i+2]<<(b-24) | ip[i+3]<<(b-32);
    }
#undef GA_URS_IFLOOP
    return( samples_read );
}

MediaType CLameEncoder::get_audio_mp4_fileinfo(
			     bool *mpeg4,
			     bool *isma_compliant,
			     uint8_t *audioProfile,
			     uint8_t **audioConfig,
			     uint32_t *audioConfigLen,
			     uint8_t *mp4AudioType)
{
  *mpeg4 = true; // legal in an mp4 - create an iod
  *isma_compliant = false;
  *audioProfile = 0xfe;
  *audioConfig = NULL;
  *audioConfigLen = 0;
  if (mp4AudioType != NULL) {
    *mp4AudioType = MP4_MP3_AUDIO_TYPE;
  }
  return MP3AUDIOFRAME;
}
