/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: MediaDivision.cpp,v 1.3 2004/02/14 12:01:09 gadibergman Exp $
 */

/************************************************************
			MediaDivision.cpp
			=================

  A mechanism of dividing one media type to a number of
  media providers.

  Media frames/samples are written to each provider's
  standard output buffers object, according to the provider's
  registered play/pause times.

************************************************************/


// PCH includes
#include "stdafx.h"

// STL includes
#include <set>
#include <list>

// Project includes
#include "MediaDivision.h"
#include "TimeKeepingReaderWriter.h"
#include "Common/MutexProtection.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

class CTimeBasedMediaProvider : public CMediaProvider {

private:
	void PlayOrPause();
	void Stop();

	void AddTimestamp();

	unsigned long WriteBufferPart
		(
			const void *buf,
			unsigned long len,
			Timestamp tsBuf,
			Duration duration,
			Timestamp tsStart,
			Timestamp tsEnd
		);

	std::list<Timestamp> m_vtsPlayPauseTimes;
	
	CMediaDivision& m_MediaDivision;
	
	TimeKeepingWriter* m_pWriter;

	unsigned long m_ulAtomSize, m_ulAtomsPerSec;

	Timestamp m_tsBufferFirst, m_tsLastEnd;

	LONGLONG m_llNumBytesProvided;

	bool m_bStopRequested, m_bIsPlaying, m_bHasStarted;

	Duration m_PausedDuration;

	LONGLONG GetNumBytesProvided() const
	{
		return m_llNumBytesProvided;
	}

	HANDLE m_hWriterExclusion;

public:
	CTimeBasedMediaProvider	(
								CMediaDivision& mediaDivision,
								TimeKeepingWriter *pWriter,
								unsigned long ulAtomSize,
								unsigned long ulAtomsPerSec
							);

	virtual ~CTimeBasedMediaProvider();

	unsigned long WriteBuffer(	const void * buf, 
								unsigned long len,
								Timestamp timestamp,
								bool& bTerminate
							);

	void DeclareWritingEnded()
	{
		m_pWriter->DeclareWritingEnded();
	}

};

class CMediaProvidersSet : 
public std::set<CTimeBasedMediaProvider*> 
{
public:

	void WriteBuffer(	const void * buf, 
						unsigned long len,
						Timestamp timestamp	
					)
	{
		iterator iter = begin();

		while (iter != end()) 
		{
			CTimeBasedMediaProvider* pMediaProvider = *iter;
			bool bTerminate;
			pMediaProvider->WriteBuffer(buf, len, timestamp, bTerminate);
			if (bTerminate)
				iter = erase(iter);
			else
				++iter;
		}
	}

	void DeclareWritingEnded()
	{
		for (iterator iter = begin(); iter != end(); iter = erase(iter)) {
			CTimeBasedMediaProvider* pMediaProvider = *iter;
			pMediaProvider->DeclareWritingEnded();
			delete pMediaProvider;
		}
	}

	virtual ~CMediaProvidersSet() {
		for (iterator iter = begin(); iter != end(); ++iter) {
			CTimeBasedMediaProvider* pMediaProvider = *iter;
			delete pMediaProvider;
		}
	}
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMediaDivision::CMediaDivision(
								unsigned long ulAtomSize,
								unsigned long ulAtomsPerSec,
								bool bIsRelativeTimestamp
							   ):
m_ulAtomSize(ulAtomSize),
m_ulAtomsPerSec(ulAtomsPerSec),
m_bIsRelativeTimestamp(bIsRelativeTimestamp),
m_ulBytesWritten(0),
m_tsGlobalFirst(GetTimestamp()),
m_tsBufferFirst(0),
m_bHasStarted(false),
m_pvMediaProviders(new CMediaProvidersSet)
{
}

CMediaDivision::~CMediaDivision()
{
	delete m_pvMediaProviders;
}

CMediaProvider* CMediaDivision::CreateMediaProvider(TimeKeepingWriter* pWriter)
{
	CTimeBasedMediaProvider* pMediaProvider = 
		new CTimeBasedMediaProvider	(
										*this, 
										pWriter, 
										m_ulAtomSize,
										m_ulAtomsPerSec
									);

	m_pvMediaProviders->insert(pMediaProvider);
	return pMediaProvider;
}

unsigned long CMediaDivision::WriteBuffer
					(
						const void * buf, 
						unsigned long len,
						Timestamp timestamp	
					)
{
	if (!m_bHasStarted) {
		m_tsBufferFirst = 
			(m_bIsRelativeTimestamp) ? 0 : timestamp;
		m_bHasStarted = true;
	}

	m_pvMediaProviders->WriteBuffer(buf, len, timestamp);

	m_ulBytesWritten += len;
	
	return len;
}

void CMediaDivision::DeclareWritingEnded()
{
	m_pvMediaProviders->DeclareWritingEnded();
}

CTimeBasedMediaProvider::CTimeBasedMediaProvider
	(
			CMediaDivision& mediaDivision,
			TimeKeepingWriter *pWriter,
			unsigned long ulAtomSize,
			unsigned long ulAtomsPerSec
	):
m_MediaDivision(mediaDivision),
m_pWriter(pWriter),
m_ulAtomSize(ulAtomSize),
m_ulAtomsPerSec(ulAtomsPerSec),
m_bStopRequested(false),
m_bIsPlaying(false),
m_bHasStarted(false),
m_PausedDuration(0),
m_llNumBytesProvided(0),
m_hWriterExclusion(CreateMutex(NULL, FALSE, NULL))
{
	AddTimestamp(); // For playing time
}

CTimeBasedMediaProvider::~CTimeBasedMediaProvider()
{
	BOOL bResult = CloseHandle(m_hWriterExclusion);
	ASSERT(bResult);
}

void CTimeBasedMediaProvider::AddTimestamp()
{
	m_vtsPlayPauseTimes.push_back(m_MediaDivision.GetLocalTimestamp());
}

void CTimeBasedMediaProvider::PlayOrPause()
{
	MutexProtection prot(m_hWriterExclusion);
	AddTimestamp();
}

void CTimeBasedMediaProvider::Stop()
{
	MutexProtection prot(m_hWriterExclusion);
	AddTimestamp();
	m_bStopRequested = true;
}

unsigned long CTimeBasedMediaProvider::WriteBuffer
	(	
		const void * buf, 
		unsigned long len,
		Timestamp timestamp,
		bool& bTerminate
	)
{
	bTerminate = false;

	MutexProtection prot(m_hWriterExclusion);

	unsigned long ulAtoms = len / m_ulAtomSize;
	ASSERT(!(len % m_ulAtomSize));

	Timestamp tsRelativeStart = 
		timestamp - m_MediaDivision.GetFirstBufferTimestamp();
	
	Timestamp tsBegin = tsRelativeStart;

	while (!m_vtsPlayPauseTimes.empty()) {
		tsBegin = *m_vtsPlayPauseTimes.begin();
		if (tsBegin > tsRelativeStart) {
			tsBegin = tsRelativeStart;
			break;
		}

		m_bIsPlaying = !m_bIsPlaying;
		m_vtsPlayPauseTimes.pop_front();
	}

	if (tsBegin < tsRelativeStart)
		tsBegin = tsRelativeStart;

	if ( m_bHasStarted && !m_bIsPlaying )
		m_PausedDuration += (tsRelativeStart - m_tsLastEnd);	

	Duration duration = TimestampTicks * ulAtoms / m_ulAtomsPerSec;
	Timestamp tsRelativeEnd = tsRelativeStart + duration;
	m_tsLastEnd = tsRelativeEnd;

	Timestamp tsEnd = tsRelativeEnd;

	unsigned long ulBytesWritten(0);

	while (!m_vtsPlayPauseTimes.empty()) {
		tsEnd = *m_vtsPlayPauseTimes.begin();
		if (tsEnd > tsRelativeEnd) {
			tsEnd = tsRelativeEnd;
			break;
		}

		ulBytesWritten += WriteBufferPart
			(
				buf, len, tsRelativeStart, duration, 
				tsBegin, tsEnd
			);
		
		tsBegin = tsEnd;

		m_bIsPlaying = !m_bIsPlaying;
		m_vtsPlayPauseTimes.pop_front();
	}

	if (m_bStopRequested && 
		m_vtsPlayPauseTimes.empty())
	{
		m_pWriter->DeclareWritingEnded();
		bTerminate = true;
		return ulBytesWritten;	
	}

	ulBytesWritten += WriteBufferPart
		(
			buf, len, tsRelativeStart, duration, 
			tsBegin, tsEnd
		);

	return ulBytesWritten;
}

unsigned long CTimeBasedMediaProvider::WriteBufferPart
		(
			const void *buf,
			unsigned long len,
			Timestamp tsBuf,
			Duration duration,
			Timestamp tsStart,
			Timestamp tsEnd
		)
{
		if (m_bIsPlaying) 
		{
			if (!m_bHasStarted)
			{
				m_tsBufferFirst = tsStart;
				m_bHasStarted = true;
			}

			unsigned long 
				ulStartOffset =	(unsigned long) ( (double) m_ulAtomSize * 0.5 + (double)(__int64)(tsStart - tsBuf) * len / duration ) ,
				ulEndOffset = (unsigned long) ( (double) m_ulAtomSize * 0.5 + (double)(__int64)(tsEnd - tsBuf) * len / duration );

			ulStartOffset -= ulStartOffset % m_ulAtomSize;
			ulEndOffset -= ulEndOffset % m_ulAtomSize;

			unsigned long ulLength = ulEndOffset - ulStartOffset;

			if (!ulLength)
				return 0;

			unsigned long dwBytesPerSec = m_ulAtomSize * m_ulAtomsPerSec;

			Duration dPart = (Duration) ulLength * TimestampTicks / dwBytesPerSec;

			const void * pStart = 
				(const char *) buf + ulStartOffset;

			ASSERT(ulStartOffset + ulLength <= len);

			ASSERT(!(ulLength % m_ulAtomSize));

			m_llNumBytesProvided += ulLength;

			return m_pWriter->WriteBytes(pStart, ulLength,
				tsStart - m_tsBufferFirst - m_PausedDuration, dPart);
		}
		else 
		{
			if (m_bHasStarted)
				m_PausedDuration += (tsEnd - tsStart);

			return 0;
		}
}
