/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Mp4Encoder.cpp,v 1.2 2004/03/27 13:47:16 gadibergman Exp $
 */

/************************************************************
			Mp4Encoder.cpp
			==============

  An encapsulation of the MPEG4IP-mp4live encoder of
  audio samples and video frames into one mp4 file.

  The encoder makes use of:
  - an audio encoder that implements CAudioEncoder. Results
  have only been successfully observed using the 
  FAAC audio encoder.
  - a video encoder in two parts: RGB->YUV & YUV->XVID
  - a file mp4live based combination

************************************************************/


// PCH includes
#include "StdAfx.h"

// Usual includes
#include "MPEGRecorder.h"
#include "Mp4Encoder.h"

// MPEG 4 Includes
#include "mpeg4/combined/file_mp4_recorder.h"
#include "mpeg4/combined/media_frame.h"

// DirectX based receivers
#include "DXSoundReceiver.h"
#include "DXVideoReceiver.h"

// Project includes
#include "TimeBasedPipe.h"
#include "FileNameGenerator.h"
#include "MediaDivision.h"

// Encoders includes
#include "AudioEncoder.h"
#include "AudioEncoderGenerator.h"
#include "rgb2yuv/RGB2YUVConvertor.h"
#include "XVidEncoder.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMp4Encoder::~CMp4Encoder()
{
	m_pAudioProvider->Stop();
	m_pVideoProvider->Stop();

	// Wait for completion of encoding
	delete m_pAudioEncoder;
	delete m_pRGB2YUVConvertor;
	delete m_pXVidEncoder;

	delete m_pPCMAudioPipe;
	delete m_pRGBVideoPipe;
	delete m_pYUVVideoPipe;
	delete m_pAudioTrackWriter;
	delete m_pVideoTrackWriter;
	delete m_pMp4Combination;
}

void CMp4Encoder::Init()
{
	CString sDirectory, sFileName;
	m_FileNameGenerator.GenerateFileName(sDirectory, sFileName);

	CString sFilePath = sDirectory;
	sFilePath += sFileName;

	m_pMp4Combination = new CMp4Recorder(
		m_pSoundReceiver->GetSampleRate(),
		m_pSoundReceiver->GetChannels(),
		m_pVideoReceiver->GetWidth(),
		m_pVideoReceiver->GetHeight(),
		m_pVideoReceiver->GetFrameRate(),
		sFilePath);

	m_pAudioTrackWriter = 
		m_pMp4Combination->CreateTrackWriter(
			m_AudioEncoderGenerator.GetMp4MediaFrameType());

	m_pVideoTrackWriter = 
		m_pMp4Combination->CreateTrackWriter(MPEG4VIDEOFRAME);

	m_pPCMAudioPipe = new CTimeBasedPipe();
	m_pRGBVideoPipe = new CTimeBasedPipe();
	m_pYUVVideoPipe = new CTimeBasedPipe();

	m_pAudioEncoder = 
		m_AudioEncoderGenerator.CreateAudioEncoder(		
			m_pPCMAudioPipe,
			m_pAudioTrackWriter,
			m_pSoundReceiver->GetChannels(),
			m_pSoundReceiver->GetBytesPerSample(),
			m_pSoundReceiver->GetSampleRate(),
			CAudioEncoderGenerator::MP4_FILE_CONTAINER
			);

	m_pRGB2YUVConvertor = new RGB2YUVConvertor(
		m_pRGBVideoPipe,
		m_pYUVVideoPipe,
		m_pVideoReceiver->GetWidth(),
		m_pVideoReceiver->GetHeight(),
		m_dDesiredFrameRate
		);

	m_pXVidEncoder = new CXVidEncoder(
		m_pYUVVideoPipe,
		m_pVideoTrackWriter,
		m_pVideoReceiver->GetWidth(),
		m_pVideoReceiver->GetHeight(),
		//m_pVideoReceiver->GetFrameRate()
		m_dDesiredFrameRate
		);

	m_pMp4Combination->SetAudioEncoder(m_pAudioEncoder);
	m_pMp4Combination->Start();

	m_pAudioProvider = m_pSoundReceiver->CreateMediaProvider(m_pPCMAudioPipe);
	m_pVideoProvider = m_pVideoReceiver->CreateMediaProvider(m_pRGBVideoPipe);
}


void CMp4Encoder::GetProgress(
	double&	time,
	LONGLONG &pcmBytes, 
	LONGLONG &aacBytes, 
	LONGLONG &rgbBytes, 
	LONGLONG &yuvBytes, 
	LONGLONG &xvidBytes
	) const
{
	time = m_pMp4Combination->GetTime();
	pcmBytes = m_pPCMAudioPipe->GetNumBytesWritten();
	aacBytes = m_pAudioTrackWriter->GetNumBytesWritten();
	rgbBytes = m_pRGBVideoPipe->GetNumBytesWritten();
	yuvBytes = m_pYUVVideoPipe->GetNumBytesWritten();
	xvidBytes = m_pVideoTrackWriter->GetNumBytesWritten();
}

void CMp4Encoder::Stop()
{
	m_pAudioProvider->Stop();
	m_pVideoProvider->Stop();
}

void CMp4Encoder::PlayOrPause()
{
	m_pAudioProvider->PlayOrPause();
	m_pVideoProvider->PlayOrPause();
}