/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: RecorderDlg.cpp,v 1.3 2004/04/12 15:36:49 gadibergman Exp $
 */

/************************************************************
			RecorderDlg.cpp
			===============

  Audio Recorder Dialog Box using MFC

  When the recorder pops up, it immediately starts encoding
  audio frames, using the audio encoder, generated by the
  audio encoder generator.

************************************************************/

// PCH includes
#include "stdafx.h"

// Project includes
#include "MPEGRecorder.h"
#include "RecorderDlg.h"
#include "TimeBasedPipe.h"
#include "DXSoundReceiver.h"
#include "LameEncoder.h"
#include "CFileWriter.h"
#include "FileNameGenerator.h"
#include "AudioEncoderGenerator.h"
#include "RenameFile.h"
#include "MediaDivision.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CRecorderDlg dialog


CRecorderDlg::CRecorderDlg(
   const GUID * pRecordingDeviceID,
   CFileNameGenerator & fileNameGenerator,
   CAudioEncoderGenerator & audioEncoderGenerator,
   bool& bInTransition,
   CRecorderDlg ** ppMe,
   CSoundReceiverBase * pSoundReceiver /*=NULL*/,
   CWnd* pParent /*=NULL*/)
	: CDialog(CRecorderDlg::IDD, pParent),
	m_gRecordingDeviceID(*pRecordingDeviceID),
	m_bInTransition(bInTransition),
	m_ppMe(ppMe),
	m_pPipe(new CTimeBasedPipe),
	m_pSoundReceiver(NULL),
	m_pAudioEncoder(NULL),
	m_pFileWriter(NULL),
	m_tInitialTime(0),
	m_FollowUp(NO_FOLLOWUP),
	m_FileNameGenerator(fileNameGenerator),
	m_AudioEncoderGenerator(audioEncoderGenerator),
	m_pSoundProvider(NULL)
{
	//{{AFX_DATA_INIT(CRecorderDlg)
	m_sFileName = _T("");
	m_sMP3Size = _T("");
	m_sTime = _T("");
	m_sWaveSize = _T("");
	m_sEncodedSizeLabel = _T("");
	//}}AFX_DATA_INIT

	m_bInTransition = true;

	CDialog::Create(IDD,pParent);
	StopBitmap.LoadBitmap(IDB_BITMAP_STOP_BUTTON);
	StopThenPlayBitmap.LoadBitmap(IDB_BITMAP_STOP_THEN_PLAY_BUTTON);
	StopThenRecordBitmap.LoadBitmap(IDB_BITMAP_STOP_THEN_RECORD_BUTTON);

	m_StopButton.SetBitmap((HBITMAP) StopBitmap);
	m_StopThenPlayButton.SetBitmap((HBITMAP) StopThenPlayBitmap);
	m_StopThenRecordButton.SetBitmap((HBITMAP) StopThenRecordBitmap);

	m_sEncodedSizeLabel = 
		audioEncoderGenerator.GetDefaultEncodedFileExtension();
	m_sEncodedSizeLabel += _T(" Size:");

	try {
		fileNameGenerator.GenerateFileName(m_sDirectory,
			m_sOriginalFileName);
		
		m_sFileName = m_sOriginalFileName;

		UpdateData(FALSE);
		
		CString sOriginalFilePath = m_sDirectory;
		sOriginalFilePath += m_sOriginalFileName;

		m_pFileWriter = new CFileWriter(sOriginalFilePath);
		
		m_pAudioEncoder = 
			audioEncoderGenerator.CreateAudioEncoder(
			m_pPipe, m_pFileWriter,
			CHANNELS, BYTES_PER_SAMPLE, SAMPLE_RATE, 
			CAudioEncoderGenerator::SIMPLE_FILE_CONTAINER);

		if (pSoundReceiver)
			m_pSoundReceiver = pSoundReceiver;
		else
		{
			m_pSoundReceiver = new CDXSoundReceiver(
				&m_gRecordingDeviceID,
				CHANNELS, BYTES_PER_SAMPLE, SAMPLE_RATE);

			m_pSoundReceiver->Start();
		}

		m_pSoundProvider = m_pSoundReceiver->CreateMediaProvider(m_pPipe);

		m_tInitialTime = time(NULL);
		SetTimer(TIMER_EVENT_ID,300,NULL);
	}
	catch (std::string * errMess) {
		AfxMessageBox(_T(errMess->c_str()));
		delete errMess;
	}
}

void CRecorderDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CRecorderDlg)
	DDX_Control(pDX, IDC_STOP_BUTTON, m_StopButton);
	DDX_Control(pDX, IDC_STOP_THEN_RECORD_BUTTON, m_StopThenRecordButton);
	DDX_Control(pDX, IDC_STOP_THEN_PLAY, m_StopThenPlayButton);
	DDX_Text(pDX, IDC_FILE_NAME, m_sFileName);
	DDX_Text(pDX, IDC_MP3_SIZE, m_sMP3Size);
	DDX_Text(pDX, IDC_TIME, m_sTime);
	DDX_Text(pDX, IDC_WAVE_SIZE, m_sWaveSize);
	DDX_Text(pDX, IDC_EncodedBytesLabel, m_sEncodedSizeLabel);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CRecorderDlg, CDialog)
	//{{AFX_MSG_MAP(CRecorderDlg)
	ON_BN_CLICKED(IDC_STOP_BUTTON, OnStopButton)
	ON_BN_CLICKED(IDC_STOP_THEN_PLAY, OnStopThenPlayButton)
	ON_BN_CLICKED(IDC_STOP_THEN_RECORD_BUTTON, OnStopThenRecordButton)
	ON_WM_TIMER()
	ON_WM_DESTROY()
	ON_WM_CLOSE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CRecorderDlg message handlers

CRecorderDlg::~CRecorderDlg()
{
	ReleaseStructures();

	* m_ppMe = NULL;
}

void CRecorderDlg::OnStopButton() 
{
	// TODO: Add your control notification handler code here
	m_FollowUp = NO_FOLLOWUP;
	Stop();
}

void CRecorderDlg::OnStopThenPlayButton() 
{
	// TODO: Add your control notification handler code here
	m_FollowUp = PLAY_FILE;
	Stop();
}

void CRecorderDlg::OnStopThenRecordButton() 
{
	// TODO: Add your control notification handler code here
	m_FollowUp = RECORD_NEW;
	Stop();
}

void LongDecimalNumberFormat(long num, CString & sNum) {

	sNum = _T("");
	CString sSign = _T("");
	if (num < 0) {
		num = -num;
		sSign = _T("-");
	}

	do {

		long dividedNum = num / 1000;

		TCHAR * sFormat = (dividedNum) ? _T("%03d") : _T("%d");
	
		TCHAR tempBuf[50];
		_stprintf(tempBuf, sFormat, num % 1000);
		
		num = dividedNum;

		CString sTemp;

		if (num > 0)
			sTemp = _T(",");
		else
			sTemp = _T("");

		sTemp += tempBuf;
		sTemp += sNum;

		sNum = sTemp;

	} while (num > 0);

	sNum = sSign + sNum;
}

void CRecorderDlg::OnTimer(UINT nIDEvent) 
{
	if (!m_hWnd)
		return;

	// TODO: Add your message handler code here and/or call default
	if (nIDEvent == TIMER_EVENT_ID) {
		unsigned long uSecondsElapsed = 
			time(NULL) - m_tInitialTime;

		TCHAR sTime[50];
		_stprintf(sTime, "%02d:%02d", uSecondsElapsed / 60,
			uSecondsElapsed % 60);

		UpdateData(TRUE);

		LongDecimalNumberFormat(m_pSoundProvider->GetNumBytesProvided(), m_sWaveSize);
		LongDecimalNumberFormat(m_pFileWriter->GetNumBytesWritten(), m_sMP3Size);

		m_sTime = sTime;

		UpdateData(FALSE);
	}

	CDialog::OnTimer(nIDEvent);
}

#define FREE_AND_ZERO(x) \
{\
	delete x;\
	x = NULL;\
}

void CRecorderDlg::ReleaseStructures() {
	// Note: the order is critical
	FREE_AND_ZERO(m_pSoundReceiver);
	FREE_AND_ZERO(m_pAudioEncoder);
	FREE_AND_ZERO(m_pPipe);
	FREE_AND_ZERO(m_pFileWriter);	
}

void CRecorderDlg::Stop()
{
	UpdateData(TRUE);

	if (m_FollowUp == RECORD_NEW) {
		new CRecorderDlg(&m_gRecordingDeviceID,
			m_FileNameGenerator, m_AudioEncoderGenerator,
			m_bInTransition, m_ppMe,
			m_pSoundReceiver);
		m_pSoundProvider->Stop();
		m_pSoundReceiver = NULL; // Don't release this object here
	}
	else 
		m_pSoundReceiver->Stop();

	ReleaseStructures();

	switch (m_FollowUp) {
	case PLAY_FILE:
		{
			CString sOriginalFilePath = m_sDirectory;
			sOriginalFilePath += m_sOriginalFileName;

			ShellExecute(NULL, "play", sOriginalFilePath, 
				NULL, NULL, SW_SHOW);
		}
		break;
	case NO_FOLLOWUP:
	case RECORD_NEW:
		{
			CRenameFile * renameFile = 
				new CRenameFile(m_sDirectory,
				m_sOriginalFileName, m_sFileName);

		}
		break;
	}

	DestroyWindow();
}


void CRecorderDlg::OnDestroy() 
{
	CDialog::OnDestroy();
	
	delete this;
}

void CRecorderDlg::OnClose() 
{
	Stop();
}

BOOL CRecorderDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	*m_ppMe = this;
	m_bInTransition = false;
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}
