/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: RecordingEventDialog.cpp,v 1.1.1.1 2003/12/24 19:41:40 gadibergman Exp $
 */

/************************************************************
			RecordingEventDialog.cpp
			========================

  A dialog box that enables the user to edit a repeatable
  recording event. This type of recording events is enabled
  if either the video viewer is closed or is held in control
  mode "externally planned".

************************************************************/

#include "stdafx.h"
#include "resource.h"
#include "RecordingEventDialog.h"
#include "RepeatingTimeEvent.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CRecordingEventDialog dialog


CRecordingEventDialog::CRecordingEventDialog(CWnd* pParent /*=NULL*/)
	: CDialog(CRecordingEventDialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(CRecordingEventDialog)
	m_csCurrentDate = _T("");
	m_csCurrentTime = _T("");
	m_Date = 0;
	m_iDayType = -1;
	m_uiDurationInMinutes = 0;
	m_EndTime = 0;
	m_iFrameLimitType = -1;
	m_StartTime = 0;
	//}}AFX_DATA_INIT
}


void CRecordingEventDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CRecordingEventDialog)
	DDX_Control(pDX, IDC_DAYS_LIST, m_DaysList);
	DDX_Text(pDX, IDC_CURRENT_DATE, m_csCurrentDate);
	DDX_Text(pDX, IDC_CURRENT_TIME, m_csCurrentTime);
	DDX_DateTimeCtrl(pDX, IDC_DATE, m_Date);
	DDX_Radio(pDX, IDC_DATE_RADIO, m_iDayType);
	DDX_Text(pDX, IDC_DURATION_IN_MINUTES, m_uiDurationInMinutes);
	DDX_DateTimeCtrl(pDX, IDC_END_TIME, m_EndTime);
	DDX_Radio(pDX, IDC_END_TIME_RADIO, m_iFrameLimitType);
	DDX_DateTimeCtrl(pDX, IDC_START_TIME, m_StartTime);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CRecordingEventDialog, CDialog)
	//{{AFX_MSG_MAP(CRecordingEventDialog)
	ON_WM_TIMER()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CRecordingEventDialog message handlers

#define THROW_CSTR(x) throw CString(_T(#x))

bool CRecordingEventDialog::EditRecordingEvent(CRepeatingTimeEvent &tEvent)
{
	pTimeEvent = &tEvent;

	int iExitCode = DoModal();

	return (iExitCode == IDOK);
}

void CRecordingEventDialog::SetByTimeEvent(const CRepeatingTimeEvent &tEvent)
{
	const CSimpleDatesVector& dates = tEvent.GetDatesVector();
	const CWeekdaysSelection& wdays = tEvent.GetWeekdaysSelection();
	const CSimpleTimeFramesVector& tframes = tEvent.GetTimeFrames();

	if (dates.size() > 1)
		THROW_CSTR(more than one date specified);

	int i, iNumDaysSelected(0);
	bool vbDaysSelection[1 + DAYS_IN_WEEK];
	for (i = 1; i <= DAYS_IN_WEEK; ++i)
	{
		bool bIsSelected = wdays.IsDaySelected(i);
		vbDaysSelection[i] = bIsSelected;
		iNumDaysSelected += bIsSelected;
	}

	if (iNumDaysSelected && dates.size())
		THROW_CSTR(selection of both date and weekday);

	if (tframes.size() > 1)
		THROW_CSTR(more than one simple time frame selected);

	if (dates.size())
	{
		m_iDayType = DAY_IS_DATE;
		const CSimpleDate& date = *dates.begin();
		m_Date = CTime(
			date.m_year, 
			date.m_month,
			date.m_day,
			0, 0, 0);
	}
	else
	if (iNumDaysSelected)
	{
		m_iDayType = DAY_IS_WEEKDAY;
		for (i = 1; i <= DAYS_IN_WEEK; ++i)
		{
			m_DaysList.SetSel(i - 1, vbDaysSelection[i]);
		}		
	}
	else
	{
		m_iDayType = DAY_IS_DATE;
		m_Date = CTime::GetCurrentTime();
		m_DaysList.SetSel(m_Date.GetDayOfWeek() - 1, true);
	}

	CTime tCurrent = CTime::GetCurrentTime();

	if (tframes.empty())
	{
		m_StartTime = CTime(
			tCurrent.GetYear(),
			tCurrent.GetMonth(),
			tCurrent.GetDay(),
			tCurrent.GetHour(),
			tCurrent.GetMinute(),
			0);
		m_iFrameLimitType = TIME_FRAME_LIMIT_IS_DURATION;
		m_uiDurationInMinutes = 30;
		m_EndTime = m_StartTime + CTimeSpan( 0, 1, 0, 0 ); 
	}
	else
	{
		const CSimpleTimeFrame& tframe = *tframes.begin();
		m_StartTime = CTime(
			tCurrent.GetYear(),
			tCurrent.GetMonth(),
			tCurrent.GetDay(),
			tframe.m_tStart.m_hour,
			tframe.m_tStart.m_minute,
			0);

		switch (tframe.m_LimitType) {
		case CSimpleTimeFrame::LIMIT_BY_DURATION:
			m_iFrameLimitType = TIME_FRAME_LIMIT_IS_DURATION;
			m_uiDurationInMinutes = 
				tframe.m_tDuration.m_minutes + 
				60 * (tframe.m_tDuration.m_hours + 
				24 * tframe.m_tDuration.m_days);
			break;
		case CSimpleTimeFrame::LIMIT_BY_END_TIME:
			m_iFrameLimitType = TIME_FRAME_LIMIT_IS_END_TIME;
			m_EndTime = CTime(
				m_Date.GetYear(),
				m_Date.GetMonth(),
				m_Date.GetDay(),
				tframe.m_tEnd.m_hour,
				tframe.m_tEnd.m_minute,
				0);
			break;
		}
	}
	
	UpdateData(FALSE);
}

void CRecordingEventDialog::SetTimeEvent(CRepeatingTimeEvent &tEvent) const
{
	CSimpleDatesVector& dates = tEvent.GetDatesVector();
	CWeekdaysSelection& wdays = tEvent.GetWeekdaysSelection();
	CSimpleTimeFramesVector& tframes = tEvent.GetTimeFrames();

	dates.clear();
	tframes.clear();
	wdays.SetNoSelection();
	int i, iNumDaysSelected(0);

	switch (m_iDayType) {
	case DAY_IS_DATE:
		++iNumDaysSelected;
		dates.push_back(
			CSimpleDate(
				m_Date.GetYear(),
				m_Date.GetMonth(),
				m_Date.GetDay()
				));
		break;
	case DAY_IS_WEEKDAY:
		for (i = 1; i <= DAYS_IN_WEEK; ++i)
		{
			bool bIsSelected = !!m_DaysList.GetSel(i - 1);
			iNumDaysSelected += bIsSelected;
			wdays.SetDaySelection(i, bIsSelected);
		}
		break;
	default:
		THROW_CSTR(No date type selected);
	}

	if (!iNumDaysSelected)
		THROW_CSTR(No day selected);

	CSimpleTime tStart(
		m_StartTime.GetHour(),
		m_StartTime.GetMinute()
		);

	switch (m_iFrameLimitType) {
	case TIME_FRAME_LIMIT_IS_DURATION:
		{
			int iMinutes = m_uiDurationInMinutes;
			int iHours = iMinutes / 60;
			iMinutes %= 60;
			int iDays = iHours / 24;
			iHours %= 24;
			SimpleDuration duration(iDays, iHours, iMinutes);

			tframes.push_back(
				CSimpleTimeFrame(tStart,duration)
				);

		}
		break;
	case TIME_FRAME_LIMIT_IS_END_TIME:
		{
			CSimpleTime tEnd(
				m_EndTime.GetHour(),
				m_EndTime.GetMinute()
				);

			tframes.push_back(
				CSimpleTimeFrame(tStart,tEnd)
				);
		}
		break;
	default:
		THROW_CSTR(No frame ending selected);
	}
}

void CRecordingEventDialog::FillDaysList()
{
	m_DaysList.ResetContent( );

	for (int i = 0; i < DAYS_IN_WEEK; ++i)
	{
		m_DaysList.AddString(days_of_the_week[i]);
	}
}

BOOL CRecordingEventDialog::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	FillDaysList();
	m_uiTimerID = SetTimer(MINUTE_TIMER_EVENT_ID, MINUTE_IN_MS, NULL);
	DisplayCurrentTime();

	try {
		SetByTimeEvent(*pTimeEvent);
	}
	catch (CString cstrErr) {
		AfxMessageBox(cstrErr);
	}

	return TRUE;	            
}

void CRecordingEventDialog::OnOK() 
{
	BOOL bRes = KillTimer(m_uiTimerID);
	ASSERT(bRes);

	try {
		UpdateData(TRUE);
		SetTimeEvent(*pTimeEvent);		
	}
	catch (CString cstrErr) {
		AfxMessageBox(cstrErr);
		return;
	}
	
	CDialog::OnOK();
}

void CRecordingEventDialog::OnCancel() 
{
	BOOL bRes = KillTimer(m_uiTimerID);
	ASSERT(bRes);
	
	CDialog::OnCancel();
}

void CRecordingEventDialog::OnTimer(UINT nIDEvent) 
{
	DisplayCurrentTime();
	CDialog::OnTimer(nIDEvent);
}

void CRecordingEventDialog::DisplayCurrentTime()
{
	CSimpleTimePoint tp;
	tp.SetLocalTime();
	std::string strDate, strTime;
	tp.m_Date.GetString(strDate);
	tp.m_Time.GetString(strTime);
	m_csCurrentDate = strDate.c_str();
	m_csCurrentTime = strTime.c_str();
	UpdateData(FALSE);
}
