/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: RepeatingTimeEvent.cpp,v 1.1.1.1 2003/12/24 19:41:42 gadibergman Exp $
 */

/************************************************************
			RepeatingTimeEvent.cpp
			======================

  - Manipulations of date and time.
  - Definition of time points and time frames.
  - Definition of a repeating time event.
  - Finding the earliest event occuring after a given time.

************************************************************/

#include "stdafx.h"
#include "RepeatingTimeEvent.h"
#include <time.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

LPCTSTR days_of_the_week[DAYS_IN_WEEK] =
{
	_T("Sunday"),
	_T("Monday"),
	_T("Tuesday"),
	_T("Wednesday"),
	_T("Thursday"),
	_T("Friday"),
	_T("Saturday")
};

CSimpleDate::CSimpleDate(int year, int month, int day):
m_year(year),
m_month(month),
m_day(day)
{}

CSimpleDate::CSimpleDate(const CSimpleDate& other):
m_year(other.m_year),
m_month(other.m_month),
m_day(other.m_day)
{}

CSimpleDate::CSimpleDate():
m_year(0),
m_month(0),
m_day(0)
{}

bool CSimpleDate::IsValid() const {
	return (m_year >= 1980) &&
		(m_month >= 1) && (m_month <= 12) &&
		(m_day >= 1) && (m_day <= 31);
}

bool CSimpleDate::operator < (const CSimpleDate& other) const 
{
	return (m_year < other.m_year) ||
			(m_year == other.m_year) &&
			( (m_month < other.m_month) ||
				(m_month == other.m_month) &&
				(m_day < other.m_day) );
}

bool CSimpleDate::operator == (const CSimpleDate& other) const {
	return (m_year == other.m_year) &&
		(m_month == other.m_month) &&
		(m_day == other.m_day);
}

CSimpleDate& CSimpleDate::operator ++ () {

#ifndef _DEBUG
	if (m_day < 28) {
		++m_day;
		return *this;
	}
#endif
	struct tm tmCurrent = {
		0						/* tm_sec */,
		0						/* tm_min */,
		0						/* tm_hour */,
		m_day					/* tm_mday */,
		m_month - 1				/* tm_mon */,
		m_year - 1900			/* tm_year */,
		0						/* tm_wday */,
		0						/* tm_yday */,
		0						/* tm_isdst */
	};

	time_t tCurrent = mktime(&tmCurrent);

	ASSERT(tCurrent != -1);
	
	tCurrent += (60 * 60 * 24);

	struct tm tmNew = *localtime(&tCurrent);

	m_day = tmNew.tm_mday;
	m_month = 1 + tmNew.tm_mon;
	m_year = 1900 + tmNew.tm_year;

	return *this;	
}

CSimpleDate& CSimpleDate::operator -- () {

#ifndef _DEBUG
	if (m_day > 1) {
		--m_day;
		return *this;
	}
#endif
	struct tm tmCurrent = {
		0						/* tm_sec */,
		0						/* tm_min */,
		0						/* tm_hour */,
		m_day					/* tm_mday */,
		m_month - 1				/* tm_mon */,
		m_year - 1900			/* tm_year */,
		0						/* tm_wday */,
		0						/* tm_yday */,
		0						/* tm_isdst */
	};

	time_t tCurrent = mktime(&tmCurrent);

	ASSERT(tCurrent != -1);
	
	tCurrent -= (60 * 60 * 24);

	struct tm tmNew = *localtime(&tCurrent);

	m_day = tmNew.tm_mday;
	m_month = 1 + tmNew.tm_mon;
	m_year = 1900 + tmNew.tm_year;

	return *this;	
}

void CSimpleDate::GetString(std::string& str) const
{
	char chBuf[100];
	sprintf(chBuf, "%02d/%02d/%04d",
		m_day, m_month, m_year);
	str = chBuf;
}

SimpleDuration::SimpleDuration(int days, int hours, int minutes):
m_days(days),
m_hours(hours),
m_minutes(minutes)
{}

SimpleDuration::SimpleDuration():
m_days(0),
m_hours(0),
m_minutes(0)
{}

SimpleDuration::SimpleDuration(const SimpleDuration& other):
m_days(other.m_days),
m_hours(other.m_hours),
m_minutes(other.m_minutes)
{}

void SimpleDuration::GetString(std::string& str) const
{
	char chBuf[100];
	if (m_days)
		sprintf(chBuf, "%d days, %d hours and %d minutes",
		m_days, m_hours, m_minutes);
	else
	if (m_hours)
		sprintf(chBuf, "%d hours and %d minutes",
		m_hours, m_minutes);
	else
		sprintf(chBuf, "%d minutes",
		m_minutes);
	str = chBuf;
}

CSimpleTime::CSimpleTime(int hour, int minute):
m_hour(hour),
m_minute(minute)
{}

CSimpleTime::CSimpleTime():
m_hour(0),
m_minute(0)
{}

CSimpleTime::CSimpleTime(const CSimpleTime& other):
m_hour(other.m_hour),
m_minute(other.m_minute)
{}

bool CSimpleTime::IsValid() const {
	return (m_hour >= 0) && (m_hour < 24) &&
		(m_minute >= 0) && (m_minute < 60);
}

bool CSimpleTime::operator < (const CSimpleTime& other) const {
	return (m_hour < other.m_hour) ||
		(m_hour == other.m_hour) &&
		(m_minute < other.m_minute);
}

bool CSimpleTime::operator == (const CSimpleTime& other) const {
	return (m_hour == other.m_hour) && 
		(m_minute == other.m_minute);
}

CSimpleTime& CSimpleTime::operator += (const SimpleDuration& duration) {
	m_minute += duration.m_minutes;
	m_hour += duration.m_hours + m_minute / 60;
	m_minute = m_minute % 60;
	return *this;
}

void CSimpleTime::GetString(std::string& str) const
{
	char chBuf[100];
	sprintf(chBuf, "%02d:%02d", m_hour, m_minute);
	str = chBuf;
}

CSimpleTimePoint::CSimpleTimePoint	(
									const CSimpleDate& date, 
									const CSimpleTime& time
									):
m_Date(date),
m_Time(time)
{}

bool CSimpleTimePoint::operator < (const CSimpleTimePoint& tpOther) const {
	return (m_Date < tpOther.m_Date) ||
		(m_Date == tpOther.m_Date) &&
		(m_Time < tpOther.m_Time);
}

bool CSimpleTimePoint::operator == (const CSimpleTimePoint& tpOther) const {
	return (m_Date == tpOther.m_Date) &&
		(m_Time == tpOther.m_Time);
}

CSimpleTimePoint& CSimpleTimePoint::operator += (const SimpleDuration& duration) {

	struct tm tmCurrent = {
		0						/* tm_sec */,
		m_Time.m_minute			/* tm_min */,
		m_Time.m_hour			/* tm_hour */,
		m_Date.m_day			/* tm_mday */,
		m_Date.m_month - 1      /* tm_mon */,
		m_Date.m_year - 1900	/* tm_year */,
		0						/* tm_wday */,
		0						/* tm_yday */,
		0						/* tm_isdst */
	};

	time_t tCurrent = mktime(&tmCurrent);

	ASSERT(tCurrent != -1);
	
	tCurrent +=
		(60 * (duration.m_minutes + 60 * (duration.m_hours
		+ 24 * duration.m_days)));

	struct tm tmNew = *localtime(&tCurrent);

	m_Time.m_minute = tmNew.tm_min;
	m_Time.m_hour = tmNew.tm_hour;
	m_Date.m_day = tmNew.tm_mday;
	m_Date.m_month = 1 + tmNew.tm_mon;
	m_Date.m_year = 1900 + tmNew.tm_year;

	return *this;
}

void CSimpleTimePoint::SetLocalTime()
{
	time_t currTime = time(NULL);
	struct tm tmLocal = *localtime( &currTime );
	m_Time.m_minute = tmLocal.tm_min;
	m_Time.m_hour = tmLocal.tm_hour;
	m_Date.m_day = tmLocal.tm_mday;
	m_Date.m_month = 1 + tmLocal.tm_mon;
	m_Date.m_year = 1900 + tmLocal.tm_year;	
}

void CSimpleTimePoint::GetString(std::string& str) const
{
	std::string strDate, strTime;
	m_Date.GetString(strDate);
	m_Time.GetString(strTime);
	str = strDate;
	str += ' ';
	str += strTime;
}

CSimpleTimeFrame::CSimpleTimeFrame(const CSimpleTime& tStart,
	const CSimpleTime& tEnd):
m_tStart(tStart),
m_LimitType(LIMIT_BY_END_TIME),
m_tEnd(tEnd)
{}

CSimpleTimeFrame::CSimpleTimeFrame(const CSimpleTime& tStart,
	const SimpleDuration& tDuration):
m_tStart(tStart),
m_LimitType(LIMIT_BY_DURATION),
m_tDuration(tDuration)
{}

void CSimpleTimeFrame::GetEndTime(CSimpleTime& tEnd) const {
	switch (m_LimitType) {
	case LIMIT_BY_DURATION:
		tEnd = m_tStart;
		tEnd += m_tDuration;
		return;
	case LIMIT_BY_END_TIME:
		tEnd = m_tEnd;
		return;
	default:
		ASSERT(FALSE);
	}
}

void CSimpleTimeFrame::GetEndPoint(const CSimpleDate& dStart,
								   CSimpleTimePoint& tpEnd) const {
	switch (m_LimitType) {
	case LIMIT_BY_DURATION:
		tpEnd = CSimpleTimePoint(dStart, m_tStart);
		tpEnd += m_tDuration;
		break;
	case LIMIT_BY_END_TIME:
		tpEnd = CSimpleTimePoint(dStart, m_tEnd);
		break;
	default:
		ASSERT(FALSE);
	}
}

bool CSimpleTimeFrame::GetEarliestEventAfter(
							bool bHasPrev,
							const CSimpleDate& dPrev,
							bool bHasSameDate,
							bool bHasNext,
							const CSimpleDate& dNext,
							const CSimpleTimePoint& tpStart,
							CSimpleTimePoint& tpNext,
							bool& bIsEnding ) const 
{
	if (bHasPrev) {
		GetEndPoint(dPrev, tpNext);
		if (tpStart < tpNext) {
			bIsEnding = true;
			return true;
		}
	}

	if (bHasSameDate) {
		if (tpStart.m_Time < m_tStart) {
			tpNext = CSimpleTimePoint(tpStart.m_Date, m_tStart);
			bIsEnding = false;
			return true;
		}

		GetEndPoint(tpStart.m_Date, tpNext);
		if (tpStart < tpNext) {
			bIsEnding = true;
			return true;
		}
	}

	if (bHasNext) {
		tpNext = CSimpleTimePoint(dNext, m_tStart);
		bIsEnding = false;
		return true;
	}

	return false;
}

void CSimpleTimeFrame::GetString(std::string& str) const
{
	std::string strStartTime;
	m_tStart.GetString(strStartTime);

	switch (m_LimitType) {
	case LIMIT_BY_DURATION:
		{
			m_tDuration.GetString(str);
			str += " starting from ";
			str += strStartTime;
		}
		break;
	case LIMIT_BY_END_TIME:
		{
			std::string strEndTime;
			m_tEnd.GetString(strEndTime);
			str = "from ";
			str += strStartTime;
			str += " to ";
			str += strEndTime;
		}
		break;
	default:
		ASSERT(FALSE);
	}
}

bool CSimpleTimeFramesVector::IsLegal() const {
	for (const_iterator iter1 = begin(); iter1 != end(); ++iter1) {
		const CSimpleTimeFrame& tf1 = *iter1;
		for (const_iterator iter2 = begin(); iter2 != end(); ++iter2) {
			const CSimpleTimeFrame& tf2 = *iter2;
			if ( (tf1.m_tStart < tf2.m_tStart) ||
				 (tf1.m_tStart == tf2.m_tStart) )
			{
				CSimpleTime tf1End;
				tf1.GetEndTime(tf1End);
				if (tf2.m_tStart < tf1End)
					return false;
			}
		}
	}
	return true;
}

bool CSimpleTimeFramesVector::GetEarliestEventAfter(
							bool bHasPrev,
							const CSimpleDate& dPrev,
							bool bHasSameDate,
							bool bHasNext,
							const CSimpleDate& dNext,
							const CSimpleTimePoint& tpStart,
							CSimpleTimePoint& tpNext,
							bool& bIsEnding ) const 
{
	bool bFound(false);
	for (const_iterator iter = begin(); iter != end(); ++iter) {
		const CSimpleTimeFrame& tfCurr = *iter;
		CSimpleTimePoint tpCurrNext;
		bool bCurrEnding;
		
		bool bFoundThisFrame =
		tfCurr.GetEarliestEventAfter(
										bHasPrev, dPrev,
										bHasSameDate, 
										bHasNext, dNext, 
										tpStart, tpCurrNext, 
										bCurrEnding
									);

		if ( bFoundThisFrame &&
			( !bFound || (tpCurrNext < tpNext) ) ) {
			bFound = true;
			tpNext = tpCurrNext;
			bIsEnding = bCurrEnding;
		}

	}

	return bFound;
}

void CSimpleTimeFramesVector::GetString(std::string& str) const
{
	if (empty())
	{
		str = "no time frames";
		return;
	}

	const_iterator iter = begin();
	const CSimpleTimeFrame& tFrame = *iter;
	tFrame.GetString(str);
	++iter;

	while (iter != end())
	{
		std::string strCur;
		const CSimpleTimeFrame& tFrame = *iter;
		tFrame.GetString(strCur);
		str += "; ";
		str += strCur;
		++iter;
	}
}

void CSiblingDatesGenerator::GetSiblingDates(
	const CSimpleDate& dRelative,
	bool& bHasPrev, CSimpleDate& dPrev,
	bool& bHasSameDate,
	bool& bHasNext, CSimpleDate& dNext) const
{
	bHasPrev = GetLatestDateBefore(dRelative, dPrev, false);
	bHasNext = GetEarliestDateAfter(dRelative, dNext, true);

	if (bHasNext && (dNext == dRelative)) {
		bHasSameDate = true;
		bHasNext = GetEarliestDateAfter(dRelative, dNext, false);
	}
	else
		bHasSameDate = false;
}


bool CSimpleDatesVector::
GetEarliestDateAfter(const CSimpleDate& dStartFrom,	
					 CSimpleDate& dEarliest, bool bInclude) const
{
	bool bFoundOne(false);

	for (const_iterator iter = begin(); iter != end(); ++iter) {
		const CSimpleDate& dCurr = *iter;
		if ( (dStartFrom < dCurr) ||
			bInclude && (dStartFrom == dCurr) )
		{
			if (bFoundOne) {
				if (dCurr < dEarliest)
					dEarliest = dCurr;
			}
			else {
				dEarliest = dCurr;
				bFoundOne = true;
			}
		}
	}

	return bFoundOne;
}

bool CSimpleDatesVector::
GetLatestDateBefore(const CSimpleDate& dRelative,
					CSimpleDate& dLatest, bool bInclude) const
{
	bool bFoundOne(false);

	for (const_iterator iter = begin(); iter != end(); ++iter) {
		const CSimpleDate& dCurr = *iter;
		if ( (dCurr < dRelative) ||
			 bInclude && (dCurr == dRelative) )
		{
			if (bFoundOne) {
				if (dLatest < dCurr)
					dLatest = dCurr;
			}
			else {
				dLatest = dCurr;
				bFoundOne = true;
			}
		}
	}

	return bFoundOne;
}

void CSimpleDatesVector::GetString(std::string& str) const
{
	if (empty())
	{
		str = "";
		return;
	}

	const_iterator iter = begin();
	CSimpleDate date = *iter;
	date.GetString(str);
	++iter;

	while (iter != end())
	{
		std::string strCur;
		CSimpleDate date = *iter;
		date.GetString(strCur);
		str += "; ";
		str += strCur;
		++iter;
	}
}

CWeekdaysSelection::CWeekdaysSelection() {
	SetNoSelection();
}

void CWeekdaysSelection::SetNoSelection()
{
	for (int i = 1; i <= DAYS_IN_WEEK; ++i)
		m_vbSelection[i] = false;
}

bool CWeekdaysSelection::IsDaySelected(int iDayNumber) const
{
	ASSERT((iDayNumber >= 1) && (iDayNumber <= DAYS_IN_WEEK));
	return m_vbSelection[iDayNumber];
}

void CWeekdaysSelection::SetDaySelection(int iDayNumber, bool bSelected) {
	ASSERT((iDayNumber >= 1) && (iDayNumber <= DAYS_IN_WEEK));
	m_vbSelection[iDayNumber] = bSelected;
}

bool CWeekdaysSelection::GetEarliestDateAfter(
		const CSimpleDate& dStartFrom,
		CSimpleDate& dEarliest, 
		bool bInclude
											) const
{
	struct tm tmStart = {
		0						/* tm_sec */,
		0						/* tm_min */,
		0						/* tm_hour */,
		dStartFrom.m_day		/* tm_mday */,
		dStartFrom.m_month - 1  /* tm_mon */,
		dStartFrom.m_year - 1900/* tm_year */,
		0						/* tm_wday */,
		0						/* tm_yday */,
		0						/* tm_isdst */
	};

	mktime(&tmStart);

	int iStartWeekDay = tmStart.tm_wday + 1;
	int iWeekDay = iStartWeekDay;

	dEarliest = dStartFrom;

	if (!bInclude) {
		iWeekDay = (iWeekDay % DAYS_IN_WEEK) + 1;
		++dEarliest;		
	}

	for (int i = 1; i <= DAYS_IN_WEEK; ++i) {
		if (m_vbSelection[iWeekDay])
			return true;

		iWeekDay = (iWeekDay % DAYS_IN_WEEK) + 1;
		++dEarliest;
	}

	return false;
}

bool CWeekdaysSelection::GetLatestDateBefore	(
						const CSimpleDate& dRelative,
						CSimpleDate& dLatest, 
						bool bInclude			) const
{
	struct tm tmStart = {
		0						/* tm_sec */,
		0						/* tm_min */,
		0						/* tm_hour */,
		dRelative.m_day			/* tm_mday */,
		dRelative.m_month - 1	/* tm_mon */,
		dRelative.m_year - 1900	/* tm_year */,
		0						/* tm_wday */,
		0						/* tm_yday */,
		0						/* tm_isdst */
	};

	mktime(&tmStart);

	int iStartWeekDay = tmStart.tm_wday + 1;
	int iWeekDay = iStartWeekDay;

	dLatest = dRelative;

	if (!bInclude) {
		iWeekDay = ((iWeekDay + DAYS_IN_WEEK - 2) % DAYS_IN_WEEK) + 1;
		--dLatest;
	}

	for (int i = 1; i <= DAYS_IN_WEEK; ++i) {
		if (m_vbSelection[iWeekDay])
			return true;

		iWeekDay = ((iWeekDay + DAYS_IN_WEEK - 2) % DAYS_IN_WEEK) + 1;
		--dLatest;
	}

	return false;
}

void CWeekdaysSelection::GetString(std::string& str) const
{
	str = "";

	for (int i = 1; i <= DAYS_IN_WEEK; ++i) {
		if (m_vbSelection[i])
		{
			if (str.empty())
			{
				str = days_of_the_week[i - 1];
			}
			else
			{
				str += "; ";
				str += days_of_the_week[i - 1];
			}
		}
	}
}

void CRepeatingTimeEvent::SetDaySelection(int iDayNumber, bool bSelected) {
	m_WeekdaysSelection.SetDaySelection(iDayNumber, bSelected);
}

void CRepeatingTimeEvent::AddSimpleDate(const CSimpleDate& date) {
	m_DatesVector.push_back(date);
}

void CRepeatingTimeEvent::AddTimeFrame(const CSimpleTimeFrame& tframe) {
	m_TimeFramesVector.push_back(tframe);
}

bool CRepeatingTimeEvent::GetEarliestStateChange(
							const CSimpleTimePoint& tpStart,
							CSimpleTimePoint& tpEarliest,
							bool& bIsEnding
							) const
{
	CSimpleDate 
		dWeekPrev, dSimplePrev, dPrev,
		dWeekNext, dSimpleNext, dNext;
	bool
		bWeekHasPrev, bWeekHasSame, bWeekHasNext,
		bSimpleHasPrev, bSimpleHasSame, bSimpleHasNext,
		bHasPrev, bHasSameDate, bHasNext;

	m_WeekdaysSelection.GetSiblingDates(tpStart.m_Date,
		bWeekHasPrev, dWeekPrev, bWeekHasSame, bWeekHasNext, dWeekNext);

	m_DatesVector.GetSiblingDates(tpStart.m_Date,
		bSimpleHasPrev, dSimplePrev, bSimpleHasSame,
		bSimpleHasNext, dSimpleNext);

	if (!bWeekHasPrev && !bSimpleHasPrev)
		bHasPrev = false;
	else {
		bHasPrev = true;
		if (!bWeekHasPrev)
			dPrev = dSimplePrev;
		else
		if (!bSimpleHasPrev)
			dPrev = dWeekPrev;
		else
			dPrev = (dWeekPrev < dSimplePrev) ? dSimplePrev : dWeekPrev;
	}

	bHasSameDate = bWeekHasSame || bSimpleHasSame;

	if (!bWeekHasNext && !bSimpleHasNext)
		bHasNext = false;
	else {
		bHasNext = true;
		if (!bWeekHasNext)
			dNext = dSimpleNext;
		else
		if (!bSimpleHasNext)
			dNext = dWeekNext;
		else
			dNext = (dWeekNext < dSimpleNext) ? dWeekNext : dSimpleNext;
	}
	
	return m_TimeFramesVector.GetEarliestEventAfter(
		bHasPrev, dPrev, bHasSameDate, bHasNext, dNext, tpStart,
		tpEarliest, bIsEnding);
}

void CRepeatingTimeEvent::GetString(std::string& str) const
{
	m_TimeFramesVector.GetString(str);
	std::string strDates, strWDays;
	m_DatesVector.GetString(strDates);
	m_WeekdaysSelection.GetString(strWDays);

	if (!strDates.empty() || !strWDays.empty())
	{
		str += " on ";

		str += strDates;

		if (!strDates.empty() && !strWDays.empty())
		{
			str += " and on ";
		}

		str += strWDays;
	}
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CRepeatingTimeEvent::CRepeatingTimeEvent()
{
}

CRepeatingTimeEvent::~CRepeatingTimeEvent()
{
}

bool CTimeEventsVector::GetEarliestStateChange(
					const CSimpleTimePoint& tpStart,
					CSimpleTimePoint& tpEarliest,
					bool& bIsEnding,
					CRepeatingTimeEvent& tEvent
					) const
{
	if (empty())
		return false;

	const_iterator iter = begin();
	tEvent = *iter;
	bool bFound = 
		tEvent.GetEarliestStateChange(tpStart, tpEarliest, bIsEnding);

	++iter;
	while (iter != end())
	{
		CSimpleTimePoint _tpEarliest;
		bool _bIsEnding, _bFound;
		const CRepeatingTimeEvent& _tEvent = *iter;
		_bFound = 
			_tEvent.GetEarliestStateChange(tpStart, _tpEarliest, _bIsEnding);

		if (_bFound &&
			(!bFound || (_tpEarliest < tpEarliest)) )
		{
			tpEarliest = _tpEarliest;
			bIsEnding = _bIsEnding;
			tEvent = _tEvent;
			bFound = true;
		}

		++iter;
	}

	return bFound;
}

