/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: RepeatingTimeEvent.h,v 1.1.1.1 2003/12/24 19:41:42 gadibergman Exp $
 */

/************************************************************
			RepeatingTimeEvent.h
			====================

  - Manipulations of date and time.
  - Definition of time points and time frames.
  - Definition of a repeating time event.
  - Finding the earliest event occuring after a given time.

************************************************************/

#if !defined(AFX_REPEATINGTIMEEVENT_H__F48F194D_BBF3_49D8_8676_FDCD0FAA60F2__INCLUDED_)
#define AFX_REPEATINGTIMEEVENT_H__F48F194D_BBF3_49D8_8676_FDCD0FAA60F2__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

// Includes
#include <vector>
#include <string>

enum {
	DAYS_IN_WEEK = 7
};

extern LPCTSTR days_of_the_week[DAYS_IN_WEEK];

struct CSimpleDate {
	int m_year;
	int m_month;
	int m_day;

	CSimpleDate(int year, int month, int day);
	CSimpleDate(const CSimpleDate& other);
	CSimpleDate();

	bool IsValid() const;
	bool operator < (const CSimpleDate& other) const;
	bool operator == (const CSimpleDate& other) const;

	CSimpleDate& operator ++ ();
	CSimpleDate& operator -- ();

	void GetString(std::string& str) const;
};

struct SimpleDuration {
	int m_minutes, m_hours, m_days;

	SimpleDuration(int days, int hours, int minutes);
	SimpleDuration();
	SimpleDuration(const SimpleDuration& other);

	void GetString(std::string& str) const;
};

struct CSimpleTime {
	int m_hour;
	int m_minute;

	CSimpleTime(int hour, int minute);
	CSimpleTime();
	CSimpleTime(const CSimpleTime& other);

	CSimpleTime& operator += (const SimpleDuration& duration);
	bool IsValid() const;
	bool operator < (const CSimpleTime& other) const;
	bool operator == (const CSimpleTime& other) const;

	void GetString(std::string& str) const;
};

struct CSimpleTimePoint {
	CSimpleDate m_Date;
	CSimpleTime m_Time;

	CSimpleTimePoint() {}

	CSimpleTimePoint(const CSimpleDate& date, 
		const CSimpleTime& time);

	bool operator < (const CSimpleTimePoint& tpOther) const;
	bool operator == (const CSimpleTimePoint& tpOther) const;

	CSimpleTimePoint& operator += (const SimpleDuration& duration);

	void SetLocalTime();

	void GetString(std::string& str) const;
};

struct CSimpleTimeFrame {
	CSimpleTime m_tStart;

	enum LimitType {
		LIMIT_BY_DURATION,
		LIMIT_BY_END_TIME
	} m_LimitType;

	CSimpleTime m_tEnd;
	SimpleDuration m_tDuration;

	CSimpleTimeFrame(const CSimpleTime& tStart,
		const CSimpleTime& tEnd);

	CSimpleTimeFrame(const CSimpleTime& tStart,
		const SimpleDuration& tDuration);

	void GetEndTime(CSimpleTime& tEnd) const;
	void GetEndPoint(	const CSimpleDate& dStart,
						CSimpleTimePoint& tpEnd ) const;

	bool GetEarliestEventAfter(
								bool bHasPrev,
								const CSimpleDate& dPrev,
								bool bHasSameDate,
								bool bHasNext,
								const CSimpleDate& dNext,
								const CSimpleTimePoint& tpStart,
								CSimpleTimePoint& tpNext,
								bool& bIsEnding ) const;

	void GetString(std::string& str) const;
};

class CSimpleTimeFramesVector : public std::vector<CSimpleTimeFrame> {
public:

	bool IsLegal() const;

	bool GetEarliestEventAfter(
								bool bHasPrev,
								const CSimpleDate& dPrev,
								bool bHasSameDate,
								bool bHasNext,
								const CSimpleDate& dNext,
								const CSimpleTimePoint& tpStart,
								CSimpleTimePoint& tpNext,
								bool& bIsEnding ) const;

	void GetString(std::string& str) const;
};

class CSiblingDatesGenerator {
private:
	virtual bool GetEarliestDateAfter(const CSimpleDate& dStartFrom,
		CSimpleDate& dEarliest, bool bInclude) const = 0;

	virtual bool GetLatestDateBefore(const CSimpleDate& dRelative,
		CSimpleDate& dLatest, bool bInclude) const = 0;

public:
	void GetSiblingDates(const CSimpleDate& dRelative,
		bool& bHasPrev, CSimpleDate& dPrev,
		bool& bHasSameDate,
		bool& bHasNext, CSimpleDate& dNext) const;
};

class CSimpleDatesVector : 
public std::vector<CSimpleDate>,
public CSiblingDatesGenerator
{
	bool GetEarliestDateAfter(const CSimpleDate& dStartFrom,
		CSimpleDate& dEarliest, bool bInclude) const;

	bool GetLatestDateBefore(const CSimpleDate& dRelative,
		CSimpleDate& dLatest, bool bInclude) const;
public:
	void GetString(std::string& str) const;
};

class CWeekdaysSelection : public CSiblingDatesGenerator
{
private:
	bool m_vbSelection[1 + DAYS_IN_WEEK];

	bool GetEarliestDateAfter(const CSimpleDate& dStartFrom,
		CSimpleDate& dEarliest, bool bInclude) const;

	bool GetLatestDateBefore(const CSimpleDate& dRelative,
		CSimpleDate& dLatest, bool bInclude) const;

public:
	CWeekdaysSelection();

	void SetNoSelection();
	void SetDaySelection(int iDayNumber, bool bSelected);

	bool IsDaySelected(int iDayNumber) const;

	void GetString(std::string& str) const;
};

class CRepeatingTimeEvent {
private:
	CWeekdaysSelection m_WeekdaysSelection;
	CSimpleDatesVector m_DatesVector;
	CSimpleTimeFramesVector m_TimeFramesVector;

public:
	void SetDaySelection(int iDayNumber, bool bSelected);
	void AddSimpleDate(const CSimpleDate& date);
	void AddTimeFrame(const CSimpleTimeFrame& tframe);

	bool GetEarliestStateChange(
								const CSimpleTimePoint& tpStart,
								CSimpleTimePoint& tpEarliest,
								bool& bIsEnding
								) const;

	CWeekdaysSelection& GetWeekdaysSelection()
	{ return m_WeekdaysSelection; }

	const CWeekdaysSelection& GetWeekdaysSelection() const
	{ return m_WeekdaysSelection; }

	CSimpleDatesVector& GetDatesVector()
	{ return m_DatesVector; }

	const CSimpleDatesVector& GetDatesVector() const
	{ return m_DatesVector; }

	CSimpleTimeFramesVector& GetTimeFrames()
	{ return m_TimeFramesVector; }

	const CSimpleTimeFramesVector& GetTimeFrames() const
	{ return m_TimeFramesVector; }

	CRepeatingTimeEvent();
	virtual ~CRepeatingTimeEvent();

	void GetString(std::string& str) const;

};

class CTimeEventsVector : public std::vector<CRepeatingTimeEvent>
{
public:
	bool GetEarliestStateChange(
								const CSimpleTimePoint& tpStart,
								CSimpleTimePoint& tpEarliest,
								bool& bIsEnding,
								CRepeatingTimeEvent& tEvent
								) const;
};

#endif // !defined(AFX_REPEATINGTIMEEVENT_H__F48F194D_BBF3_49D8_8676_FDCD0FAA60F2__INCLUDED_)
