/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: SpecificApplicationData.cpp,v 1.2 2004/01/25 17:45:33 gadibergman Exp $
 */

/************************************************************
			SpecificApplicationData.cpp
			===========================

  The persistency of the application properties.
  This object contains all application variables that 
  need to be saved between sessions, loads them from the
  registry when the application starts to run and saves them
  to the registry every time the user presses "ok" in the 
  properties sheet UI.
  
************************************************************/


// PCH includes
#include "StdAfx.h"

// Project includes
#include "SpecificApplicationData.h"

void TellSystemError(LONG lastError) {

	LPVOID lpMsgBuf;
	FormatMessage( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM | 
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		lastError /*GetLastError()*/,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &lpMsgBuf,
		0,
		NULL 
	);
	// Process any inserts in lpMsgBuf.
	// ...
	// Display the string.
	MessageBox( NULL, (LPCTSTR)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION );
	// Free the buffer.
	LocalFree( lpMsgBuf );
}

class RegistryKeyKeeper {
private:
	HKEY m_hKey;
public:
	RegistryKeyKeeper(HKEY hKey):
	m_hKey(hKey)
	{}

	~RegistryKeyKeeper() {
		RegCloseKey(m_hKey);
	}
};

bool CSpecificApplicationData::Load() {
	return UniteWithRegistryValues(false);
}

void CSpecificApplicationData::Save() {

	UniteWithRegistryValues(true);
}

bool CSpecificApplicationData::
SetRegistryValue(HKEY hRegKey, LPCTSTR valuename, const CString & value)
{
	LONG lResult = RegSetValueEx(hRegKey, valuename, 0,
		REG_SZ, (LPBYTE) (LPCTSTR) value, 
		value.GetLength());

	return lResult == ERROR_SUCCESS;
}

bool CSpecificApplicationData::
SetRegistryValue(HKEY hRegKey, LPCTSTR valuename, DWORD value)
{
	LONG lResult = RegSetValueEx(hRegKey, valuename, 0,
		REG_DWORD, (LPBYTE) &value, 
		sizeof(value));

	return lResult == ERROR_SUCCESS;
}

bool CSpecificApplicationData::GetRegistryValue(HKEY hRegKey, LPCTSTR valuename, CString &value)
{
	enum {
		NUM_BUF_CHARS = 256
	};
	TCHAR strBuf[NUM_BUF_CHARS];
	DWORD valueType, bufSize = sizeof(strBuf);

	LONG lResult = RegQueryValueEx(hRegKey, (LPCTSTR) valuename,
		NULL, &valueType, (LPBYTE) strBuf, &bufSize);

	if (lResult == ERROR_SUCCESS) {
		value = strBuf;
		return true;
	}

	//TellSystemError(lResult);

	return false;

}

bool CSpecificApplicationData::
GetRegistryValue(HKEY hRegKey, LPCTSTR valuename, DWORD & value)
{
	DWORD valueType, bufSize = sizeof(value);

	LONG lResult = RegQueryValueEx(hRegKey, (LPCTSTR) valuename,
		NULL, &valueType, (LPBYTE) &value, &bufSize);

	return lResult == ERROR_SUCCESS;
}

bool CSpecificApplicationData::
UniteWithRegistryKey(HKEY hRegKey, bool bSave, LPCTSTR valuename, CString &value)
{
	if (bSave)
		return SetRegistryValue(hRegKey, valuename, value);
	else
		return GetRegistryValue(hRegKey, valuename, value);
}

bool CSpecificApplicationData::
UniteWithRegistryKey(HKEY hRegKey, bool bSave, LPCTSTR valuename, DWORD & value)
{
	if (bSave)
		return SetRegistryValue(hRegKey, valuename, value);
	else
		return GetRegistryValue(hRegKey, valuename, value);
}

#define SOFTWARE_REG_KEY "Software\\GSoft\\MPEG Recorder"

bool CSpecificApplicationData::UniteWithRegistryValues(bool bSave)
{
	HKEY hKey;

	REGSAM samDesired = (bSave) ? KEY_WRITE : KEY_READ;

	LONG lResult = RegOpenKeyEx(HKEY_CURRENT_USER, _T(SOFTWARE_REG_KEY),
		0, samDesired, &hKey);
	if (lResult != ERROR_SUCCESS) {
		DWORD dwRes;
		lResult = RegCreateKeyEx(HKEY_CURRENT_USER, _T(SOFTWARE_REG_KEY), 0,
			_T(""), 0, samDesired, NULL, &hKey, &dwRes);
		if (lResult != ERROR_SUCCESS)
		{
			TellSystemError(lResult);
			MessageBox(NULL, _T(
				"The following key must be defined in your registry:\n"
				"HKEY_CURRENT_USER\\" SOFTWARE_REG_KEY),
				_T("Registry Information"),
				MB_OK);
			return false;
		}
		ASSERT(dwRes == REG_CREATED_NEW_KEY);
		return false;
	}
	
	RegistryKeyKeeper regKeeper(hKey);

	bool success = true;

	success &= UniteWithRegistryKey(hKey, bSave, _T("Recording Directory"), m_sRecordingDirectory);
	success &= UniteWithRegistryKey(hKey, bSave, _T("Audio Input Device"), m_sRecordingDeviceName);
	success &= UniteWithRegistryKey(hKey, bSave, _T("Video Input Device"), m_sVideoCaptureDevName);
	success &= UniteWithRegistryKey(hKey, bSave, _T("Audio Encoder Number"), m_uAudioEncoderNumber);
	success &= UniteWithRegistryKey(hKey, bSave, _T("AV Format"), m_sAVFormatName);
	success &= UniteWithRegistryKey(hKey, bSave, _T("TV Video Format Number"), m_uTVVideoFormatNumber);
	success &= UniteWithRegistryKey(hKey, bSave, _T("TV Channel Number"), m_uTVChannelNumber);
	success &= UniteWithRegistryKey(hKey, bSave, _T("Audio KBit Rate"), m_uAudioKBitRate);
	success &= UniteWithRegistryKey(hKey, bSave, _T("Video Frame Rate"), m_uVideoFrameRate);
	success &= UniteWithRegistryKey(hKey, bSave, _T("Preferred Image Size"), m_sPreferredImageSize);

	return success;
}
