/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: TimeBasedProgressSelection.cpp,v 1.1.1.1 2003/12/24 19:41:42 gadibergman Exp $
 */

/************************************************************
			TimeBasedProgressSelection.cpp
			==============================

  Synchronization solution for delaying threads according
  to the predicted timestamp of the current media buffer that
  they are waiting to write.
  
  The prediction of their timestamp = timestamp + duration
  of their previous buffer.
  
************************************************************/

#include "stdafx.h"
#include "TimeBasedProgressSelection.h"
#include "ProgramLog.h"

// Includes
#include <assert.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

struct ThreadInfo {
	HANDLE m_hProgressEvent;
	Timestamp m_Timestamp;
	bool m_bActive;

	ThreadInfo():
	m_hProgressEvent(CreateEvent(NULL, FALSE, FALSE, NULL)),
	m_Timestamp(0),
	m_bActive(true)
	{}

	void Wait() {
		DWORD dwResult =
			WaitForSingleObject(m_hProgressEvent, INFINITE);
		assert(dwResult == WAIT_OBJECT_0);
	}

	void Wait(Timestamp timestamp) {
		Wait();
		m_Timestamp = timestamp;
	}

	virtual ~ThreadInfo() {
		CloseHandle(m_hProgressEvent);
	}

private:
	// Instances of this class may not be copied
	ThreadInfo(const ThreadInfo&);
	ThreadInfo& operator = (const ThreadInfo&);
};

class ThreadsInfoVector {
private:

	ThreadInfo* m_pThreads;
	int m_iNumThreads;

	void Enable(int iIndex) {
		m_iNextEnabledIndex = iIndex;

		LOG_STREAM("Enabling stream number " << iIndex)

		BOOL bResult = SetEvent(m_pThreads[iIndex].m_hProgressEvent);
		assert(bResult);
	}

public:

	void EnableEarliest() {
		Timestamp tsMin(0); 
		int iMinIndex(0);
		bool bFound(false);
		for (int iIndex = 0; iIndex < m_iNumThreads; ++iIndex) {
			ThreadInfo& thInfo = m_pThreads[iIndex];
			if ( ( !bFound || (thInfo.m_Timestamp < tsMin) ) && thInfo.m_bActive )
			{
				tsMin = thInfo.m_Timestamp;
				iMinIndex = iIndex;
				bFound = true;
			}
		}
		if (bFound)
			Enable(iMinIndex);
	}

	int m_iNextEnabledIndex;

	ThreadsInfoVector(int numThreads):
	m_pThreads(new ThreadInfo[numThreads]),
	m_iNumThreads(numThreads)
	{
		assert(numThreads);
		Enable(0);
	}

	virtual ~ThreadsInfoVector() {
		delete[] m_pThreads;
	}

	void Wait(int iThreadIndex, Timestamp timestamp) {

		LOG_STREAM("Waiting as stream number " << iThreadIndex
			<< " with time = " << 
			(double) (__int64) m_pThreads[iThreadIndex].m_Timestamp / TimestampTicks)

		m_pThreads[iThreadIndex].Wait(timestamp);

		LOG_STREAM("Finished waiting as stream number " << iThreadIndex
			<< ". Next time = " << 
			(double) (__int64) m_pThreads[iThreadIndex].m_Timestamp / TimestampTicks)
	}

	void DeclareInactive(int iThreadIndex) {
		LOG_STREAM("Declaring stream number " << iThreadIndex << "inactive")
		m_pThreads[iThreadIndex].Wait();
		m_pThreads[iThreadIndex].m_bActive = false;
	}
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTimeBasedProgressSelection::CTimeBasedProgressSelection(int numThreads):
m_pThreadsInfo(new ThreadsInfoVector(numThreads))
{
}

CTimeBasedProgressSelection::~CTimeBasedProgressSelection()
{
	delete m_pThreadsInfo;
}

void CTimeBasedProgressSelection::Wait(int iThreadIndex, Timestamp timestamp)
{
	m_pThreadsInfo->Wait(iThreadIndex, timestamp);
}

void CTimeBasedProgressSelection::DeclareInactive(int iThreadIndex)
{
	m_pThreadsInfo->DeclareInactive(iThreadIndex);
}

void CTimeBasedProgressSelection::Release()
{
	m_pThreadsInfo->EnableEarliest();
}
