/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: VideoViewer.cpp,v 1.7 2004/04/12 15:36:49 gadibergman Exp $
 */

/************************************************************
			VideoViewer.cpp
			===============

  - UI for recording combined video+audio file
  - UI for TV and cross-bar control
  - UI for basic scheduling of a recording start/stop time
  
************************************************************/

// PCH includes
#include "stdafx.h"

// C includes
#include <stdio.h>

// DX includes
#include <vfwmsgs.h>

// Dummy Filter
#include "DummyFilter.h"

// Project includes
#include "AudioEncoderGenerator.h"
#include "AVEncoder.h"
#include "ImageCaptureDlg.h"
#include "DXSoundReceiver.h"
#include "DXVideoReceiver.h"
#include "ExternalVideoViewer.h"
#include "FileNameGenerator.h"
#include "Mp4Encoder.h"
#include "MPEGRecorder.h"
#include "TimeKeepingReaderWriter.h"
#include "VideoViewer.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------
#define SAFE_DELETE(p)  { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_RELEASE(p) { if(p) { (p)->Release(); (p)=NULL; } }

struct FilterGraphWrapper
{
	IGraphBuilder *m_pGraph;
	ICaptureGraphBuilder2 *m_pBuilder;
	bool m_bIsOK;

	FilterGraphWrapper(IGraphBuilder* pGraph, ICaptureGraphBuilder2* pBuilder):
	m_pGraph(pGraph),
	m_pBuilder(pBuilder),
	m_bIsOK(false)
	{
		if ( !m_pGraph )
		{
			// Create the filter graph.
			CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC,
					IID_IGraphBuilder, (void **)&m_pGraph);

			if ( !m_pGraph )
				return;
		}

		m_pGraph->AddRef();

		if ( !m_pBuilder )
		{
			// Create the capture graph builder.
			CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, CLSCTX_INPROC, 
					IID_ICaptureGraphBuilder2, (void **)&m_pBuilder);

			if ( !m_pBuilder )
				return;

			m_pBuilder->SetFiltergraph(m_pGraph);
		}

		m_pBuilder->AddRef();

		m_bIsOK = true;
	}

	virtual ~FilterGraphWrapper()
	{
		if ( m_pBuilder )
			m_pBuilder->Release();

		if ( m_pGraph )
			m_pGraph->Release();
	}

};

/////////////////////////////////////////////////////////////////////////////
// CVideoViewer dialog


CVideoViewer::CVideoViewer(
	IBaseFilter * pVideoIn,
	const GUID & audioRecordingDeviceID,
	CAudioEncoderGenerator & audioEncoderGenerator,
	CFileNameGenerator & fileNameGenerator,
	RecorderControlMode controlMode,
	long lAnalogVideoStandard,
	DWORD & lChannelNumberRef,
	unsigned int uDesiredFrameRate,
	const std::string & strPreferredImageSize,
	bool& bInTransition,
	CVideoViewer ** ppMe,
	CWnd* pParent):
CDialog(CVideoViewer::IDD, pParent),
m_pVideoFilter(pVideoIn),
m_AudioRecordingDeviceID(audioRecordingDeviceID),
m_AudioEncoderGenerator(audioEncoderGenerator),
m_FileNameGenerator(fileNameGenerator),
m_ControlMode(controlMode),
m_lAnalogVideoStandard(lAnalogVideoStandard),
m_lChannelNumberRef(lChannelNumberRef),
m_uDesiredFrameRate(uDesiredFrameRate),
m_strPreferredImageSize(strPreferredImageSize),
m_bInTransition(bInTransition),
m_ppMe(ppMe),
m_pVideoReceiver(NULL),
m_pSoundReceiver(NULL),
m_pMPEGEncoder(NULL),
m_pVideoProps(NULL),
m_pTVTuner(NULL),
m_pTVProps(NULL),
m_pCrossbar(NULL),
m_pCrossbarProps(NULL),
m_pTVAudio(NULL),
m_pTVAudioProps(NULL),
m_pExternalVideoViewer(NULL),
m_pFGW(NULL),
m_pDummyFilter(NULL)
{
	//{{AFX_DATA_INIT(CVideoViewer)
	m_lChannelNumber = 0;
	m_sAACBytes = _T("");
	m_sPCMBytes = _T("");
	m_sRGBBytes = _T("");
	m_sXVidBytes = _T("");
	m_sYUVBytes = _T("");
	m_sSampleRate = _T("");
	m_sEncodedBytesLabel = _T("");
	m_sTime = _T("");
	m_sFrameInformation = _T("");
	m_DialogEventTime = 0;
	m_uiDialogEventRemainingMinutes = 0;
	//}}AFX_DATA_INIT

	if ( !pVideoIn )
		return;

	m_pVideoFilter->AddRef();

	m_bInTransition = true;

	Create(IDD, pParent);

	if ( !m_bInitSuccess )
	{
		delete this;
		return;
	}

	m_PauseBitmap.LoadBitmap(IDB_BITMAP_PAUSE_BUTTON);
	m_StopBitmap.LoadBitmap(IDB_BITMAP_STOP_BUTTON);
	m_StopThenRecordBitmap.LoadBitmap(IDB_BITMAP_STOP_THEN_RECORD_BUTTON);

	m_PauseButton.SetBitmap((HBITMAP) m_PauseBitmap);
	m_StopButton.SetBitmap((HBITMAP) m_StopBitmap);
	m_StopThenRecordButton.SetBitmap((HBITMAP) m_StopThenRecordBitmap);

}

CVideoViewer::~CVideoViewer()
{
	SAFE_RELEASE( m_pTVProps )
	SAFE_RELEASE( m_pCrossbarProps )
	SAFE_RELEASE( m_pTVAudioProps )
	SAFE_RELEASE( m_pVideoProps )

	if (m_pTVTuner)
		m_pTVTuner->Logout();

	if (m_pExternalVideoViewer) {
		m_pVideoReceiver->SetPreviewWindow(GetDlgItem(IDC_VIDEO_WINDOW)->m_hWnd);
		delete m_pExternalVideoViewer;
	}

	SAFE_DELETE( m_pFGW )

	SAFE_RELEASE( m_pVideoFilter )
	SAFE_RELEASE( m_pTVTuner )
	SAFE_RELEASE( m_pCrossbar )
	SAFE_RELEASE( m_pTVAudio )

	delete m_pVideoReceiver;
	delete m_pSoundReceiver;

	SAFE_DELETE( m_pDummyFilter )

	* m_ppMe = NULL;
}

void CVideoViewer::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CVideoViewer)
	DDX_Control(pDX, IDC_VIDEO_BUTTON, m_VideoButton);
	DDX_Control(pDX, IDC_TV_AUDIO_BUTTON, m_TVAudioButton);
	DDX_Control(pDX, IDC_PAUSE_BUTTON, m_PauseButton);
	DDX_Control(pDX, IDC_EXTERNAL_WINDOW_BUTTON, m_ExternalVideoWindowButton);
	DDX_Control(pDX, IDC_CROSSBAR_BUTTON, m_CrossbarButton);
	DDX_Control(pDX, IDC_RECORD_BUTTON, m_StopThenRecordButton);
	DDX_Control(pDX, IDC_STOP_BUTTON, m_StopButton);
	DDX_Control(pDX, IDC_TUNER_MODES_COMBO, m_TunerModesCombo);
	DDX_Control(pDX, IDC_TV_BUTTON, m_TVButton);
	DDX_Control(pDX, IDC_CHANNELS_SLIDER, m_ChannelsSlider);
	DDX_Text(pDX, IDC_CHANNEL_NUMBER, m_lChannelNumber);
	DDX_Text(pDX, IDC_AAC_BYTES, m_sAACBytes);
	DDX_Text(pDX, IDC_PCM_BYTES, m_sPCMBytes);
	DDX_Text(pDX, IDC_RGB_BYTES, m_sRGBBytes);
	DDX_Text(pDX, IDC_XVID_BYTES, m_sXVidBytes);
	DDX_Text(pDX, IDC_YUV_BYTES, m_sYUVBytes);
	DDX_Text(pDX, IDC_SAMPLE_RATE, m_sSampleRate);
	DDX_Text(pDX, IDC_AUDIO_BYTES_LABEL, m_sEncodedBytesLabel);
	DDX_Text(pDX, IDC_TIME, m_sTime);
	DDX_Text(pDX, IDC_FRAME_INFO, m_sFrameInformation);
	DDX_DateTimeCtrl(pDX, IDC_EVENT_TIME, m_DialogEventTime);
	DDX_Text(pDX, IDC_EVENT_MINUTES, m_uiDialogEventRemainingMinutes);
	DDV_MinMaxUInt(pDX, m_uiDialogEventRemainingMinutes, 0, 1000);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CVideoViewer, CDialog)
	//{{AFX_MSG_MAP(CVideoViewer)
	ON_BN_CLICKED(IDC_CLOSE, OnCloseButton)
	ON_WM_DESTROY()
	ON_WM_CLOSE()
	ON_WM_HSCROLL()
	ON_BN_CLICKED(IDC_TV_BUTTON, OnTVProperties)
	ON_EN_CHANGE(IDC_CHANNEL_NUMBER, OnChangeChannelNumber)
	ON_CBN_SELCHANGE(IDC_TUNER_MODES_COMBO, OnSelchangeTunerModesCombo)
	ON_WM_TIMER()
	ON_BN_CLICKED(IDC_STOP_BUTTON, OnStop)
	ON_BN_CLICKED(IDC_RECORD_BUTTON, OnRecord)
	ON_BN_CLICKED(IDC_CROSSBAR_BUTTON, OnCrossbarProperties)
	ON_BN_CLICKED(IDC_EXTERNAL_WINDOW_BUTTON, OnPressedExternalWindowButton)
	ON_BN_CLICKED(IDC_PAUSE_BUTTON, OnPause)
	ON_BN_CLICKED(IDC_MANUAL_REC_RADIO, OnManualRecSelected)
	ON_BN_CLICKED(IDC_EVENT_TIME_RADIO, OnEventTimeSelected)
	ON_BN_CLICKED(IDC_EXT_PLANNED_REC_RADIO, OnExternalPlanSelected)
	ON_BN_CLICKED(IDC_EVENT_MINUTES_RADIO, OnEventMinutesSelected)
	ON_BN_CLICKED(IDC_CAPTURE_IMAGE_BUTTON, OnCaptureImage)
	ON_BN_CLICKED(IDC_TV_AUDIO_BUTTON, OnTVAudioProperties)
	ON_BN_CLICKED(IDC_VIDEO_BUTTON, OnVideoProperties)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CVideoViewer message handlers

void CVideoViewer::OnCloseButton() 
{
	OnClose();
}

BOOL CVideoViewer::OnInitDialog() 
{
	CDialog::OnInitDialog();

	try
	{
		// Start audio input
		m_pSoundReceiver = new CDXSoundReceiver(
			&m_AudioRecordingDeviceID,
					CHANNELS, BYTES_PER_SAMPLE, SAMPLE_RATE);

		// Start video viewer on window
		m_pVideoReceiver = new CDXVideoReceiver(
			m_pVideoFilter, 
			GetDlgItem(IDC_VIDEO_WINDOW)->m_hWnd,
			m_strPreferredImageSize
			);

		IGraphBuilder* pGraph = NULL;
		ICaptureGraphBuilder2* pBuilder = NULL;
		m_pVideoReceiver->GetGraphBuilderRef(&pGraph, &pBuilder);
		// The following line is commented because it generates a bug
		// that makes the second popping of this viewer fail...
		m_pFGW = new FilterGraphWrapper(pGraph, pBuilder);
		pGraph->Release();
		pBuilder->Release();

		//m_pFGW = new FilterGraphWrapper(NULL);
		//m_pFGW->m_pGraph->AddFilter( m_pVideoFilter, L"Video Capture" );

		//SAFE_DELETE( m_pFGW )

		EnableSpecialDeviceProperties();

		// Start video DirectX graph
		m_pVideoReceiver->Start();

		// Start sound DirectX graph
		m_pSoundReceiver->Start();

		// Pass sample rate and frame rate to window
		TCHAR strbuf[100];
		_stprintf(strbuf, _T("%d samples per second"), SAMPLE_RATE);
		m_sSampleRate = strbuf;
		
		_stprintf( strbuf, _T("%d x %d pixels per frame"),
			m_pVideoReceiver->GetWidth(),
			m_pVideoReceiver->GetHeight() );
		m_sFrameInformation = strbuf;
		
		m_sEncodedBytesLabel = 
			m_AudioEncoderGenerator.GetDefaultEncodedFileExtension();
		m_sEncodedBytesLabel += _T(" Bytes:");

		SetControlMode(m_ControlMode);

		CTime tCurr = CTime::GetCurrentTime();
		m_DialogEventTime = CTime(
			tCurr.GetYear(),
			tCurr.GetMonth(),
			tCurr.GetDay(),
			(tCurr.GetHour() + 1) % 24,
			0, 0 );

		SetTimer(TIMER_EVENT_ID,300,NULL);

		UpdateData(FALSE);

		* m_ppMe = this;
		m_bInTransition = false;
		m_bInitSuccess = true;

	}
	catch (std::string* pstrError)
	{
		AfxMessageBox(_T(pstrError->c_str()));
		delete pstrError;
		m_bInitSuccess = false;
	}

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CVideoViewer::EnableSpecialDeviceProperties()
{
	m_pVideoProps = GetPropertiesPages(m_pVideoFilter);

	if (m_pVideoProps)
		m_VideoButton.EnableWindow(TRUE);

	GetTVTuner();

	if (m_pTVTuner)
		m_pTVProps = GetPropertiesPages(m_pTVTuner);

	FindCrossbar();

	if (m_pCrossbar)
	{
		m_pCrossbarProps = GetPropertiesPages(m_pCrossbar);

		if (m_pCrossbarProps)
			m_CrossbarButton.EnableWindow();

		CComQIPtr< IBaseFilter > pFilter( m_pCrossbar );
		m_pFGW->m_pGraph->AddFilter( pFilter, L"Crossbar" );
	}

	InitTVControl();
}

void CVideoViewer::OnDestroy() 
{
	CDialog::OnDestroy();
}

void CVideoViewer::OnClose() 
{
	Stop();
	DestroyWindow();
}

void CVideoViewer::Stop()
{
	StopRecording();
	m_pVideoReceiver->Stop();
	m_pSoundReceiver->Stop();
}

class TemporaryFilter {
private:
	IBaseFilter * m_pFilter;
	IGraphBuilder * m_pBuilder;

public:
	TemporaryFilter(
		IBaseFilter * pFilter, 
		IGraphBuilder * pBuilder,
		LPCWSTR name
		):
	m_pFilter(pFilter),
	m_pBuilder(pBuilder)
	{
		pBuilder->AddFilter(pFilter, name);
	}

	virtual ~TemporaryFilter() {
		m_pBuilder->RemoveFilter(m_pFilter);
	}

};

void CVideoViewer::GetTVTuner()
{
	m_pVideoFilter->QueryInterface(__uuidof(IAMTVTuner), (void**) &m_pTVTuner);

	if (m_pTVTuner)
		return;

	m_pFGW->m_pBuilder->FindInterface(&PIN_CATEGORY_CAPTURE,
		&MEDIATYPE_Interleaved, m_pVideoFilter,
  		IID_IAMTVTuner, (void **)&m_pTVTuner);

	if (m_pTVTuner)
		return;

	m_pFGW->m_pBuilder->FindInterface(&PIN_CATEGORY_CAPTURE,
		&MEDIATYPE_Video, m_pVideoFilter,
		IID_IAMTVTuner, (void **)&m_pTVTuner);
}

void CVideoViewer::InitTVControl()
{
	if (!m_pTVTuner) 
	{
		m_ChannelsSlider.EnableWindow(FALSE);
		SendDlgItemMessage(IDC_CHANNEL_NUMBER, WM_ENABLE, 0, 0);
		m_TunerModesCombo.EnableWindow(FALSE);
		return;
	}

	if (m_pTVProps)
		m_TVButton.EnableWindow();

	m_pTVTuner->Logon(HKEY_CURRENT_USER);

	long lChannelMin, lChannelMax;
	m_pTVTuner->ChannelMinMax(&lChannelMin, &lChannelMax);
	m_ChannelsSlider.SetRange(lChannelMin, lChannelMax);

	if (m_lChannelNumberRef >= lChannelMin &&
		m_lChannelNumberRef <= lChannelMax) {
		m_lChannelNumber = m_lChannelNumberRef;
		SetTVChannel();
	}
	else {
		long lVideoChannel, lAudioChannel;
		m_pTVTuner->get_Channel(&m_lChannelNumber, &lVideoChannel, &lAudioChannel);
	}

	UpdateData(FALSE);
	m_ChannelsSlider.SetPos(m_lChannelNumber);

	LookForAudio();

	if (m_pTVAudioProps)
		m_TVAudioButton.EnableWindow();

	SetAnalogVideoFormat(m_pVideoFilter, m_lAnalogVideoStandard);

	EnumerateTunerModes();

}

void CVideoViewer::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{

	if (m_pTVTuner) {
		m_lChannelNumber = m_ChannelsSlider.GetPos();
		m_lChannelNumberRef = m_lChannelNumber;
		SetTVChannel();
		UpdateData(FALSE);
	}
	
	CDialog::OnHScroll(nSBCode, nPos, pScrollBar);
}

ISpecifyPropertyPages * CVideoViewer::GetPropertiesPages(IUnknown * pObject)
{
	ISpecifyPropertyPages *pSpec = NULL;
    HRESULT hr = pObject->QueryInterface(IID_ISpecifyPropertyPages, (void **)&pSpec);

	return pSpec;
}

void CVideoViewer::ShowProperties(HWND hWnd, IUnknown * pObject, ISpecifyPropertyPages *pSpec)
{
  CAUUID cauuid;
  HRESULT hr = pSpec->GetPages(&cauuid);
  if (hr == S_OK && cauuid.cElems > 0) {

    hr = pSpec->GetPages(&cauuid);
    hr = OleCreatePropertyFrame(hWnd, 30, 30, NULL, 1,
          &pObject, cauuid.cElems,
					(GUID *)cauuid.pElems, 0, 0, NULL);

		CoTaskMemFree(cauuid.pElems);
	}
}

void CVideoViewer::OnTVProperties() 
{
	if (!m_pTVProps)
		return;

	ShowProperties(m_hWnd, m_pTVTuner, m_pTVProps);
}

void CVideoViewer::OnChangeChannelNumber() 
{
	if (!m_pTVTuner)
		return;

	UpdateData(TRUE);
	
	long lChannelMin, lChannelMax;
	m_pTVTuner->ChannelMinMax(&lChannelMin, &lChannelMax);

	if (m_lChannelNumber >= lChannelMin &&
		m_lChannelNumber <= lChannelMax) {
		m_ChannelsSlider.SetPos(m_lChannelNumber);
		m_lChannelNumberRef = m_lChannelNumber;
		SetTVChannel();
	}
}

void CVideoViewer::SetTVChannel()
{
	long lFoundSignal;
	m_pTVTuner->AutoTune(m_lChannelNumber, &lFoundSignal);
	m_pTVTuner->StoreAutoTune();
}

void CVideoViewer::EnumerateTunerModes()
{
	if (!m_pTVTuner)
		return;

	static const char *modes_str[] = {
		"TV",
		"FM Radio",
		"AM Radio",
		"DSS"
	};

	long lModes = 0;
	m_pTVTuner->GetAvailableModes(&lModes);

	int numModes = sizeof(modes_str)/sizeof(const char*);
	for(int i = 0, flag = 1; i < numModes; ++i, flag << 1) {
		if (lModes & flag) {
			int index = m_TunerModesCombo.AddString(modes_str[i]);
			m_TunerModesCombo.SetItemData(index, flag);
		}
	}
}

void CVideoViewer::OnSelchangeTunerModesCombo() 
{
	if (!m_pTVTuner)
		return;

	m_pTVTuner->put_Mode((AMTunerModeType)
		m_TunerModesCombo.GetItemData(
			m_TunerModesCombo.GetCurSel()));
}

void HugeToReadableString(LONGLONG num, CString & sNum) {

	sNum = _T("");
	CString sSign = _T("");
	if (num < 0) {
		num = -num;
		sSign = _T("-");
	}

	do {

		LONGLONG dividedNum = num / 1000;

		TCHAR * sFormat = (dividedNum) ? _T("%03d") : _T("%d");
	
		TCHAR tempBuf[50];
		_stprintf(tempBuf, sFormat, num % 1000);
		
		num = dividedNum;

		CString sTemp;

		if (num > 0)
			sTemp = _T(",");
		else
			sTemp = _T("");

		sTemp += tempBuf;
		sTemp += sNum;

		sNum = sTemp;

	} while (num > 0);

	sNum = sSign + sNum;
}


void CVideoViewer::OnTimer(UINT nIDEvent) 
{
	double
		time(0);
	LONGLONG
		pcmBytes(0),
		aacBytes(0),
		rgbBytes(0),
		yuvBytes(0),
		xvidBytes(0);
	{
		MutexProtection prot(m_EncoderPtrMutex);
		if (m_pMPEGEncoder)
			m_pMPEGEncoder->GetProgress
				(
					time,
					pcmBytes, 
					aacBytes, 
					rgbBytes, 
					yuvBytes, 
					xvidBytes
				);
	}

	TCHAR sTime[50];
	_stprintf(sTime, "%02d:%02d", (int) time / 60,
		(int) time % 60);

	UpdateData(TRUE);

	m_sTime = sTime;

	HugeToReadableString(pcmBytes, m_sPCMBytes);
	HugeToReadableString(aacBytes, m_sAACBytes);
	HugeToReadableString(rgbBytes, m_sRGBBytes);
	HugeToReadableString(yuvBytes, m_sYUVBytes);
	HugeToReadableString(xvidBytes, m_sXVidBytes);

	UpdateData(FALSE);
	
	CDialog::OnTimer(nIDEvent);
}

void CVideoViewer::OnStop() 
{
	StopRecording();
}

void CVideoViewer::OnRecord() 
{
	StartRecording();
}

HRESULT FindCrossbarPin(
    IAMCrossbar *pXBar,                 // Pointer to the crossbar.
    PhysicalConnectorType PhysicalType, // Pin type to match.
    PIN_DIRECTION Dir,                  // Pin direction.
    long *pIndex)       // Receives the index of the pin, if found.
{
    BOOL bInput = (Dir == PINDIR_INPUT ? TRUE : FALSE);

    // Find out how many pins the crossbar has.
    long cOut, cIn;
    HRESULT hr = pXBar->get_PinCounts(&cOut, &cIn);
    if (FAILED(hr)) return hr;
    // Enumerate pins and look for a matching pin.
    long count = (bInput ? cIn : cOut);
    for (long i = 0; i < count; i++)
    {
        long iRelated = 0;
        long ThisPhysicalType = 0;
        hr = pXBar->get_CrossbarPinInfo(bInput, i, &iRelated,
            &ThisPhysicalType);
        if (SUCCEEDED(hr) && ThisPhysicalType == PhysicalType)
        {
            // Found a match, return the index.
            *pIndex = i;
            return S_OK;
        }
    }
    // Did not find a matching pin.
    return E_FAIL;
}

HRESULT ConnectAudio(IAMCrossbar *pXBar, BOOL bActivate)
{
    // Look for the Audio Decoder output pin.
    long i = 0;
    HRESULT hr = FindCrossbarPin(pXBar, PhysConn_Audio_AudioDecoder,
        PINDIR_OUTPUT, &i);
    if (SUCCEEDED(hr))
    {
        if (bActivate)  // Activate the audio. 
        {
            // Look for the Audio Tuner input pin.
            long j = 0;
            hr = FindCrossbarPin(pXBar, PhysConn_Audio_Tuner, 
                PINDIR_INPUT, &j);
            if (SUCCEEDED(hr))
            {
                return pXBar->Route(i, j);
            }
        }
        else  // Mute the audio
        {
            return pXBar->Route(i, -1);
        }
    }
    return E_FAIL;
}

HRESULT ActivateAudio(ICaptureGraphBuilder2 *pBuild, IBaseFilter *pSrc,
  BOOL bActivate)
{
    // Search upstream for a crossbar.
    IAMCrossbar *pXBar1 = NULL;
    HRESULT hr = pBuild->FindInterface(&LOOK_UPSTREAM_ONLY, NULL, pSrc,
        IID_IAMCrossbar, (void**)&pXBar1);
    if (SUCCEEDED(hr)) 
    {
        hr = ConnectAudio(pXBar1, bActivate);
        if (FAILED(hr))
        {
            // Look for another crossbar.
            IBaseFilter *pF = NULL;
            hr = pXBar1->QueryInterface(IID_IBaseFilter, (void**)&pF);
            if (SUCCEEDED(hr)) 
            {
                // Search upstream for another one.
                IAMCrossbar *pXBar2 = NULL;
                hr = pBuild->FindInterface(&LOOK_UPSTREAM_ONLY, NULL, pF,
                    IID_IAMCrossbar, (void**)&pXBar2);
                pF->Release();
                if (SUCCEEDED(hr))
                {
                    hr = ConnectAudio(pXBar2, bActivate);
                    pXBar2->Release();
                }
            }
        }
        pXBar1->Release();
    }
    return hr;
}

void CVideoViewer::LookForAudio()
{
	if (!m_pTVTuner)
		return;

	CComQIPtr< IBaseFilter > pTunerFilter( m_pTVTuner );

	ASSERT(pTunerFilter);

	m_pFGW->m_pGraph->AddFilter(pTunerFilter, L"TV Tuner");

	m_pFGW->m_pBuilder->FindInterface(&LOOK_DOWNSTREAM_ONLY,
		&MEDIATYPE_Audio, pTunerFilter,
  		IID_IAMTVAudio, (void **)&m_pTVAudio);

	if ( m_pTVAudio )
		m_pTVAudioProps = GetPropertiesPages(m_pTVAudio);

	if ( m_pCrossbar )
	{
		//HRESULT hRes = ConnectAudio( m_pCrossbar, TRUE );
		HRESULT hRes = ActivateAudio( m_pFGW->m_pBuilder, 
			m_pVideoFilter, TRUE );

// The following line causes a weird crash on window-close
#if 0
		RenderUnconnectedCrossbarOutputPins();
#endif
	}
}

void CVideoViewer::SetAnalogVideoFormat(
		IBaseFilter *pVideoFilter, 
		long lAnalogVideoStandard
		) 
{
	CComQIPtr< IAMAnalogVideoDecoder > pVideoDecoder( pVideoFilter );

	if (!pVideoDecoder)
		return;

	pVideoDecoder->put_TVFormat(lAnalogVideoStandard);
}

void CVideoViewer::FindCrossbar()
{
	m_pVideoFilter->QueryInterface(__uuidof(IAMCrossbar), (void**) &m_pCrossbar);

	if (m_pCrossbar)
		return;

	m_pFGW->m_pBuilder->FindInterface(&PIN_CATEGORY_CAPTURE,
		&MEDIATYPE_Interleaved, m_pVideoFilter,
  		IID_IAMCrossbar, (void **)&m_pCrossbar);

	if (m_pCrossbar)
		return;

	m_pFGW->m_pBuilder->FindInterface(&PIN_CATEGORY_CAPTURE,
		&MEDIATYPE_Video, m_pVideoFilter,
		IID_IAMCrossbar, (void **)&m_pCrossbar);
}

void CVideoViewer::OnCrossbarProperties() 
{
	if (!m_pCrossbarProps)
		return;

	ShowProperties(m_hWnd, m_pCrossbar, m_pCrossbarProps);
}

CMPEGEncoder* CVideoViewer::CreateMPEGEncoder()
{
	if (0 == _tcsicmp(m_FileNameGenerator.GetFileExtension(), _T(".mp4")))
		return new CMp4Encoder(
			m_pSoundReceiver,
			m_pVideoReceiver,
			m_AudioEncoderGenerator,
			m_FileNameGenerator,
			m_uDesiredFrameRate
			);

	return new CAVEncoder(
			m_pSoundReceiver,
			m_pVideoReceiver,
			m_AudioEncoderGenerator,
			m_FileNameGenerator,
			m_uDesiredFrameRate
			);
}

void CVideoViewer::RestoreInternalPreview()
{
	m_ExternalVideoWindowButton.SetWindowText(_T("External Window"));
	m_pVideoReceiver->SetPreviewWindow(GetDlgItem(IDC_VIDEO_WINDOW)->m_hWnd);
	delete m_pExternalVideoViewer;
	m_pExternalVideoViewer = NULL;
}

void CVideoViewer::OnPressedExternalWindowButton()
{
	if (m_pExternalVideoViewer) {
		RestoreInternalPreview();
		return;
	}

	m_pExternalVideoViewer = 
		new CExternalVideoViewer(this);

	m_ExternalVideoWindowButton.SetWindowText(_T("Internal View"));
}

void CVideoViewer::OnPause() 
{
	if (m_pMPEGEncoder)
		m_pMPEGEncoder->PlayOrPause();
}

BOOL CVideoViewer::OnCommand(WPARAM wParam, LPARAM lParam) 
{
	if (wParam == 2) // Pressed Esc
	{
		Stop();
		DestroyWindow();
		return TRUE;
	}
	
	return CDialog::OnCommand(wParam, lParam);
}

void CVideoViewer::ExternalStartRecording()
{
	if (m_ControlMode == EXTERNAL_CONTROL)
		StartRecording();
}

void CVideoViewer::ManualStartRecording()
{
	if (m_ControlMode == MANUAL_CONTROL)
		StartRecording();
}

void CVideoViewer::ExternalStopRecording()
{
	if (m_ControlMode == EXTERNAL_CONTROL)
		StopRecording();
}

void CVideoViewer::OnManualRecSelected() 
{
	SetControlMode(MANUAL_CONTROL);
}

void CVideoViewer::OnExternalPlanSelected() 
{
	SetControlMode(EXTERNAL_CONTROL);
}

void CVideoViewer::OnEventMinutesSelected() 
{
	SetControlMode(DURATION_CONTROL);
}

void CVideoViewer::OnEventTimeSelected() 
{
	SetControlMode(TIME_CONTROL);	
}

struct ControlRadioButtonInfo {
	CVideoViewer::RecorderControlMode m_Mode;
	int m_iButtonID;
};

static ControlRadioButtonInfo vControlButtons[] = {
	{ CVideoViewer::MANUAL_CONTROL, IDC_MANUAL_REC_RADIO },
	{ CVideoViewer::EXTERNAL_CONTROL, IDC_EXT_PLANNED_REC_RADIO },
	{ CVideoViewer::DURATION_CONTROL, IDC_EVENT_MINUTES_RADIO },
	{ CVideoViewer::TIME_CONTROL, IDC_EVENT_TIME_RADIO }
};

static int vManualControlButtonIDs[] = {
	IDC_RECORD_BUTTON,
	IDC_STOP_BUTTON,
	IDC_PAUSE_BUTTON,
	IDC_CLOSE
};

void CVideoViewer::SetControlMode(CVideoViewer::RecorderControlMode mode)
{
	int i;

	// Manual Radio-Button manipulation 
	// required because the "group" attribute is enabled for all these radio buttons
	// and that's because we want to get click events from every selection change
	int iNumControlButtons = sizeof(vControlButtons) / sizeof(ControlRadioButtonInfo);
	for (i = 0; i < iNumControlButtons; ++i)
	{
		bool bIsSelected = (vControlButtons[i].m_Mode == mode);
		SendDlgItemMessage(vControlButtons[i].m_iButtonID,
			BM_SETCHECK, (WPARAM) bIsSelected, 0);
	}

	bool bManualEnabled = (mode == MANUAL_CONTROL);

	int iNumManualControlButtons = sizeof(vManualControlButtonIDs) / sizeof(int);
	for (i = 0; i < iNumManualControlButtons; ++i)
	{
		GetDlgItem(vManualControlButtonIDs[i])->EnableWindow(bManualEnabled);
	}

	m_ControlMode = mode;
}

void CVideoViewer::OnMinutePassed()
{
	UpdateData(TRUE);

	switch (m_ControlMode)
	{
	case DURATION_CONTROL:
		{
			if (m_uiDialogEventRemainingMinutes)
			{
				--m_uiDialogEventRemainingMinutes;
				if (0 == m_uiDialogEventRemainingMinutes)
					PerformDefaultRecordingEvent();
			}
		}
		break;
	case TIME_CONTROL:
		{
			CTime curTime = CTime::GetCurrentTime();
			bool bSameTime = 
				(curTime.GetHour() == m_DialogEventTime.GetHour()) &&
				(curTime.GetMinute() == m_DialogEventTime.GetMinute());

			if (bSameTime)
				PerformDefaultRecordingEvent();
		}
		break;
	}

	UpdateData(FALSE);
}

void CVideoViewer::PerformDefaultRecordingEvent()
{
	if (m_pMPEGEncoder)
		StopRecording();
	else
		StartRecording();

	SetControlMode(MANUAL_CONTROL);
}

void CVideoViewer::StartRecording()
{
	CMPEGEncoder * pLastEncoder = m_pMPEGEncoder;

	MutexProtection prot(m_EncoderPtrMutex);

	m_pMPEGEncoder = CreateMPEGEncoder();

	if (pLastEncoder)
	{
		pLastEncoder->Stop();
		delete pLastEncoder;
	}
}

void CVideoViewer::StopRecording()
{
	MutexProtection prot(m_EncoderPtrMutex);

	if (m_pMPEGEncoder)
	{
		CMPEGEncoder* pLastEncoder = m_pMPEGEncoder;
		m_pMPEGEncoder = NULL;
		pLastEncoder->Stop();
		delete pLastEncoder;
	}
}

void CVideoViewer::OnCaptureImage() 
{
	CImageCaptureDlg icapDialog(this);
	icapDialog.CaptureImage(m_pVideoReceiver);
}

void CVideoViewer::RenderUnconnectedCrossbarOutputPins()
{
	CComQIPtr< IBaseFilter > pFilter( m_pCrossbar );

	if ( !pFilter )
		return;

    CComPtr< IEnumPins > pEnum;
    HRESULT hr = pFilter->EnumPins(&pEnum);
    if(FAILED(hr)) 
        return;

    ULONG ulFound;
    IPin *pPin;
    hr = E_FAIL;
    while(S_OK == pEnum->Next(1, &pPin, &ulFound))
    {
        PIN_DIRECTION pindir = (PIN_DIRECTION)3;
        pPin->QueryDirection(&pindir);
        if(pindir == PINDIR_OUTPUT)
        {
			IPin * pPinIn = NULL;
			HRESULT hRes = pPin->ConnectedTo(&pPinIn);
			if ( pPinIn )
			{
				pPinIn->Release();
			}
			else
			{
				PIN_INFO info;
				pPin->QueryPinInfo(&info);

				if ( info.pFilter )
					info.pFilter->Release();

			    if ( wcsstr( info.achName, L"Audio" ) || wcsstr( info.achName, L"audio" ) )
				{
					//HRESULT hRes = fgw.m_pGraph->Render(pPin);
					m_pDummyFilter = new CDummySinkFilter(pPin);
					HRESULT hRes = m_pFGW->m_pGraph->AddFilter( m_pDummyFilter, L"Dummy Glove Sink Filter" );
					ATLASSERT(hRes == S_OK);
					//hRes = fgw.m_pGraph->AddFilter( pFilter, L"Crossbar" );
					//ATLASSERT(hRes == S_OK);
					hRes = m_pFGW->m_pGraph->Connect( pPin, m_pDummyFilter->GetInputPin() );
					int iRes = 0;
					switch (hRes)
					{
					case VFW_S_PARTIAL_RENDER:
						iRes = 1;
						break;
					case E_ABORT:
						iRes = 2;
						break;
					case E_POINTER:
						iRes = 3;
						break;
					case VFW_E_CANNOT_CONNECT:
						iRes = 4;
						break;
					case VFW_E_NOT_IN_GRAPH:
						iRes = 5;
						break;
					default:
						iRes = 6;
						break;
					}
					//ATLASSERT(hRes == S_OK);
				}

			}
        } // if
        pPin->Release();
    } // while
}


void CVideoViewer::OnTVAudioProperties() 
{
	if ( !m_pTVAudioProps )
		return;

	ShowProperties(m_hWnd, m_pTVAudio, m_pTVAudioProps);
}

void CVideoViewer::OnVideoProperties() 
{
	if ( !m_pVideoProps )
		return;

	ShowProperties(m_hWnd, m_pVideoFilter, m_pVideoProps);
}

void CVideoViewer::PostNcDestroy() 
{
	CDialog::PostNcDestroy();

	delete this;
}
