/*
 * MPEG Recorder
 * Copyright (c) 2003 Gadi Bergman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: XVidEncoder.cpp,v 1.2 2004/03/27 13:47:16 gadibergman Exp $
 */

/************************************************************
			XVidEncoder.cpp
			===============

  An MPEG-4 Video encoder, using code from MPEG4IP project.
  It reads YUV frames and writes XVid data.
  
************************************************************/

// PCH include
#include "stdafx.h"

// Project includes
#include "MPEGRecorder.h"
#include "XVidEncoder.h"
#include "TimeKeepingReaderWriter.h"
#include "Common/MutexProtection.h"
#include <process.h>

// MPEG 4 includes
#define MPEG4IP
#include "mpeg4/xvid/divx4.h"

#include <string>

using namespace std;

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CXVidEncoder::
CXVidEncoder(TimeKeepingReader * pReader, TimeKeepingWriter * pWriter,
			 unsigned int uFrameWidth, unsigned int uFrameHeight,
			 double dFramesPerSecond):
m_pReader(pReader),
m_pWriter(pWriter),
m_uFrameWidth(uFrameWidth),
m_uFrameHeight(uFrameHeight),
m_hThreadCompletion(CreateEvent(NULL, FALSE, TRUE, NULL)),
m_uCurrentFrameNumber(0),
m_dFramesPerSecond(dFramesPerSecond)
{
	Init();
}

CXVidEncoder::~CXVidEncoder()
{
	DWORD dwResult = WaitForSingleObject(m_hThreadCompletion, INFINITE);
	ASSERT(dwResult == WAIT_OBJECT_0);

	CloseHandle(m_hThreadCompletion);
}

static void CXVidEncoder_EncodingThread(LPVOID pParam) {
	CXVidEncoder * pEncoder = (CXVidEncoder *) pParam;
	ThreadCompletionAnnouncement announce(pEncoder->m_hThreadCompletion);

	try {
		pEncoder->Encode();
	}
	catch (string * errMess) {
		AfxMessageBox(errMess->c_str());
		delete errMess;
	}
}

void CXVidEncoder::Init()
{
	_beginthread(CXVidEncoder_EncodingThread, 0, this);
}

void CXVidEncoder::Encode()
{
	/* variables controlable from command line */
	//enum {
	//	BIT_RATE = 500000
	//};

	ASSERT((m_dFramesPerSecond < 100) && (m_dFramesPerSecond > 5));

	int short_headers = 0;                  /* --shortheaders */

	/* internal variables */
	void* myHandle = NULL;
	ENC_PARAM encParams;
	ENC_FRAME encFrame;
	ENC_RESULT encResult;
	unsigned long yuvSize;
	unsigned char encVopBuffer[128 * 1024];
	time_t startTime;

	/* initialize encoder */
	encParams.x_dim = m_uFrameWidth;
	encParams.y_dim = m_uFrameHeight;
	encParams.framerate = m_dFramesPerSecond;
	encParams.bitrate = m_uFrameWidth * m_uFrameHeight * 5; //BIT_RATE;
	encParams.rc_period = 2000;
	encParams.rc_reaction_period = 10;
	encParams.rc_reaction_ratio = 20;
	encParams.max_key_interval = (int) m_dFramesPerSecond;
	encParams.use_bidirect = 0;
	encParams.deinterlace = 0;
	encParams.quality = 5;
	encParams.obmc = 0;
	encParams.handle = NULL;

	if (encore(myHandle, ENC_OPT_INIT, &encParams, NULL) != ENC_OK)
		throw new string("Failed to initialize XVID Encoder");

	myHandle = encParams.handle;

	yuvSize = (m_uFrameWidth * m_uFrameHeight * 3) / 2;
	encFrame.image = malloc(yuvSize);
	encFrame.bitstream = encVopBuffer;
	encFrame.colorspace = ENC_CSP_I420;
	encFrame.quant = 0;
	encFrame.intra = -1;
	encFrame.mvs = NULL;
	encFrame.general = 0;
	if (short_headers) {
	  encFrame.general |= DEC_SHORT_HEADERS;
	}
	startTime = time(0);

	while(true)
	{
		Timestamp timestamp;
		Duration duration;

		/* read yuv data */
		int rc = m_pReader->ReadBytes(encFrame.image, yuvSize, timestamp, duration);

		if (!rc)
			break;

		long lNumExpectedFrames = 
			m_dFramesPerSecond * (double) (int64_t) ( timestamp + duration ) / 
			TimestampTicks;

		duration = TimestampTicks / m_dFramesPerSecond;

		for (; m_uCurrentFrameNumber < lNumExpectedFrames; ++m_uCurrentFrameNumber)
		{
	
			encFrame.length = sizeof(encVopBuffer);

			/* call xvid encoder */
			rc = encore(myHandle, ENC_OPT_ENCODE, &encFrame, &encResult);

			if (rc != ENC_OK) {
				if (rc == ENC_MEMORY)
					throw new string("XVid Memory Error");
				else if (rc == ENC_BAD_FORMAT)
					throw new string("XVid Bad Format Error");
				else 
					throw new string("Undefined XVid Error");
			}

			/* write results to output file */
			m_pWriter->WriteBytes(encFrame.bitstream, 
				encFrame.length, timestamp, duration);

			timestamp += duration;
		}
	}

	/* cleanup */
	encore(myHandle, ENC_OPT_RELEASE, NULL, NULL);

	m_pWriter->DeclareWritingEnded();
}
