/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is MPEG4IP.
 * 
 * The Initial Developer of the Original Code is Cisco Systems Inc.
 * Portions created by Cisco Systems Inc. are
 * Copyright (C) Cisco Systems Inc. 2000, 2001.  All Rights Reserved.
 * 
 * Contributor(s): 
 *		Dave Mackie		dmackie@cisco.com
 *		Bill May 		wmay@cisco.com
 */

#include "StdAfx.h"

#include "../../AudioEncoder.h"

//#include "mp4live.h"
#include "file_mp4_recorder.h"
//#include "video_v4l_source.h"
//#include "audio_encoder.h"
#include "../../TimeKeepingReaderWriter.h"
#include "../../TimeBasedProgressSelection.h"

enum {
  RTP_PAYLOAD_SIZE = 1460,
  VIDEO_PROFILE_ID = 1
};


void error_message(const char * mess) {
	MessageBox(NULL, mess, _T("Error"), MB_OK);
}

void create_mp4_video_hint_track (
				  MP4FileHandle mp4file,
				  MP4TrackId trackId);


void create_mp4_audio_hint_track (
				  MP4FileHandle mp4file,
				  MP4TrackId trackId)
{
    MP4AV_RfcIsmaHinter(mp4file, 
			trackId, 
			false, 
			RTP_PAYLOAD_SIZE);
}

int CMp4Recorder::ThreadMain(void) 
{
	while (SDL_SemWait(m_myMsgQueueSemaphore) == 0) {
		CMsg* pMsg = m_myMsgQueue.get_message();
		
		if (pMsg != NULL) {
			switch (pMsg->get_value()) {
			case MSG_NODE_STOP_THREAD:
				DoStopRecord();
				delete pMsg;
				return 0;
			case MSG_NODE_START:
				DoStartRecord();
				break;
			case MSG_NODE_STOP:
				DoStopRecord();
				break;
			case MSG_SINK_FRAME:
				uint32_t dontcare;
				DoWriteFrame((CMediaFrame*)pMsg->get_message(dontcare));
				break;
			}

			delete pMsg;
		}
	}

	return -1;
}

void CMp4Recorder::DoStartRecord()
{
	// already recording
	if (m_sink) {
		return;
	}

	m_tsRecent = 0;

	m_makeIod = true;
	m_makeIsmaCompliant = true;
	m_rawVideoTrackId = MP4_INVALID_TRACK_ID;
	m_encodedVideoTrackId = MP4_INVALID_TRACK_ID;
	m_rawAudioTrackId = MP4_INVALID_TRACK_ID;
	m_encodedAudioTrackId = MP4_INVALID_TRACK_ID;

	m_audioTimeScale = m_uAudioSampleRate;

	// are we recording any video? yes
	m_recordVideo = true;
	m_canRecordAudio = false;	// audio must wait for video
	m_movieTimeScale = m_videoTimeScale;

	bool isma_compliant = false;
	bool createIod = false;

 	// get the mp4 file setup
	bool hugeFile = true;

 	// enable huge file mode in mp4 
 	// if duration is very long or if estimated size goes over 1 GB
	/*
 	u_int64_t duration = m_pConfig->GetIntegerValue(CONFIG_APP_DURATION) 
 		* m_pConfig->GetIntegerValue(CONFIG_APP_DURATION_UNITS)
 		* m_movieTimeScale;
 	bool hugeFile = 
 		(duration > 0xFFFFFFFF) 
 		|| (m_pConfig->m_recordEstFileSize > 1000000000LLU);
	*/
 
	//debug_message("Creating huge file - %s", hugeFile ? "yes" : "no");
 	u_int32_t verbosity =
 		MP4_DETAILS_ERROR /* DEBUG | MP4_DETAILS_WRITE_ALL */;
 
 	m_mp4File = MP4Create((LPCTSTR) m_sFileName, verbosity, hugeFile);
 
 	if (!m_mp4File) {
 		return;
	}
	MP4SetTimeScale(m_mp4File, m_movieTimeScale);

	m_encodedVideoFrameNumber = 1;
	bool vIod, vIsma;
	uint8_t videoProfile;
	uint8_t *videoConfig;
	uint32_t videoConfigLen;
	uint8_t videoType;

	m_encodedVideoFrameType = 
	  get_video_mp4_fileinfo(
				 &vIod,
				 &vIsma,
				 &videoProfile,
				 &videoConfig,
				 &videoConfigLen,
				 &videoType);
				 
	m_encodedVideoTrackId = MP4AddVideoTrack(
		m_mp4File,
		m_videoTimeScale,
		MP4_INVALID_DURATION,
		m_uFrameWidth, 
		m_uFrameHeight,
		videoType);

	if (vIod == false) m_makeIod = false;
	if (vIsma == false) m_makeIsmaCompliant = false;

	if (m_encodedVideoTrackId == MP4_INVALID_TRACK_ID) {
		error_message("can't create encoded video track");
		goto start_failure;
	}

	MP4SetVideoProfileLevel(
		m_mp4File, 
		videoProfile);

	MP4SetTrackESConfiguration(
		m_mp4File, 
		m_encodedVideoTrackId,
		videoConfig,
		videoConfigLen);

	m_rawAudioFrameNumber = 1;
	m_rawAudioDuration = 0;
	m_encodedAudioFrameNumber = 1;
	m_encodedAudioDuration = 0;

	u_int8_t audioType;
	uint8_t audioProfile;
	uint8_t *pAudioConfig;
	uint32_t audioConfigLen;
	ASSERT(m_pAudioEncoder);
	m_encodedAudioFrameType = 
		m_pAudioEncoder->get_audio_mp4_fileinfo(
				 &createIod,
				 &isma_compliant,
				 &audioProfile,
				 &pAudioConfig,
				 &audioConfigLen,
				 &audioType);
				   
	if (createIod == false) m_makeIod = false;
	if (isma_compliant == false) 
	  m_makeIsmaCompliant = false;
	MP4SetAudioProfileLevel(m_mp4File, audioProfile);
	m_encodedAudioTrackId = MP4AddAudioTrack(
		m_mp4File, 
		m_audioTimeScale, 
		MP4_INVALID_DURATION,
		audioType);

	if (m_encodedAudioTrackId == MP4_INVALID_TRACK_ID) {
		error_message("can't create encoded audio track");
		goto start_failure;
	}

	if (pAudioConfig) {
	  MP4SetTrackESConfiguration(
					 m_mp4File, 
					 m_encodedAudioTrackId,
					 pAudioConfig, 
					 audioConfigLen);
	}

	m_sink = true;
	return;

start_failure:
	MP4Close(m_mp4File);
	m_mp4File = NULL;
	return;
}

void CMp4Recorder::DoWriteFrame(CMediaFrame* pFrame)
{
	// dispose of degenerate cases
	if (pFrame == NULL) {
		return;
	}

	if (!m_sink) {
	  if (pFrame->RemoveReference()) {
		delete pFrame;
	  }
	  return;
	}

	m_tsRecent = pFrame->GetTimestamp() + pFrame->GetDuration();

	// check if this is an audio frame that we want to record
	// and if so setup some local variables

	bool doRawAudioFrame = false;
	bool doEncodedAudioFrame = false;
	MP4TrackId audioTrackId = MP4_INVALID_TRACK_ID;
	u_int32_t audioFrameNumber = 0;
	Duration audioDuration = 0;

	if (pFrame->GetType() == m_encodedAudioFrameType) {
		doEncodedAudioFrame = true;
		audioTrackId = m_encodedAudioTrackId;
		audioFrameNumber = m_encodedAudioFrameNumber;
		audioDuration = m_encodedAudioDuration;
	}

	// process an audio frame
	if ((doRawAudioFrame || doEncodedAudioFrame)) {
		// need special processing for very first audio frame
		if (audioFrameNumber == 1) {

			// can't record yet, awaiting first video frame
			if (!m_canRecordAudio) {
			  if (pFrame->RemoveReference()) {
				delete pFrame;
			  }
			  return;
			}

			// initialize variables for audio timeline
			if (doRawAudioFrame) {
				m_rawAudioStartTimestamp = pFrame->GetTimestamp();
			} else {
				m_encodedAudioStartTimestamp = pFrame->GetTimestamp();
			}

			// if just recording audio
			if (!m_recordVideo) {
				if (m_rawAudioFrameNumber == 1 
				  && m_encodedAudioFrameNumber == 1) {
					m_movieStartTimestamp = pFrame->GetTimestamp();
				}
			} else {
				// drop any errant audio frames that are too early
				if (pFrame->GetTimestamp() < m_movieStartTimestamp) {
				  if (pFrame->RemoveReference())
					delete pFrame;
				  return;
				}
			}

		} // end of first audio frame processing

		Duration audioGapInTicks = 
			(pFrame->GetTimestamp() - m_movieStartTimestamp) - audioDuration;

		MP4Duration audioGapInSamples = 0;

		if (audioGapInTicks > 0) {
			audioGapInSamples =
				MP4ConvertToTrackDuration(
					m_mp4File, 
					audioTrackId,
					audioGapInTicks,
					TimestampTicks);
		}

		// if there is an audio gap
		if (audioGapInSamples >= m_audioGapThresholdInSamples) {
			// write a null audio sample to fill the gap
			MP4WriteSample(
				m_mp4File,
				audioTrackId,
				NULL,
				0,
				audioGapInSamples);

			if (doRawAudioFrame) {
				m_rawAudioDuration += audioGapInTicks;
			} else {
				m_encodedAudioDuration += audioGapInTicks;
			}
		}

		// write the audio frame
		MP4WriteSample(
			m_mp4File,
			audioTrackId,
			(u_int8_t*)pFrame->GetData(), 
			pFrame->GetDataLength(),
			pFrame->ConvertDuration(m_audioTimeScale));

		if (doRawAudioFrame) {
			m_rawAudioFrameNumber++;
#if 0
			m_rawAudioDuration += 
				pFrame->ConvertDuration(TimestampTicks);
#else
			m_rawAudioDuration = pFrame->GetTimestamp() - m_movieStartTimestamp;
			m_rawAudioDuration += pFrame->ConvertDuration(TimestampTicks);
#endif
		} else {
			m_encodedAudioFrameNumber++;
#if 0
			m_encodedAudioDuration += 
				pFrame->ConvertDuration(TimestampTicks);
#else
			m_encodedAudioDuration = pFrame->GetTimestamp() - m_movieStartTimestamp;
			m_encodedAudioDuration += pFrame->ConvertDuration(TimestampTicks);
#endif
		}

	} else if (pFrame->GetType() == MPEG4VIDEOFRAME) {

		bool isIFrame = (MP4AV_Mpeg4GetVopType(
				(u_int8_t*)pFrame->GetData(), 
				pFrame->GetDataLength()) 
			== 'I');

		// ensure we start recording with an I frame
		if (m_encodedVideoFrameNumber == 1) {
			if (!isIFrame) {
			  if (pFrame->RemoveReference())
			    delete pFrame;
			  return;
			}

			m_encodedVideoStartTimestamp = pFrame->GetTimestamp();
			m_movieStartTimestamp = m_encodedVideoStartTimestamp;
			m_canRecordAudio = true;
		}

		MP4WriteSample(
			m_mp4File,
			m_encodedVideoTrackId,
			(u_int8_t*)pFrame->GetData(),
			pFrame->GetDataLength(),
			pFrame->ConvertDuration(m_videoTimeScale),
			0,
			isIFrame);
		
		m_encodedVideoFrameNumber++;
	}
	if (pFrame->RemoveReference())
	  delete pFrame;
	return;
}

void CMp4Recorder::DoStopRecord()
{
	if (!m_sink) {
		return;
	}

	bool optimize = false;

	// create hint tracks
	if (MP4_IS_VALID_TRACK_ID(m_encodedVideoTrackId)) {
	  create_mp4_video_hint_track(
					  m_mp4File, 
					  m_encodedVideoTrackId);
	}

	if (MP4_IS_VALID_TRACK_ID(m_encodedAudioTrackId)) {
	  create_mp4_audio_hint_track( 
		m_mp4File, m_encodedAudioTrackId);
	}

	// close the mp4 file
	MP4Close(m_mp4File);
	m_mp4File = NULL;

	// add ISMA style OD and Scene tracks
	bool useIsmaTag = false;

	// if AAC track is present, can tag this as ISMA compliant content
	useIsmaTag = m_makeIsmaCompliant;
	if (m_makeIod) {
	  MP4MakeIsmaCompliant(
				   (LPCTSTR) m_sFileName,
				   0,
				   useIsmaTag);
	}

	if (optimize) {
		MP4Optimize((LPCTSTR) m_sFileName);
	}

	m_sink = false;
}

// Extracted from video_encoder.cpp

MediaType CMp4Recorder::get_video_mp4_fileinfo (
				bool *createIod,
				bool *isma_compliant,
				uint8_t *videoProfile,
				uint8_t **videoConfig,
				uint32_t *videoConfigLen,
				uint8_t *mp4_video_type)
{
	*createIod = true;
	*isma_compliant = true;
	*videoProfile = VIDEO_PROFILE_ID;

	*videoConfig = (u_int8_t*)malloc(256);
	*videoConfigLen = 0;

	MP4AV_Mpeg4CreateVosh(
		videoConfig,
		videoConfigLen,
		// profile_level_id, default is 3, Simple Profile @ Level 3
		3);

	MP4AV_Mpeg4CreateVo(
		videoConfig,
		videoConfigLen,
		1);

	MP4AV_Mpeg4CreateVol(
		videoConfig,
		videoConfigLen,
		1,
		m_dFrameRate,
		true, // short time - true if we haven't set the # of bits
		true,	// variableRate
		m_uFrameWidth,
		m_uFrameHeight,
		0 // quantType, H.263
		);
		//&pConfig->m_videoTimeIncrBits);

	//free(pConfig->m_videoMpeg4Config);
	//pConfig->m_videoMpeg4Config = pMpeg4Config;
	//pConfig->m_videoMpeg4ConfigLength = mpeg4ConfigLength;

	//*videoConfig = pConfig->m_videoMpeg4Config;
	//*videoConfigLen = pConfig->m_videoMpeg4ConfigLength;
	if (mp4_video_type) {
	  *mp4_video_type = MP4_MPEG4_VIDEO_TYPE;
	}

	return MPEG4VIDEOFRAME;
}

void create_mp4_video_hint_track (
				  MP4FileHandle mp4file,
				  MP4TrackId trackId)
{
    MP4AV_Rfc3016Hinter(mp4file, 
			trackId,
			RTP_PAYLOAD_SIZE);
}


class CMp4TrackWriter : public TimeKeepingWriter {
private:
	unsigned int Write(
		const void * buf, 
		unsigned int numBytes,
		Timestamp timestamp,
		Duration duration
		);

	MediaType m_MediaType;
	CMp4Recorder & m_Mp4Recorder;

public:
	CMp4TrackWriter(CMp4Recorder & mp4Recorder, MediaType mediatype):
	m_Mp4Recorder(mp4Recorder),
	m_MediaType(mediatype)
	{}

	void DeclareWritingEnded();

	virtual ~CMp4TrackWriter() {}
};

unsigned int CMp4TrackWriter::Write(
	const void * buf, 
	unsigned int numBytes,
	Timestamp timestamp,
	Duration duration
	)
{
	//m_Mp4Recorder.Wait(m_MediaType, timestamp + duration);

	m_Mp4Recorder.EnqueueFrame(
		new CMediaFrame(m_MediaType, buf, numBytes,
		timestamp, duration));

	//m_Mp4Recorder.ReleaseProgressSelection();

	return numBytes;
}

void CMp4TrackWriter::DeclareWritingEnded() {
	m_Mp4Recorder.DeclareWritingEnded(m_MediaType);
}

TimeKeepingWriter* CMp4Recorder::CreateTrackWriter(MediaType mediatype)
{
	return new CMp4TrackWriter(*this, mediatype);
}

void CMp4Recorder::DeclareWritingEnded(MediaType mediatype)
{
	switch (mediatype) {
	case AACAUDIOFRAME:
	case MP3AUDIOFRAME:
		m_pProgressSelection->DeclareInactive(1);
		m_bAudioDone = true;
		break;
	case MPEG4VIDEOFRAME:
		m_pProgressSelection->DeclareInactive(0);
		m_bVideoDone = true;
		break;
	}

	if (m_bAudioDone && m_bVideoDone) {
		Stop();
		SetEvent(m_hCompletionEvent);
		m_pProgressSelection->Release();
		delete m_pProgressSelection;
		m_pProgressSelection = NULL;
	}
	else
		m_pProgressSelection->Release();
}

void CMp4Recorder::Start()
{
	m_pProgressSelection = new CTimeBasedProgressSelection(2);
	ResetEvent(m_hCompletionEvent);

	StartThread();
	m_bAudioDone = false;
	m_bVideoDone = false;
	CMediaNode::Start();
}


void CMp4Recorder::Wait(MediaType mediatype, Timestamp nextTimeStamp)
{
	int iMediaIndex = -1;

	switch (mediatype) {
	case AACAUDIOFRAME:
	case MP3AUDIOFRAME:
		iMediaIndex = 1;
		break;
	case MPEG4VIDEOFRAME:
		iMediaIndex = 0;
		break;
	default:
		ASSERT(false);
	}

	m_pProgressSelection->Wait(iMediaIndex, nextTimeStamp);
}

void CMp4Recorder::ReleaseProgressSelection()
{
	m_pProgressSelection->Release();
}
