/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is MPEG4IP.
 * 
 * The Initial Developer of the Original Code is Cisco Systems Inc.
 * Portions created by Cisco Systems Inc. are
 * Copyright (C) Cisco Systems Inc. 2000, 2001.  All Rights Reserved.
 * 
 * Contributor(s): 
 *		Dave Mackie		dmackie@cisco.com
 */

// PCH include
#include "StdAfx.h"

// Includes
#include "rgb2yuv.h"
#include "RGB2YUVConvertor.h"
#include "../TimeKeepingReaderWriter.h"
#include "../Common/MutexProtection.h"
#include <string>
#include <process.h>

using namespace std;

typedef unsigned char BYTE;

RGB2YUVConvertor::RGB2YUVConvertor(
		TimeKeepingReader * pReader, 
		TimeKeepingWriter * pWriter, 
		unsigned long lFrameWidth, 
		unsigned long lFrameHeight,
		double dDesiredFrameRate
		):
m_pReader(pReader),
m_pWriter(pWriter),
m_lFrameWidth(lFrameWidth),
m_lFrameHeight(lFrameHeight),
m_dDesiredFrameRate(dDesiredFrameRate),
m_pRGB(NULL),
m_pY(NULL),
m_pU(NULL),
m_pV(NULL),
m_lFramesWritten(0),
m_hThreadTermination(CreateEvent(NULL, FALSE, TRUE, NULL))
{
	AllocateBuffers();
	StartConversionThread();
}

RGB2YUVConvertor::~RGB2YUVConvertor() {

	DWORD dwResult = WaitForSingleObject(m_hThreadTermination, INFINITE);
	ASSERT(dwResult == WAIT_OBJECT_0);

	CloseHandle(m_hThreadTermination);

	delete [] m_pRGB;
	delete [] m_pYUV;
}

void RGB2YUVConvertor::AllocateBuffers() {
	unsigned long nFramePixels = m_lFrameWidth * m_lFrameHeight;

	m_pRGB = new BYTE[nFramePixels * 3];
	m_pYUV = new BYTE[nFramePixels * 3 / 2];
	m_pY = m_pYUV;
	m_pU = m_pY + nFramePixels;
	m_pV = m_pU + nFramePixels / 4;
}

void RGB2YUVConvertor::ConvertAndWrite(Timestamp timestamp, Duration duration)
{
	unsigned long nFramePixels = m_lFrameWidth * m_lFrameHeight;

	RGB2YUV(m_lFrameWidth, m_lFrameHeight, m_pRGB, m_pY, m_pU, m_pV, false);

	m_pWriter->WriteBytes(m_pYUV, nFramePixels * 3 / 2, timestamp, duration);

	m_lFramesWritten++;
}

void RGB2YUVConvertor::DoConvert() 
{
	unsigned long nFramePixels = m_lFrameWidth * m_lFrameHeight;
	unsigned long uNumFrames(0);

	while(true) 
	{
		Timestamp timestamp;
		Duration duration;

		unsigned int nRead = 
			m_pReader->ReadBytes(m_pRGB, nFramePixels * 3, timestamp, duration); 

		if (!nRead)
			break;

		if ( m_dDesiredFrameRate > 0 )
		{
			long dNumExpectedFrames = 
				m_dDesiredFrameRate * (double) (int64_t) ( timestamp + duration ) / 
				TimestampTicks;

			if (uNumFrames >= dNumExpectedFrames)
				continue;
		}

		ConvertAndWrite(timestamp, duration);
	}

	m_pWriter->DeclareWritingEnded();
}

static void RGB2YUVConvertor_ThreadFunc(LPVOID pParam) {
	
	RGB2YUVConvertor * pConvertor = (RGB2YUVConvertor*) pParam;
	ThreadCompletionAnnouncement announce(pConvertor->m_hThreadTermination);

	try {
		pConvertor->DoConvert();
	}
	catch (string * errMess) {
		AfxMessageBox(errMess->c_str());
		delete errMess;
		return;
	}
	catch (...) {
		AfxMessageBox(_T("Unhandled exception in RGB-2-YUV conversion thread"));
	}

}

void RGB2YUVConvertor::StartConversionThread() {
	_beginthread(RGB2YUVConvertor_ThreadFunc, 0, this);
}
